/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

/**
 * <p>This class contains some general functions for reading values
 * from a byte array and for writing values to a byte array.</p>
 *
 * <p>Note that the functionality in this class closely resembles that
 * of java.lang.reflect.Array as well as java.nio.ByteBuffer. It is
 * not entirely clear why neither of these built in classes are used
 * instead of code in this class. In the case of java.nio.ByteBuffer,
 * Java NIO was not added to the JDK until version 1.4; the code in
 * this class predates version 1.4. Also, deployment of HyperCast to
 * hand-held computers is limited by the fact that such platforms do
 * not, as of this writing, support Java features after 1.2. In the
 * case of java.lang.reflect.Array, perhaps the reflection was seen as
 * a performance penalty.</p>
 *
 * <h3>Bitwise Arithmetic</h3>
 *
 * <p>A note on Java bitwise arithmetic: The code in this class does a
 * lot of arithmetic using byte values. Care must be used in such
 * operations since Java "sign-extends" values when promoting a
 * smaller type to a larger type. For example, the byte value of 0x80
 * will be 0xFF80 if cast to a short, whereas the byte value 0x7F will
 * be 0x007F if cast to a short. Said another way: the value of the
 * high order bit of a byte is very important/tricky. As a result, the
 * code in this class contains many explicit bitwise AND operations
 * that clearly specify which bits will be retained so that
 * sign-extended values to not cause incorrect results. This means
 * that when converting the byte value to a larger type, e.g short, an
 * explicit bitwise AND operation will be used as in:</p>
 *
 * <p><code>(data[index] & 0xFF) << 8</code></p>
 *
 * <p>Although this bitwise AND seems superfluous it is not; if
 * <code>data[index]</code> had its high order bit set to 1, the
 * result could be incorrect.</p>
 *
 * <h3>Unsigned Values</h3>
 *
 * <p>A note about signed and unsigned values: The Java Programming
 * Language 3rd ed. section 6.2 "Types and Literals" has a paragraph
 * concerning unsigned types. It says that a Java program needing to
 * use unsigned types should use a "larger signed type" for
 * arithmetic. HyperCast does not fully espouse this doctrine, rather
 * when storing an unsigned quantity (i.e. an instance variable of an
 * object) HyperCast uses a type that is the same size as the unsigned
 * quantity that it is using. For example, HyperCast will use a
 * <code>short</code> (2 bytes) to represent a 2 byte unsigned
 * quantity instead of the larger signed type of <code>int</code>, or
 * an <code>int</code> to store an unsigned 4 byte quantity rather
 * than a <code>long</code> (note that the Java built-in 2 byte
 * unsigned <code>char</code> type is never used in HyperCast). This
 * means that HyperCast must be careful when performing arithmetic on
 * quantities that are interpretted as unsigned. The arithmetic
 * operations must use the next larger signed type and then convert
 * back to the smaller type. An example of this is seen in
 * <code>OL_Message.decrementHopLimit()</code>.</p>
 *
 * @see OL_Message#decrementHopLimit()
 *
 * <p>The <code>ByteArrayUtility</code> class provides several methods
 * that claim to read "unsigned" values from a byte array. These
 * methods conform to the "larger signed type" idea described
 * above. That is, they read an amount of data from a byte array that
 * is smaller than the type that they return, this ensures that the
 * high order bit is not used so negative numbers cannot be
 * represented. For example, an unsigned integer can be created using
 * <code>toUnsignedInt</code> from 1, 2, or 3, but not 4 bytes. Using
 * a fourth byte would allow the high order bit of the integer to
 * indicate a negative value. (NB in Java 2's complement is used so
 * there is no "sign bit" per se as in "signed-magnitude" bit
 * encoding, nevertheless, all negative values in 2's complement have
 * the high order bit set to 1).</p>
 * 
 * @author HyperCast Team
 * @author Tyler Beam
 * @author Dongwen Wang
 * @author Michael Nahas
 * @author Greg Mattes
 * @version September 2004
 */
public class ByteArrayUtility
{
    /**
     * Converts two bytes in a byte array to a signed short integer.
     *
     * @param data a byte array having a two byte short quantity encoded beginning at position <code>index</code>.
     * @param index position in the array <code>data</code> where a two byte short quantity is encoded.
     *
     * @return a short integer based on the value of the two bytes beginning at position <code>index</code> of the array <code>data</data>.
     * @throws ArrayIndexOutOfBoundsException if the value of <code>index</code> causes an illegal array access.
     */
    public static short toShort (final byte [] data, final int index)
    {
        return (short)
            (((data[index + 0] & 0xFF) << (8*1)) | 
             ((data[index + 1] & 0xFF) << (8*0)));
    }

    /**
     * Converts a byte in a byte array to an unsigned short
     * integer. Can be thought of as an unsigned byte contained in the
     * next larger integral type, i.e. short.
     *
     * @param data a byte array having a one byte unsigned short quantity encoded at position <code>index</code>.
     * @param index position in the array <code>data</code> where a one byte unsigned short quantity is encoded.
     *
     * @return a short integer based on the value of the byte at position <code>index</code> of the array <code>data</data>.
     * @throws ArrayIndexOutOfBoundsException if the value of <code>index</code> causes an illegal array access.
     */
    public static short toUnsignedShort (final byte [] data, final int index)
    {
        return (short) (data[index] & 0xFF);
    }

    /**
     * Converts an short integer to a byte array of length 2.
     * @param value a short integer
     * @return a byte array representing the short integer
     */
    public static byte [] toByteArray (final short value)
    {
        byte [] result = new byte [2];

        result[1] = (byte) ((value >>> (8*0)) & 0xFF);
        result[0] = (byte) ((value >>> (8*1)) & 0xFF);
        
        return result;
    }

    /**
     * Converts four bytes of a byte array to an integer.
     *
     * @param data a byte array having an integer quantity encoded beginning at position <code>index</code>.
     * @param index position in the array <code>data</code> where an integer quantity is encoded.
     *
     * @return an integer based on the value of the specifed bytes beginning at position <code>index</code> of the array <code>data</data>.
     *
     * @throws ArrayIndexOutOfBoundsException if the value of <code>index</code> causes an illegal array access.
     */
    public static int toInteger (final byte [] data, final int index)
    {
        return (int)
            (((data[index + 0] & 0xFF) << (8*3)) | 
             ((data[index + 1] & 0xFF) << (8*2)) |
             ((data[index + 2] & 0xFF) << (8*1)) |
             ((data[index + 3] & 0xFF) << (8*0)));
    }

    /**
     * Converts one, two, or three bytes of a byte array to an
     * unsigned integer. Can be thought of as using a larger, i.e. 32
     * bit integer, integral type to hold an unsigned 8 bit, 16 bit, or 24 bit value.
     *
     * @param data a byte array having an integer quantity encoded beginning at position <code>index</code>.
     * @param index position in the array <code>data</code> where an integer quantity is encoded.
     *
     * @param length the number of bytes comprising the integer data,
     * can be 1, 2, or 3. Anything longer than 3 bytes will allow
     * negative integer values to be encoded, breaking the contract
     * that an unsigned value is returned.
     *
     * @return an integer of <code>length</code> bytes based on the
     * value of the specifed bytes beginning at position
     * <code>index</code> of the array <code>data</data>.
     *
     * @throws IllegalArgumentException if length is not 1, 2, or 3.
     * @throws ArrayIndexOutOfBoundsException if the values of <code>index</code> and <code>length</code> cause an illegal array access.
     */
    public static int toUnsignedInteger (final byte [] data, final int index, final int length)
    {
        int result = 0;
        int offset = index;
        switch (length)
        {
            case 3: // 24 bit quantity
                result |= (data[offset] & 0xFF) << 16;
                ++offset;
                /* fall-through */

            case 2: // 16 bit short quantity
                result |= (data[offset] & 0xFF) << 8;
                ++offset;
                /* fall-through */

            case 1: // 8 bit byte quantity
                result |= (data[offset] & 0xFF) << 0;
                return result;

            default:
                throw new IllegalArgumentException ("Can only convert 1, 2, or 3 bytes to unsigned int");
        }
    }

    /**
     * Converts an integer to a byte array of length 4
     * @param value an integer
     * @return a byte array representing the integer
     */
    public static byte [] toByteArray (final int value)
    {
        byte[] result = new byte[4];

        result[3] = (byte) ((value >>> (8*0)) & 0xFF);
        result[2] = (byte) ((value >>> (8*1)) & 0xFF);
        result[1] = (byte) ((value >>> (8*2)) & 0xFF);
        result[0] = (byte) ((value >>> (8*3)) & 0xFF);
        
        return result;
    }

    /**
     * Converts a byte array to a long integer.
     * @param data a byte array
     * @param index a starting index in the array
     * @return a long integer
     * @throws ArrayIndexOutOfBoundsException if the value of <code>index</code> causes an illegal array access.
     */
    public static long toLong (final byte [] data, final int index)
    {
        return (long)
            (((data[index + 0] & 0xFFL) << (8*7)) | 
             ((data[index + 1] & 0xFFL) << (8*6)) |
             ((data[index + 2] & 0xFFL) << (8*5)) |
             ((data[index + 3] & 0xFFL) << (8*4)) |
             ((data[index + 4] & 0xFFL) << (8*3)) |
             ((data[index + 5] & 0xFFL) << (8*2)) |
             ((data[index + 6] & 0xFFL) << (8*1)) |
             ((data[index + 7] & 0xFFL) << (8*0)));
    }

    /**
     * Converts between one and seven bytes of a byte array to an
     * unsigned long. Can be thought of as using a larger, i.e. 64 bit
     * long, integral type to hold an unsigned quantity represented as
     * an array of size ranging fom one to seven bytes.
     *
     * @param data a byte array having a long quantity encoded beginning at position <code>index</code>.
     * @param index position in the array <code>data</code> where an long quantity is encoded.
     *
     * @param length the number of bytes comprising the long data,
     * can be 1-7 inclusive. Anything longer than 7 bytes will allow
     * negative long values to be encoded, breaking the contract
     * that an unsigned value is returned.
     *
     * @return a long of <code>length</code> bytes based on the value
     * of the specifed bytes beginning at position <code>index</code>
     * of the array <code>data</data>.
     *
     * @throws IllegalArgumentException if length is not in the range 1-7 inclusive.
     * @throws ArrayIndexOutOfBoundsException if the values of <code>index</code> and <code>length</code> cause an illegal array access.
     */
    public static int toUnsignedLong (final byte [] data, final int index, final int length)
    {
        int result = 0;
        int offset = index;
        switch (length)
        {
            case 7: // 56 bit quantity
                result |= (data[offset] & 0xFF) << (8*6);
                ++offset;
                /* fall-through */

            case 6: // 48 bit quantity
                result |= (data[offset] & 0xFF) << (8*5);
                ++offset;
                /* fall-through */

            case 5: // 40 bit short quantity
                result |= (data[offset] & 0xFF) << (8*4);
                ++offset;
                /* fall-through */

            case 4: // 32 bit byte quantity
                result |= (data[offset] & 0xFF) << (8*3);
                ++offset;
                /* fall-through */

            case 3: // 24 bit quantity
                result |= (data[offset] & 0xFF) << (8*2);
                ++offset;
                /* fall-through */

            case 2: // 16 bit short quantity
                result |= (data[offset] & 0xFF) << (8*1);
                ++offset;
                /* fall-through */

            case 1: // 8 bit byte quantity
                result |= (data[offset] & 0xFF) << (8*0);
                return result;

            default:
                throw new IllegalArgumentException ("Can only convert 1-7 bytes to unsigned long");
        }
    }

    /**
     * Converts a long to a byte array
     * @param value  a long integer
     * @return a byte array representing the long integer
     */
    public static byte [] toByteArray (final long value)
    {
        byte[] result = new byte[8];

        result[7] = (byte) ((value >>> (8*0)) & 0xFF);
        result[6] = (byte) ((value >>> (8*1)) & 0xFF);
        result[5] = (byte) ((value >>> (8*2)) & 0xFF);
        result[4] = (byte) ((value >>> (8*3)) & 0xFF);
        result[3] = (byte) ((value >>> (8*4)) & 0xFF);
        result[2] = (byte) ((value >>> (8*5)) & 0xFF);
        result[1] = (byte) ((value >>> (8*6)) & 0xFF);
        result[0] = (byte) ((value >>> (8*7)) & 0xFF);
        
        return result;
    }

    /**
     * Checks out if the specified part of a byte array is all zero.
     * @param data    a byte array
     * @param offset   the start position to check
     * @param length  the total length of bytes to be checked
     * 
     * @return true, if all the bytes checked out to be all zero, false if not
     * @throws ArrayIndexOutOfBoundsException if the value of <code>index</code> causes an illegal array access.
     */
    public static boolean isBytesZero (final byte [] data, final int offset, final int length)
    {
        int index = offset;
        int stop = index + length;
        while (index < stop)
        {
            if (data [index++] != 0)
            {
                return false;
            }
        }
        return true;
    }

    /**
     * Sets a specified part of a byte array to all zeros.
     *
     * @param data    a byte array
     * @param offset   starting position
     * @param length  number of bytes to set to zero
     * @throws ArrayIndexOutOfBoundsException if the value of <code>index</code> causes an illegal array access.
     */
    public static void setBytesZero (final byte [] data, final int offset, final int length)
    {
        // are either of these faster than the while loop? They both carry the overhead of object creation.
        // data = new byte [length];
        // - or -
        // System.arraycopy (new byte [length], 0, data, index, length);

        int index = offset;
        int stop = index + length;
        while (index < stop)
        {
            data [index++] = 0;
        }
    }

    /**
     * Concatenates two byte arrays into one new byte array
     *
     * @param source1 the first byte array (will be at the beginning of the new array)
     * @param source2 the second byte array (will be at the end of the new array)
     * @return the concatenation of the two arrays
     * @throws ArrayIndexOutOfBoundsException if the value of <code>index</code> causes an illegal array access.
     */
    public static byte [] concatenate (final byte [] source1, final byte [] source2)
    {
        byte [] result = new byte [source1.length + source2.length];

        System.arraycopy (source1, 0, result, 0, source1.length);
        System.arraycopy (source2, 0, result, source1.length, source2.length);

        return result;
    }

    /** Length of byte primitive type in bits */
    public static final int SIZEOF_BYTE  =  8; // bits

    /** Length of short primitive type in bits */
    public static final int SIZEOF_SHORT = 16; // bits

    /** Length of int primitive type in bits */
    public static final int SIZEOF_INT   = 32; // bits

    /** Length of long primitive type in bits */
    public static final int SIZEOF_LONG  = 64; // bits

    /**
     * Creates a string that represents the bit pattern of the specified byte value.
     *
     * @param x A byte value
     * @return A string representing the bit pattern of the specified byte.
     */
    public static String byteToBits (final byte x)
    {
        return generalToBits ((long) x, SIZEOF_BYTE);
    }

    /**
     * Creates a string that represents the bit pattern of the specified short value.
     *
     * @param x A short value
     * @return A string representing the bit pattern of the specified short.
     */
    public static String shortToBits (final short x)
    {
        return generalToBits ((long) x, SIZEOF_SHORT);
    }

    /**
     * Creates a string that represents the bit pattern of the specified int value.
     *
     * @param x A int value
     * @return A string representing the bit pattern of the specified int.
     */
    public static String intToBits (final int x)
    {
        return generalToBits ((long) x, SIZEOF_INT);
    }

    /**
     * Creates a string that represents the bit pattern of the specified long value.
     *
     * @param x A long value
     * @return A string representing the bit pattern of the specified long.
     */
    public static String longToBits (final long x)
    {
        return generalToBits (x, SIZEOF_LONG);
    }

    private static String generalToBits (final long x, final int typesize)
    {
        StringBuffer result = new StringBuffer (typesize);

        for (int i = (typesize - 1); i >= 0; --i)
        {
            long mask = 1L << i;
            boolean isBitOn = (x & mask) != 0;
            String nextBit = isBitOn ? "1" : "0";
            result.append (nextBit);
        }
        return result.toString();
    }

    public static void main (final String args [])
    {
        System.out.println ("Testing for ByteArrayUtility");


        byte b0 = (byte) 0xFF;
        short s0 = (short) b0;

        System.out.println ("Byte b0: " + ByteArrayUtility.byteToBits (b0));
        System.out.println ("Short s0: " + ByteArrayUtility.shortToBits (s0));

        byte b1 = (byte) 0x7F;
        short s1 = (short) b1;

        System.out.println ("Byte b1: " + ByteArrayUtility.byteToBits (b1));
        System.out.println ("Short s1: " + ByteArrayUtility.shortToBits (s1));


        byte [] longarray = new byte []
            {
                (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x00,
                (byte) 0x00, (byte) 0x00, (byte) 0x00, (byte) 0x80,
            };

        for (int i = 0; i < longarray.length; ++i)
            System.out.print (Long.toHexString (longarray[i] & 0xFFL) + ":");
        System.out.println ();

        long l = ByteArrayUtility.toLong (longarray, 0);
        System.out.println ("Decoded this long: " + l);
        System.out.println ("long to hex: " + Long.toHexString (l));
        System.out.println ("long to bits: " + ByteArrayUtility.longToBits (l));

        long l2 = Long.decode ("0x0000008000000080").longValue();
        System.out.println ("Decoded this long: " + l2);
        System.out.println ("long to hex: " + Long.toHexString (l2));
        System.out.println ("long to bits: " + ByteArrayUtility.longToBits (l2));


        // 0x83: 1000:0011
        // 0x0A: 0000:1010
        // 0x00: 0000:0000
        // 0x5D: 0101:1101
        //byte [] array = new byte [] { (byte) 0x83, (byte) 0x0A, (byte) 0x00, (byte) 0x5D };
        //byte [] array = new byte [] { (byte) 0x83, (byte) 0xD0, (byte) 0x82, (byte) 0x5D };
        //byte [] array = new byte [] { (byte) 0x83, (byte) 0x81, (byte) 0xaf, (byte) 0x5D };
        //byte [] array = new byte [] { (byte) 0x83, (byte) 0x80, (byte) 0x71, (byte) 0x5D };
        //byte [] array = new byte [] { (byte) 0x83, (byte) 0xa6, (byte) 0xeb, (byte) 0x5D };
        byte [] array = new byte [] { (byte) 0x83, (byte) 0x80, (byte) 0x80, (byte) 0x80 };

        short decodedShort = ByteArrayUtility.toShort (array, 1);
        System.out.println ("Decoded this short: " + decodedShort);
        System.out.println ("short to bits: " + ByteArrayUtility.shortToBits (decodedShort));

        // 2^7 + 2^1 + 2^0 = 128 + 2 + 1 = 131
        //int decodedInt1 = ByteArrayUtility.toUnsignedInteger (array, 0, 1);
        //System.out.println ("Decoded this int: " + decodedInt1);

        // 2^15 + 2^9 + 2^8 + 2^3 + 2^1 = 32768 + 512 + 256 + 8 + 2 = 33546
        int decodedInt2 = ByteArrayUtility.toUnsignedInteger (array, 0, 3);
        System.out.println ("Decoded this int: " + decodedInt2);
        System.out.println ("int to bits: " + ByteArrayUtility.intToBits (decodedInt2));

        // 2^23 + 2^17 + 2^16 + 2^11 + 2^9 = 8388608 + 131072 + 65536 + 2048 + 512 = 8587776
        //int decodedInt3 = ByteArrayUtility.toUnsignedInteger (array, 0, 3);
        //System.out.println ("Decoded this int: " + decodedInt3);

        System.out.println ("Hex: " + Long.toHexString (53378));
        System.out.println ("Hex: " + Long.toHexString (33199));
        System.out.println ("Hex: " + Long.toHexString (32881));
        System.out.println ("Hex: " + Long.toHexString (42731));

// 53378 -> 65410 : 15 14 12 7 1 -> 1101:0000:1000:0010 -> 0xD0 0x82
// 33199 -> 65455 0x81af
// 32881 ok       0x8071
// 42731 -> 65515 0xa6eb

/*
        long port = Long.parseLong(args[0]);
      
        byte[] bytePort = ByteArrayUtility.toByteArray (port);
        for (int i = 0; i < bytePort.length; i++)
            System.out.println("[" + i + "] = " + bytePort[i]);
        long read_port = (ByteArrayUtility.toLong(bytePort, 0));
        System.out.println("read_port = " + read_port);
*/
    }
}
