/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

import hypercast.*;

import java.net.*;
import java.util.*;

/** A logical address on the Delaunay triangulation.
 * It extends DT_Point.
 * 
 * It is immutable.
 *
 * @author HyperCast Team
 * @author Michael Nahas
 * @version	2.0, Feb. 11, 2001
 * @author Jianping Wang
 * @version 3.0, Jan. 08, 2004 
 */

final public class DT_LogicalAddress extends DT_Point implements I_LogicalAddress 
{
    /** length of byte array required to hold the address */
private final static Random rand = new Random();
    
public final static int ADDRESS_ARRAY_SIZE = 8;

/** constructor with the coodinates pair */
public DT_LogicalAddress(int xx, int yy) 
    {
        super(xx,yy);
    }
    
public DT_LogicalAddress(DT_LogicalAddress dtla) 
    {
        super(dtla);
    }

/** Reads coordinates from a byte array.
 * It reads 8 bytes - the first 4 making the x-coordinate, the last 4 making the y-coordinate.
 */
public DT_LogicalAddress(byte[] array, int offset) 
    {
        super(ByteArrayUtility.toInteger(array, offset), ByteArrayUtility.toInteger(array, offset+4));
    }



/** Reads coordinates from string in format "\d+,\d+".  (e.g. "12345,67890")
 */
public DT_LogicalAddress(String s) 
    {
        super(s);
    }
    
/** Creates coordinate such that if 2 nodes are on the same network, they have close coordinates.
 * This is accomplished by taking every even-number bit in the IP_Address to make the x-coordinate,
 * and every odd-numbered bit to make the y-coordinate.
 * Note, implication does not go the other way.  (i.e. if coordinates are close, it does not mean nodes 
 * are on the same network.)
 */

public DT_LogicalAddress(INETv4OnePort a) 
    {
        //To run multiple nodes on a host, enlarge the range of x and y coordinate
        //super( everyOtherBit(a, true), everyOtherBit(a, false) );
        super( everyOtherBit(a, true)*50 + rand.nextInt() % 50, everyOtherBit(a, false)*50 +  rand.nextInt() % 50);
    }


/** This function converts the IP_Address to a long, and
 * then creates an integer made from every other bit of the IP_Address.
 * It is called by <code>DT_LogicalAddress(IP_Address)</code>
 */

private static int everyOtherBit(INETv4OnePort a, boolean evenBits) 
    {
        InetAddress ia = a.getInetAddress();
        byte[] b = ia.getAddress();

        // addressBites 0 to 7 are zero.
        //addressBits 8 to 39 contain the IP_Address
        // bits 40 to 55 are the port #.
        // bites 55 to 63 are random.
        int random8BitValue = Math.abs(rand.nextInt())%255;
        long addressBits = ((((((long) ByteArrayUtility.toInteger(b, 0)) & 0xFFFFFFFFL) << 16) + a.getPort()) << 8) +  random8BitValue;

        if (!evenBits)
        {
            addressBits >>>= 1;
        }

        int result = 0;
        for (int i = 0; i < 32; i++) 
        {
            result |= ((addressBits & 1) << i);
            addressBits >>>= 2;
        }

        return result;
    }

/*  
// Test program to make sure everyOtherBit works correctly.
// My outcome was 202508,9150976.
public static void main(String argv[]) 
    {
        try 
        {
            byte[] UVAWebserver = new byte[4];
            UVAWebserver[0] = (byte) 128;
            UVAWebserver[1] = (byte) 143;
            UVAWebserver[2] = (byte) 137;
            UVAWebserver[3] = (byte) 29;
            
            
            IP_Address pa = new IP_Address(UVAWebserver, 80);
            DT_LogicalAddress la = new DT_LogicalAddress(pa); 
        
            System.out.println("PA: " + pa.toString());
            System.out.println("LA: " + la.toString());


        }
        catch (Exception e) 
        {
            System.err.println("ERROR: Exception caught.");
        }
    }
*/    



/** Implements method of I_Address interface. */

public boolean greaterThan(I_Address a) 
    {
        if (a == null || !(a instanceof DT_LogicalAddress))
            throw new IllegalArgumentException("Attempted to compare DT_LogicalAddress with something of a different type!");
        DT_LogicalAddress dtla = (DT_LogicalAddress) a;
        
        return DT_Point.greaterThan(this, dtla);
    }


	/** Check if the given address is equal to this address. */

	public boolean equals(I_Address a) 
		{
			if (a == null || !(a instanceof DT_LogicalAddress))
				throw new IllegalArgumentException("Attempted to compare DT_LogicalAddress with something of a different type!");
			DT_LogicalAddress dtla = (DT_LogicalAddress) a;
        
			return super.equals(dtla);
		}


/** Returns the size of byte[] used to store address.
 * @return 8
 */
    
public int getSize() 
    {
        return ADDRESS_ARRAY_SIZE;
    }


/** Converts the object into a byte array for transfer in a message.
 * This is the exact inverse of the constructor that takes a byte array.
 */
public byte[] toByteArray() 
    {
        byte[] array = new byte[8];

        System.arraycopy(ByteArrayUtility.toByteArray(x), 0, array, 0, 4);
        System.arraycopy(ByteArrayUtility.toByteArray(y), 0, array, 4, 4);

        return array;
    }
}







