/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

/**
 * This class represents a point in two dimensions, where the coordinates
 * of the point are ints.
 *
 * It is immutable.
 *
 * @author HyperCast Team
 * @author Michael Nahas
 * @version	2.0, Feb. 11, 2001
 */

public class DT_Point
{
    protected int x;
    protected int y;
    
    public DT_Point(int xx, int yy)
    {
        x = xx;
        y = yy;
    }

    public DT_Point(DT_Point p)
    {
        x = p.x;
        y = p.y;
    }

/** Reads coordinates from string in format "\d+,\d+".  (e.g. "12345,67890")
 */

public DT_Point(String s) 
    {
        try 
        {
            x = Integer.parseInt(s.substring(0, s.indexOf(',')));
            y = Integer.parseInt(s.substring(s.indexOf(',')+1, s.length()));
        }
        catch (Exception e) 
        {
            throw new IllegalArgumentException("String did not match regular expression of a point. \"\\d+,\\d+\"");
        }
    }
    

/** Creates a string following the in x-coordinate,y-coordinate fashion. */
public String toString() 
    {
        return Integer.toString(x) +"," + Integer.toString(y);
    }


/** Compares a DT_Point to another object. 
 * Throws exception if the other object is not an instance of DT_Point.
 * Returns false if the other object is null.
 */
public boolean equals(Object o) 
    {
        if (o == null)  
            return false;
        
        if (!(o instanceof DT_Point))
            throw new IllegalArgumentException("DT_Point.equals() pass a object that was not a DT_Point!");
        
        DT_Point p = (DT_Point) o;
        return x == p.x && y == p.y;
    }

/** Overrides Object.hashCode so that this can be used as a key in a hashtable. */
    public int hashCode() 
    {
        return ((x ^ (y >> 16)) ^ (y << 16));
    }
    
public int getX() 
    {
        return x;
    }
    
public int getY() 
    {
        return y;
    }

/** Returns true if DT_Point <code>a</code> comes before <code>b</code> in
 * the total ordering of points.   A point is greater than another point if its y-coordinate
 * is greater or if the y-coordinates are the same and its x-coordinate is greater.
 */

public final static boolean greaterThan(DT_Point a, DT_Point b) {
    return (a.y > b.y) || ((a.y == b.y) && (a.x > b.x));
}


/** Returns the euclidean distance between two points. */
public static double distance(DT_Point a, DT_Point b)
  {
    long dx = (long) a.x - (long) b.x;
    long dy = (long) a.y - (long) b.y;
    
    return Math.sqrt(dx*dx + dy*dy);
  }





/** Returns the angle (in radians) of datum to horizontal ray going from the pivot in the 
* positive x direction.  
* @return value in range (-Math.PI, Math.PI)
*/
static public double angleToHorizon(DT_Point pivot, DT_Point datum)
  {
      return Math.atan2((long) datum.y - (long) pivot.y, (long) datum.x - (long) pivot.x);
  }


/* Returns the positive angle (in radians) of datum to a ray going from the pivot to
 * the reference.
* @ returns value in range (-Math.PI, Math.PI]
 */
static public double angleRelative(DT_Point reference, DT_Point pivot, DT_Point datum) 
    {
        double angle_reference_pivot_datum = angleToHorizon(pivot, datum) - angleToHorizon(pivot, reference);

        if (angle_reference_pivot_datum > Math.PI)
            return angle_reference_pivot_datum - 2.0 * Math.PI;
        else if (angle_reference_pivot_datum <= -Math.PI)
            return angle_reference_pivot_datum + 2.0 * Math.PI;
        else
            return angle_reference_pivot_datum;
    }

/* Returns true if datum lies on the circle inscribing the points a, b, and c.
 * (Actually, it affords an error of +/-DT_DoublePoint.EPSILON.)
 * If no circle exists (all points are colinear) then it returns false.
 */
static public boolean isOnCircle(DT_Point a, DT_Point b, DT_Point c, DT_Point datum) 
{
    DT_DoublePoint center = findCenterOfCircle(a, b, c);

    if (center == null) // a,b,c are colinear
        return false;

    double center_to_datum = DT_DoublePoint.distance(center, new DT_DoublePoint(datum));
    double center_to_a = DT_DoublePoint.distance(center, new DT_DoublePoint(a));
    

    if ( center_to_datum < center_to_a + DT_DoublePoint.EPSILON && 
        center_to_datum > center_to_a - DT_DoublePoint.EPSILON)
        return true;
    else
        return false;
}


/* Returns true if datum lies inside the circle inscribing the points a, b, and c.
 * There is no accountment for error here.
 * If no circle exists (all points are colinear) then it returns true.
 */
static public boolean isInCircle(DT_Point a, DT_Point b, DT_Point c, DT_Point datum) 
{
    
    DT_DoublePoint center = findCenterOfCircle(a, b, c);

    if (center == null) // a,b,c are colinear
        return true;


    if (DT_DoublePoint.distance(center, new DT_DoublePoint(datum)) < DT_DoublePoint.distance(center, new DT_DoublePoint(a)))
        return true;
    else
        return false;
}


/* Returns true if datum lies inside or on the circle inscribing the points a, b, and c.
 * There is no accountment for error here.
 * If no circle exists (all points are colinear) then it returns true.
 */
static public boolean isInOrOnCircle(DT_Point a, DT_Point b, DT_Point c, DT_Point datum) 
{
    
    DT_DoublePoint center = findCenterOfCircle(a, b, c);

    if (center == null) // a,b,c are colinear
        return true;


    if (DT_DoublePoint.distance(center, new DT_DoublePoint(datum)) < DT_DoublePoint.distance(center, new DT_DoublePoint(a)) + DT_DoublePoint.EPSILON)
        return true;
    else
        return false;
}



/** Find the center of the circle inscribing 3 points.  
 * If all 3 points are the same, it returns that point.
 * Else, if 3 points lie on a line, it returns null.
 */

public static DT_DoublePoint findCenterOfCircle(DT_Point a, DT_Point b, DT_Point c) 
{
    // The center is the point of intersection of any 2 of the 3 perpendicular 
    // bisectors of line segments connecting a, b, and c.
    // We now select 2 line segments so that their perpendicular bisectors
    // do not have infinite slopes.
    // p1, p2 are the endpoints of the 1st line segment
    // q1, q2 are the endpoints of the 2nd line segment

	DT_Point p1, p2;
	DT_Point q1, q2;
    if (a.y == b.y && b.y == c.y) // all colinear
    {
        if (a.x == b.x && b.x == c.x)  // all the same point!  
            return new DT_DoublePoint(a); 
        else
            return null;
    }
    else if (a.y != b.y && a.y != c.y) 
    {
	    p1 = a;
	    p2 = b;
	    q1 = a;
	    q2 = c;
    }
    else if (a.y != b.y && b.y != c.y) 
    {
	    p1 = a;
	    p2 = b;
	    q1 = b;
	    q2 = c;
    }
    else // a.y == b.y && b.y != c.y && a.y != c.y
    {
	    p1 = b;
	    p2 = c;
	    q1 = a;
	    q2 = c;
    }
    

    // calculate x-intercepts and slopes for two lines.
    // (NOTE: divide by zero should never happen, but
    // given that the slopes could be infinitely small,
    // I'll catch the expeption anyway.)

    try 
    {
        double f, s;   // x intercept and slope for first line

        s = ((long)p2.x - (long)p1.x) / (double) ((long)p1.y - (long)p2.y);  
        f = ((long)p1.y + (long)p2.y)/2.0 + (((long)p1.x + (long)p2.x)*-0.5*s);

        double g, t;   // x intercept and slope for second line

        t = ((long)q2.x - (long)q1.x) / (double) ((long)q1.y - (long)q2.y);  
        g = ((long)q1.y + (long)q2.y)/2.0 + (((long)q1.x + (long)q2.x)*-0.5*t);


        if (s == t)  // either 0 or infinite solutions...
            return null;
        else
            return DT_DoublePoint.findInterceptOfLines(f, s, g, t);
    }
    catch (ArithmeticException ae) 
    {
        System.err.println("Caught ArithmeticException in DT_Point.findCenterOfCircle: " + ae.getMessage());
        return null;  //seems safe to say there is no center...
    }
}
}










