/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

/**
 * This class defines the OL_Message extension header. It contains all the functions needed to 
 * maintain the Extension header. The header applies to OL_Message.FSM extension. 
 * 
 * @author HyperCast Team
 */

public class FSM_Extension extends Extension
{
    protected static final int EXTENSION_HEADER_LENGTH =
    2 + // Service number
    1 + // Extension type
    4;  // unique id

    
    protected static final int ID_HASH_LENGTH = 2;
    protected static final int ID_COUNTER_LENGTH = 2; // sizeof (int) unsigned
    
    protected static final int ID_LENGTH = ID_HASH_LENGTH + ID_COUNTER_LENGTH;
    protected static final int DELIVERY_MODE_LENGTH = 1; // TODO: shouldn't this be defined in OL_Message? delivery mode is really an attribute of an OL_Message, also the delivery mode constants are defined there.

    /*
     * Extension types
     */
    protected static final byte EXTENSION_TYPE_PAYLOAD = (byte)(1 << 7); // has application payload

    protected static final int SEQUENCE_NUMBER_LENGTH = 8; // sizeof (long)
    
    protected static final long MAX_SEQUENCE_NUMBER_VALUE = Long.MAX_VALUE ; // max value of long

    /** value of the streamid (uniqueid) field used with the naming service */
    protected static final byte [] NAMING_STREAM_ID =
    new byte []
    {
        (byte) 0x4E, // N
        (byte) 0x41, // A
        (byte) 0x4D, // M
        (byte) 0x45, // E
    };

    /** Type (8 bits) */
    private byte type;
    /** FSM ID */
    private short fsmid;

    /**
     * The unique identifier is message id for individual messages
     * and stream id for stream data.
     */
    private byte [] uniqueID;

    /** sequence number  */
    private long sequenceNumber = -1;

    /** sender logical address, only used for control messages */
    private byte[] senderAddress = null;

    /**delivery mode */
    private byte deliveryMode = OL_Message.DELIVERY_MODE_MULTICAST;

    /** extension length */
    private int extensionLength;

    /** The payload of the extension that appears after the extension header information */
    // TODO: maybe select a better name
    private byte [] controlData;

    /**
     * Constructs an Extension header object for Message-FSM-data
     * finite state machine id (2 bytes) type(1 byte) 
     * byte array of the message id (8 bytes) // TODO: I think that this comment is wrong, should be 4 byte 
     */
    public FSM_Extension (final short fsmid, final byte type, final byte [] uniqueID)
    {
        /*
         * check if a valid mesage id
         */
        if(uniqueID.length != ID_LENGTH)
            throw new HyperCastWarningRuntimeException("The message id has an unexpected length"); 
        /*
         * set the header information
         */
        this.fsmid 	= fsmid;
        this.type = type ;  
        this.uniqueID = uniqueID;
        this.extensionLength = EXTENSION_HEADER_LENGTH;
    }

    /**
     * Constructs an Extension header object for FSM-Control
     * finite state machine id (2 bytes) type(1 byte)
     * byte array of the message id (8 bytes), delivery mode (1 byte),the logical address of the root 
     */
    public FSM_Extension (final short fsmid, final byte type, final byte [] uniqueID, final byte deliveryMode, final I_LogicalAddress la)
    {
        /*
         * check if a valid mesage id
         */
        if(uniqueID.length != ID_LENGTH)
            throw new HyperCastWarningRuntimeException("The message id has an unexpected length"); 
        /*
         * set the header information
         */
        this.fsmid = fsmid;
        this.type = type ;
        this.uniqueID = uniqueID;
        this.senderAddress = la.toByteArray();
        this.deliveryMode = deliveryMode;
        this.extensionLength = EXTENSION_HEADER_LENGTH + DELIVERY_MODE_LENGTH +la.toByteArray().length;
    }

    /**
     * Constructs an Extension header object for FSM-Control
     * finite state machine id (2 bytes) type(1 byte)
     * byte array of the message id (8 bytes), delivery mode (1 byte),the logical address of the root 
     */
    public FSM_Extension (final short fsmid, final byte type, final byte [] uniqueID, final I_LogicalAddress la) 
    {
        /*
         * check if a valid mesage id
         */
        if (uniqueID.length != ID_LENGTH)
            throw new HyperCastWarningRuntimeException("The message id has an unexpected length"); 
        /*
         * set the header information
         */
        this.fsmid = fsmid;
        this.type = type ;
        this.uniqueID = uniqueID;
        this.senderAddress = la.toByteArray();
        this.extensionLength = EXTENSION_HEADER_LENGTH + DELIVERY_MODE_LENGTH +la.toByteArray().length;
    }

    /**
     * Constructs an Extension header object for Stream-Data
     * header(1 byte) type(1 byte) finite state machine id (2 bytes)
     * stream id length(1 byte) stream id(no fixed length),sequence number(8 bytes) 
     */
    public FSM_Extension (final short fsmid, final byte type, final byte [] uniqueID, final long sequenceNumber)
    {
        /*
         * check if a valid mesage id
         */
        if (uniqueID.length != ID_LENGTH)
            throw new HyperCastWarningRuntimeException ("The message id has an unexpected length");  
        /*
         * set the header information
         */
        this.fsmid = fsmid;
        this.type = type;
        this.uniqueID = uniqueID;
        this.sequenceNumber = sequenceNumber;
        this.extensionLength = EXTENSION_HEADER_LENGTH + SEQUENCE_NUMBER_LENGTH;
    }

    /**
     * Constructs an Extension header object for Stream-Control
     * finite state machine id (2 bytes)
     * type (1 byte)
     * unique id (no fixed length)
     * control payload (no fixed length) 
     */
    public FSM_Extension (final short fsmid, final byte type, final byte [] uniqueID, final byte [] controlData)
    {
        /*
         * check if a valid mesage id
         */
        if (uniqueID.length != ID_LENGTH)
            throw new HyperCastWarningRuntimeException ("The message id has an unexpected length"); 

        this.fsmid = fsmid;
        this.type = type;
        this.uniqueID = uniqueID;
        this.controlData = controlData;
        this.extensionLength = EXTENSION_HEADER_LENGTH + uniqueID.length + controlData.length;
    }

    /**
     * Constructs an Extension header object using byte array
     */
    protected FSM_Extension (final byte [] e)
    {
        /*
         * set the header information
         */
        /*
         * use the index to point to the current position in the byte array
         */
        int index = 0;

        /*
         * set finite state machine id
         */
        fsmid = ByteArrayUtility.toShort(e,index);
        index +=2;

        /*
         * set type
         */
        type = e[index];
        index += 1;
        
        /*
         * set the message id
         */
        uniqueID = new byte[ID_LENGTH];
        System.arraycopy(e,index,uniqueID,0,ID_HASH_LENGTH+ID_COUNTER_LENGTH); // TODO: use ID_LENGTH
        index +=ID_HASH_LENGTH+ID_COUNTER_LENGTH; // TODO: use ID_LENGTH

        /*
         * FSM_MESSAGE
         */
        if(isServiceForMessage(fsmid))
        {
            /*
             * set the logical address for control message
             */
            if (! isPayloadType (type))
            {
                deliveryMode = e[index];
                ++index;
                senderAddress = new byte[e.length-index];
                System.arraycopy(e,index,senderAddress,0,e.length-index);
            }
        }
        /*
         * STREAM
         */
        else if(isServiceForStream(fsmid))
        {
            /*
             * set the sequence number for payload message
             */
            if (isPayloadType (type))
            {
                sequenceNumber = ByteArrayUtility.toLong(e,index);
                index += SEQUENCE_NUMBER_LENGTH;
            }
            /*
             * set the payload of the control message
             */
            else
            {
                int controlDataLength = e.length - index;
                controlData = new byte [controlDataLength];
                System.arraycopy (e, index, controlData, 0, controlDataLength);
            }
        }
        extensionLength = e.length;
    }


	/**
	 * Constructs an SecurityExtension Extension from an existing one.
	 */
	public FSM_Extension(FSM_Extension se)
	{
		this.controlData = se.controlData;
		this.deliveryMode = se.deliveryMode;
		this.extensionLength = se.extensionLength;
		this.fsmid = se.fsmid;
		this.senderAddress = se.senderAddress;
		this.sequenceNumber = se.sequenceNumber;
		this.type = se.type;
		this.uniqueID = se.uniqueID;
	}


	/**
	 * Return a new FSM_Extension extension instance by cloning this extension.
	 */
	public Extension Clone()
	{
		return new FSM_Extension(this);
	}
				
	/**
	 * Returns the type of this extension.
	 */
	public byte getExtensionType() {
		return Extension.FSM;
	}
		

    /**
     * Return the byte array of the extension header
     */
    public byte [] toByteArray()
    {
        /*
         * calculate the length of the byte array
         */
        int length = 0;
        
        //message FSM
        if (isServiceForMessage(fsmid))
        {
            //payload message
            if (isPayloadType (type))
            {
                length += EXTENSION_HEADER_LENGTH;
            }
            //control message
            else
            {
                length += EXTENSION_HEADER_LENGTH + DELIVERY_MODE_LENGTH + senderAddress.length;
            }
        }
        //Stream ADU
        else if (isServiceForStream(fsmid))
        {
            //payload message
            if (isPayloadType (type))
            {
                length += EXTENSION_HEADER_LENGTH + SEQUENCE_NUMBER_LENGTH;
            }
            //control message
            else
            {
                length += EXTENSION_HEADER_LENGTH + controlData.length;
            }
        }
        
        /*
         * fill the byte array
         */
        byte [] result = new byte [length];
        int index = 0;

        //finite state machine id
        result[index+0] = (byte) ((fsmid >>> 8) & 0xff);
        result[index+1] = (byte) ((fsmid >>> 0) & 0xff);
        index +=2;

        //type
        result[index] = type;
        index += 1;
 
     	System.arraycopy(uniqueID, 0, result, index,ID_LENGTH);
		index += ID_LENGTH;

        //message FSM
        if (isServiceForMessage (fsmid))
        {
            //control message

            if (! isPayloadType (type))
            {
                result[index] = deliveryMode;
                ++index;
                System.arraycopy(senderAddress,0,result,index,senderAddress.length);
            }
        }
        else if (isServiceForStream (fsmid))
        {
            //payload message
            if (isPayloadType (type))
            {
                System.arraycopy (ByteArrayUtility.toByteArray (sequenceNumber), 0, result, index, SEQUENCE_NUMBER_LENGTH);
            }
            //control message
            else
            {
                System.arraycopy (controlData, 0, result, index, controlData.length);
            }
        }
        return result;
    }


	/*
	 * This method provides a way to get the length of the byte 
	 * array of an extension without actually executing toByteArray().
	 */
	/**
	 * Return the length of the byte array of this extension.
	 */
	public int getSize()
	{
		/*
		 * calculate the length of the byte array
		 */
		int length = 0;
        
		//message FSM
		if (isServiceForMessage(fsmid))
		{
			//payload message
			if (isPayloadType (type))
			{
				length += EXTENSION_HEADER_LENGTH;
			}
			//control message
			else
			{
				length += EXTENSION_HEADER_LENGTH + DELIVERY_MODE_LENGTH + senderAddress.length;
			}
		}
		//Stream ADU
		else if (isServiceForStream(fsmid))
		{
			//payload message
			if (isPayloadType (type))
			{
				length += EXTENSION_HEADER_LENGTH + SEQUENCE_NUMBER_LENGTH;
			}
			//control message
			else
			{
				length += EXTENSION_HEADER_LENGTH + controlData.length;
			}
		}

		return length;
	}




    /**
     * Provides access to the control message payload of a stream
     * store control message.
     *
     * @return An array of bytes that represents the stream control
     * data, null if not a stream control message
     */
    protected byte [] getControlData()
    {
        return controlData;
    }
    
    // NAMING
    protected void setControlData (final byte [] controlData)
    {
        this.controlData = controlData;
    }
    
    /**
     * Returns the extension type
     */
    protected byte getType() {
        return type;
    }

    private boolean isPayloadType (final byte type)
    {
        return ((type >>> 7) & 0x1) == 1;
    }

    private boolean isServiceForMessage (final short serviceNumber)
    {
        return (serviceNumber == MessageStore.FSM_H2H_ACK
                ||
                serviceNumber == MessageStore.FSM_E2E_ACK
                ||
                serviceNumber == MessageStore.FSM_SYNCHRONIZATION
                ||
                serviceNumber == MessageStore.FSM_DUPLICATE_ELIMINATION
                ||
                serviceNumber == MessageStore.FSM_INCAST);
    }

    private boolean isServiceForStream (final short serviceNumber)
    {
        return (serviceNumber == MessageStore.FSM_BEST_EFFORT_ORDERING
                ||
                serviceNumber == MessageStore.FSM_NAMING);
    }

    /**
     * Returns the extension finite state machine id
     */
    protected short getFSMID() {
        return fsmid;
    }

    /**
     * Returns the UniqueID
     */
    protected byte[] getUniqueID() {
        return uniqueID;
    }

    /**Returns the sender Address
     */
    protected byte[] getSenderAddress() {
        return senderAddress;
    }

    /**Returns the length of the extension
     */
    protected int getExtensionLength() {
        return extensionLength;
    }

    /**
     * @return The sequence number of the extension.
     */
    protected long getSequenceNumber() {
        return sequenceNumber;
    }

    /**
     * @return The delivery mode of the extension.
     **/
    protected byte getDeliveryMode() {
        return deliveryMode;
    }

    /**
     * Sets the extension Flag
     */
    protected void setType (final byte t) {
        type = t;
    }

    /**
     * set the extension finite state machine id
     */
    protected void setFSMID (final short FID) {
        fsmid = FID;
    }

    /**
     * Sets the uniqueID
     */
    protected void setUniqueID (final byte [] id) {
        uniqueID = id;
    }

    /**
     * Set sequence number
     */
    protected void setSequenceNumber (final long s) {
        sequenceNumber = s;
    }

    /**
     * Sets the delivery mode
     */
    protected void setDeliveryMode (final byte mode) {
        deliveryMode = mode;
    }

    protected String payloadToString()
    {
        final byte [] payload = toByteArray();
        final String result = new String (payload);
        return result;
    }
}
