/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.events.NOTIFICATION_EVENT;

/**
 * This interface defines the operations that take place based on
 * finite machine type, not finite machines instances. Such operations
 * are needed in the MessageStore because there are various places
 * where decisions must be made regarding a message depending on the
 * finite state machine type.
 *
 * In order to implement these operations follow this convention:
 *
 * (1) In the file that implements your finite state machine create a
 * second (non-inner) class that implements this interface, we'll
 * refer to this as "FiniteStateMachineMetaOperationsChild" for the
 * purposes of this example.
 *
 * (2) Provide appropriate code for each method.
 *
 * (3) Add a static variable to your finite state machine class of
 * type "FiniteStateMachineMetaOperations."
 *
 * (4) Create a public static method in your finite state machine
 * class that returns the singleton instance of
 * "FiniteStateMachineMetaOperationsChild" initializing it if
 * necessary.
 *
 * (5) In MessageStore.java add a line of code that registers the new
 * MetaOperations object with the meta operations table.
 *
 * @author HyperCast Team
 * @author Greg Mattes
 * @version June 2004
 */
public interface FiniteStateMachineMetaOperations
{
    /**
     * This method returns true if the message store will be solely
     * responsible for forwarding this message. If the forwarding
     * engine should forward this message, this method returns false.
     */
    public boolean messageStoreWillForwardMessage (OL_Message message);

    /**
     * This method ensures that the settings of the message,
     * e.g. delivery method, is valid for the given service number.
     */
    public boolean isValidMessage (OL_Message message);

	/**
	 * This method returns true if the message store will process the intermediate
	 * unicast message.
	 */
	public boolean processIntermediateUnicastMessage(OL_Message message);
    /**
     * Returns a notification event that should be raised at message
     * send time. If no notification should be raised, this method
     * returns null.
     */
    public NOTIFICATION_EVENT notifyOnSend (OL_Message message);
}
