/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.HC;

import hypercast.*;

/**
 * This class encapsulates the functions of the Logical-Physical Address pair.
 * It provides both physical and logical address for a hypercube overlay node.
 * We also include the concept of age of the variable so that we can use this
 * same class in the Neigborhood View
 * 
 * @author HyperCast Team
 * @author Tyler Beam
 * @author by Dongwen Wang
 * @version	2.0, Feb. 27, 2001
 */

public final class HC_AddressPair implements I_AddressPair {
		/** The physical address of the protocol node */
	private I_PhysicalAddress PA;
		/** The logical address of the protocol node */
	private HC_LogicalAddress LA;
		/** The age of the protocol node (in HeartBeats).  Unlike
            humans, HC_AddressPairs start with a high value for Age
            and then it decreases over time.  If the Age ever reaches
            zero, the HC_AddressPair has become "stale".  HC_AddressPairs can be
            rejuvenated by a call to <code>setAge</code> or
            <code>resetAge</code>.*/
	private int Age;

	/** The HC_Node that created this object.  It is required for the
        values MaxAge and MaxMissingNeighbor. */
	private HC_Node parent;
	

	/** 
	 * Constructs HC_AddressPair from a HC_LogicalAddress, I_PhysicalAddress.
	 *	Just copies the arguments into internal objects and
	 *  sets the Age variable to MaxAge + MaxMissingNeighbor. Either LogicalAddress
	 *  or PhysicalAddress can be null. <br>
	 *  If the HC_AddressPair is an HRoot, then MissingNeighbor does not apply.
	 * @param	laddr	a HC_LogicalAddress object
	 * @param	paddr	an I_PhysicalAddress object
	 * @param	ishroot a flag indicating if the HC_AddressPair represents an HRoot
	 */
	HC_AddressPair(HC_Node hcnode, HC_LogicalAddress laddr, I_PhysicalAddress paddr, boolean ishroot) {
		
		parent = hcnode;

        LA = laddr;
        PA = paddr;
        
		if (ishroot)
			Age = parent.MaxAge;
		else
			Age = parent.MaxAge + parent.MaxMissingNeighbor;
	}



	/**
	 * Constructs a HC_AddressPair from a HC_AddressPair and a boolean flag.
	 *	Just copies the arguments into internal objects and
	 *	sets the Age variable to MaxAge + MaxMissingNeighbor <br>
	 *	If the HC_AddressPair is an HRoot, then MissingNeighbor does not apply.
	 * 
	 * @param	addrpair		HC_AddressPair object
	 * @param	ishroot	a flag indicating if the HC_AddressPair represents an HRoot
	 */

	HC_AddressPair(HC_Node hcnode, HC_AddressPair addrpair, boolean ishroot) {		
		parent = hcnode;
		
		if (addrpair.getLogicalAddress() == null) 
            LA = null;
		else 
            LA  = (HC_LogicalAddress)addrpair.getLogicalAddress();

		if (addrpair.getPhysicalAddress() == null) 
            PA = null;
		else 
            PA  = addrpair.getPhysicalAddress();
        
		if (ishroot)
			Age = parent.MaxAge;
		else
			Age = parent.MaxAge + parent.MaxMissingNeighbor;
	}



	/** 
	 * Ages a AddressPair entry and returns true if the AddressPair is still "alive".
	 * Decreases age by 1 if age is greater than 0, and then returns true
	 * if age is still greater than 0.
	 * @return	true if the data is not stale, and false, otherwise
	 */
	public boolean age() {
		if (Age > 0) 
            Age--;

		if (Age > 0) 
            return true; 
		else 
            return false;
	}

  
	/** 
	 * Resets Age to MaxAge + MaxMissingNeighbor.
	 */
	public void resetAge(){
		Age = parent.MaxAge + parent.MaxMissingNeighbor;
	}


	/** 
	 * Sets Age to the new age.
	 * @param	age		the new age
	 */
	public void setAge(int age) {
		Age = age;
	}


	/** 
	 * Gets the age of the node
	 * @return	age of the node
	 */
	public int getAge() {
		return Age;
	}



	/** 
	 * Gets a copy of the internal HC_LogicalAddress object,
	 * or null if the object is null.
	 * @return	copy of HC_LogicalAddress stored
	 */	
	public I_LogicalAddress getLogicalAddress() {
        return LA;
	}
	
	
	/**
	 * Gets a copy of the internal I_PhysicalAddress object,
	 * or null if the object is null.
	 * @return	copy of PhysicalAddress stored
	 */
	public I_PhysicalAddress getPhysicalAddress() {
        return PA;
	}

	
	
	
	/** 
	 * Gets the string format of the HC_AddressPair object
	 * return string format: nnn.nnn.nnn.nnn/nnnn:LA repeate
	 * @return  String representation of the object
	 * 
	 */	
    public String toString() {
		StringBuffer sb = new StringBuffer();

		//changed by Dongwen				
		if (PA != null)
		    sb.append(PA.toString());
		if (LA != null) {
			if (LA.getColexIndex() != 0x7fffffff) {
				sb.append(':');
				sb.append(LA.getColexIndex());
				sb.append(" ");
			}
		}
		
		return sb.toString();
    }
    public boolean equals(Object o){
    	if(o==null)return false;
    	if(!(o instanceof I_AddressPair))
    		return false;
    	I_AddressPair ap = (I_AddressPair)o;
    	
    	return LA.equals(ap.getLogicalAddress()) && PA.equals(ap.getPhysicalAddress());
    	
    }

	
}









