/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.HC;

import hypercast.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.*;

import java.util.*;

/**
 * This class encapsulates the HC_Neighborhood and do all functions regarding that <p>
 * Specifically, it should do the following jobs: <br>
 * 
 *  ageNeighborhood     : Should be able to age data and remove stale data <br>
 *  empty   : Should be able to check if the HC_Neighborhood is empty <br>
 *  register: It should be able to take a PhysicalAddress-LogicalAddress <
 *            pair and return true if it was able to install it/reset its
 *            age and false otherwise <br>
 *  remove  : Remove a neighbor if the physical address matches the one in
 *            the HC_Neighborhood entry <br>
 *  searchTear:
 *            Should return the HC_LogicalAddress of the lowest tear, and
 *            HC_Global.MaxAddress if none exists.  The tear reported is
 * 	          Ttimeout + Tmissing time units old <br>
 *  searchLittleTear:
 *            Should return the HC_LogicalAddress of the lowest tear, and
 *            HC_Global.MaxAddress if none exists.  This differs from searchTear in
 *	          that tears which are only Ttimeout old are reported <br>
 *  getNthNeighbor:
 *            Gets the HC_AddressPair of the Nth neighbor <br>
 *  resetNeighborhood:
 *			  Removes all entries from the neighborhood table <br>
 * 
 * @author HyperCast Team
 * @author Tyler Beam
 * @author Dongwen Wang
 * @version	2.0, Feb. 27, 2001
 */

public final class HC_Neighborhood
{
		/** AddressPair array of all neighbors */
	HC_AddressPair[] NView;
		/** AddressPair of the HRoot */
	HC_AddressPair HRoot; 
		/** HRoot sequence number */
	int HRootSeqNum;
		/** MissingNeighbor Time */
	int MNTimeOut;  
		/** Hypercast node object */
	private HC_Node parent;
	
	/**
	 * The reference to the StatsProcessor instance which handles the statistics in this object.
	 */
	protected StatsProcessor statsPro;
	
	/** 
	 * Constructs HC_Neighborhood objects with an initialization object 
	 * and a HC_Node object <p>
	 *  Initialize each of the entries in the HC_Neighborhood table to
	 *  be "unknown", that is, Physical Address is null in each of those
	 *  HC_AddressPair objects
	 * 
	 * @param	hcnode a HC_Node object with respect to which the neighborhood will be constructed.
	 */
	public HC_Neighborhood(HC_Node hcnode, StatsProcessor sPro) {
		parent = hcnode;
		
		statsPro = sPro;
		
		NView = new HC_AddressPair[HC_Node.MaxHypercubeDim];
		MNTimeOut = 0;

		// Initialize HC_Neighborhood.
		resetNeighborhood();
		
		for(int i=0;i<NView.length;i++){
			this.addNeighborToStatsProcessor(NView, i);
		}

		
	}

	/**
	 * Create a StatsElement instance for a neighbor and add it to the StatsProcessor object.
	 */
	private void addNeighborToStatsProcessor(HC_AddressPair[] neighbors, int index)
	{
		 statsPro.addStatsElement("NeighborTable", new HC_NeighborhoodStats(neighbors, index),  0/* minoccurs */, -1/* maxoccurs */);
	}
	
	/** 
	 * Checks the age of all the Neighborhood.
	 * Ages all the valid data and removes any data that has become stale
	 * @return	1 if all neighbors have expired, 0 otherwise
	 */
	public int ageNeighborhood() {
		int returnVal = 1;

		for (int i = 0; i < NView.length; i++) {
			if (NView[i] != null)  
            {
				if (!NView[i].age()) {
					NView[i] = null;
				}
				else 
                {
					returnVal = 0;
                }
            }
		}

		if (HRoot != null && !HRoot.age()) {
            HRoot = null;
        }

		return(returnVal);
	}




	/** 
	 * Checks if the neighborhood is empty.
	 * Checks if all the entries in the array are null. If yes, it returns 
	 * true, else it returns false.
	 * @return	true or false, depending on whether the neighborhood is empty or not
	 */
	public boolean empty() {
		for (int i = 0; i < NView.length; i++) {
			if (NView[i] != null)
				return false;
		}

		return true;
	}




	/** 
     * Returns the number of neighbors in neighbor table
	 */
	public int getNumOfNeighbors() {
        int count = 0;
		for (int i = 0; i < NView.length; i++) {
			if (NView[i] != null)
                count++;
		}

		return count;
	}



	/** 
	 * Registers the HC_AddressPair to the node's neighborhood table <p>
	 *  Algorithm: <br>
	 *  1. Calculate the index of the HC_LogicalAddress in the Neighborhood View <br>
	 *       (a) If it is not a neighbor return false <br>
	 *  2. Check if the index is occupied; <br>
	 *       (a) If it is, check if the PhysicalAddress is the same <br>
	 *              (i)  If it is, reset the age, and return true <br>
	 *              (ii) If it is not, return false <br>
	 *       (b) If it is not, install the Pair, and return true <br>
	 * 
	 * @param	addrpair	a LogicalAddress/PhysicalAddress pair
	 * @return	true or false
	 */
	public boolean register(HC_AddressPair addrpair) {
		int index = parent.MyLogicalAddress.findNeighborIndex((HC_LogicalAddress)addrpair.getLogicalAddress());

		if (index == HC_Node.MaxAddress()) 
            return false;

		// If the place is not occupied
		if ((NView[index] == null) || (NView[index].getPhysicalAddress()==null)) {
			NView[index] = new HC_AddressPair(parent, addrpair, false);
			return true;
		}

		// If it is occupied
		if (addrpair.getPhysicalAddress().equals(NView[index].getPhysicalAddress())) {
			NView[index].resetAge();
			return true;
		}

		return false;
	}

	

	/** 
	 * Updates HRoot with HRoot HC_LogicalAddress and sequence number
	 * @param	laddr		a HC_LogicalAddress object
	 * @param	curseqnum	current sequence number
	 * @return	true or false
	 */	
	public boolean updateHRoot(HC_LogicalAddress laddr, int curseqnum) {
		if ((curseqnum > HRootSeqNum) || (HRoot == null)) {
			HRootSeqNum = curseqnum;
			HRoot = new HC_AddressPair(parent, laddr, null, true);
			return true;
		}

		return false;
	}


	
	/** 
	 * Removes a HC_AddressPair from the neighborhood table <p>
	 * Algorithm: <br>
	 * Find the Index and check if physical address is the same <br>
	 *     (a) If yes, remove and make that point unknown <br>
	 *     (b) If no, make it false <br>
	 * 
	 * @param	addrpair	a Node Address
	 * @return	true if it was able to remove the data, false otherwise
	 */ 
	public boolean remove(HC_AddressPair addrpair) {
		if (HRoot != null && addrpair.getPhysicalAddress().equals(HRoot.getPhysicalAddress())) 
            HRoot = null;

		for (int index = 0; index < HC_Node.MaxHypercubeDim; index++)
        {
			if (NView[index] != null && addrpair.getPhysicalAddress().equals(NView[index].getPhysicalAddress())) 
            {
                NView[index] = null;
                return true;
            }
        }
        
		return false;
	}




	/** 
	 * Searches a Tear in all neighborhood <p>
	 * Algorithm: <br>  
	 *  The tear found is one where the neighborhood table entry has been missing <br>
	 *  Check each neighbor in turn, in increasing logical address order. <br>
	 *  Stop when the neighbor's LA exceeds that of the known HRoot. <br>
     *  If a neighbor is found to be absent, report it as a tear.
	 * 
	 * @return	The index of a tear if there is one, and HC_Global.MaxAddress otherwise.
	 */
	public int searchTear() {
		int a = HC_Node.MaxAddress();
		
		int i;
		if (HRoot == null)
			return a;

		for (i = 0; i < HC_Node.MaxHypercubeDim; i++) {
			if ((NView[i] == null) || (NView[i].getPhysicalAddress() == null))
            {
				if (parent.MyLogicalAddress.getNthNeighbor(i).getColexIndex() <
				((HC_LogicalAddress)HRoot.getLogicalAddress()).getColexIndex()) {
					return i;
				}
            }
		}

		return a;
	}

	
	
	/** 
	 * Searches Little Tear in all the neighborhood <p>
	 * Algorithm:  <br>
	 *  The tear found is one where the neighborhood table entry has been missing
	 * 	for Ttimeout time. <br>
	 *  Check each neighbor in turn, in increasing logical address order. <br>
	 *  Stop when the neighbor's LA exceeds that of the known HRoot. <br>
     *  If a neighbor is found to be absent, report it as a tear. <br>
	 * 
	 * @return	The index of a tear if there is one, and HC_Global.MaxAddress otherwise.
	 */
	public int searchLittleTear() {
		int a = HC_Node.MaxAddress();

		int i;
		if (HRoot == null)
			return a;

		for (i = 0; i < HC_Node.MaxHypercubeDim; i++) {
			if ((NView[i] == null) || (NView[i].getPhysicalAddress() == null)) {
				if (parent.MyLogicalAddress.getNthNeighbor(i).getColexIndex() <
                    ((HC_LogicalAddress)HRoot.getLogicalAddress()).getColexIndex())
					return i;
			}
			else 
            {
				if (NView[i] != null) 
                {
					if (NView[i].getAge() < parent.MaxMissingNeighbor) 
                    {
						if (parent.MyLogicalAddress.getNthNeighbor(i).getColexIndex() <
						((HC_LogicalAddress)HRoot.getLogicalAddress()).getColexIndex())
							return i;
                    }
                }
            }
		}

		return a;
	}
      



	/** 
	 * Gets the AddressPair of the Nth neighbor
	 * @param	n integer representing the Index of neighbor
	 * @return  a copy of the HC_AddressPair of the requested neighbor
	 */	
	public HC_AddressPair getNthNeighbor(int n) {
        return NView[n];
	}

	

	
	/** 
	 * Gets the AddressPair of a neighbor with a logical address
	 * @param	laddr a Neighbor's HC_LogicalAddress
	 * @return  returns a copy of the HC_AddressPair of the requested neighbor
	 */	
	public HC_AddressPair getNeighborPair(HC_LogicalAddress laddr) {
				
		if (NView == null)
			return null; 
		if (laddr == null)
			return null;
		
		boolean findpair = false;
		int index = 0;
		
		for (int i = 0; i < HC_Node.MaxHypercubeDim; i++) {			
			HC_LogicalAddress la = (HC_LogicalAddress)(NView[i].getLogicalAddress());
			if (la != null) {
				if ((la.equals(laddr)) && (NView[i].getPhysicalAddress() != null)) {
					findpair = true;	
					index =i;
					break;				
				}
			}
		}
			
		if (findpair == true)
			return NView[index];
		else
			return null; 		
	}


	/** 
	 * Check if there exists a neighbor with a specified logical address
	 * @param	laa  a I_LogicalAddress
	 * @return  returns true if finding a neighbor with logical address
	 * la; otherwise flase.
	 */	
	public boolean contains(I_LogicalAddress laa) {
				
		HC_LogicalAddress laddr = (HC_LogicalAddress) laa;

		if (NView == null)
			return false; 
		if (laddr == null)
			return false;
		
		boolean findpair = false;
		//int index = 0;
		
		for (int i = 0; i < HC_Node.MaxHypercubeDim; i++) {			
			HC_LogicalAddress la = (HC_LogicalAddress)(NView[i].getLogicalAddress());
			if (la != null) {
				if ((la.equals(laddr)) && (NView[i].getPhysicalAddress() != null)) {
					findpair = true;	
					//index =i;
					break;				
				}
			}
		}
			
		if (findpair == true)
			return true;
		else
			return false; 		
	}


	
	/** 
	 * Gets all neighbor's AddressPair
	 * @return  returns an array of HC_AddressPair of all neighbors
	 */
	public HC_AddressPair[] getNeighbors() {
        return NView;
	}
	

	/** 
	 * Gets HRoot's AddressPair
	 * @return	the HC_AddressPair of the HRoot
	 */
	public HC_AddressPair getHRoot() {
		return HRoot;
	}


	/** 
	 * Gets HRoot's sequence number
	 * @return	an integer representing the sequence number
	 */
	public int getHRootSeqNum() {
		return HRootSeqNum;
	}


	
	/** 
	 * Set HRoot sequence number to a new value
	 * @param	newseqnum an integer representing the new sequence number
	 */
	public void setHRootSeqNum(int newseqnum) {
		HRootSeqNum = newseqnum;
	}

	

	/** 
	 * Resets the neighborhood table
	 */
	public void resetNeighborhood() {
		for (int i = 0; i < HC_Node.MaxHypercubeDim; i++) {
			NView[i] = new HC_AddressPair(parent, new HC_LogicalAddress(HC_Node.MaxAddress(), true), null, true);
			//	NView[i] = null;
		}

		HRoot = null;
		HRootSeqNum = 0;

		parent.MyLogicalAddress = new HC_LogicalAddress(HC_Node.MaxAddress(), true);
	}

	
	/** 
	 * Converts the HC_Neighborhood object to String format
	 * @return  string representation of the object
	 */
    public String toString() {
		StringBuffer sb = new StringBuffer();
	
		sb.append(MNTimeOut);
		sb.append(" ");
		sb.append(HRootSeqNum);
		sb.append(" ");
		if (HRoot != null) {
			sb.append(HRoot.toString());
		}

		for (int i = 0; i < HC_Node.MaxHypercubeDim; i++) {
			if (NView[i] != null) {
				sb.append(NView[i].toString());
			}
		}

		return sb.toString();
	}

	

	
}



