/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.util.XmlUtil;

import java.io.IOException;
import java.io.InputStream;



 
/** This is the class used to receive data from other nodes 
 *
 * @author HyperCast Team
 */
public class HCastInputStream extends InputStream

{
	/** debug flag */
	private boolean debug = false;
	
   /** The number of bytes in the buffer */
	private int NumberOfBytes = 0;

    /** The buffer of messages. */
	private byte[] buffer;

    /** The location of the oldest message in the buffer.  (The buffer is circular, so the start moves.) */
	private int Start = 0;

    /** Timeout for reads. */
    private int soTimeout = 0;
	
    /** Number of threads waiting to read. */
    private int readersWaiting = 0;

    /** Number of threads waiting to write. */
    private int writersWaiting = 0;
    
    /** the stream id. */
    private byte[] streamID;
    
    /** the stream manager */
    private StreamManager parent;
    
    /**the flag indicating that the stream has been assigned to the application */
    private boolean assigned = false; 
    
	/** 
	 * Constructs the BoundedBuffer with a size.
	 * Simply initializes the parameters
	 */

	public HCastInputStream (StreamManager sm,byte[] streamID)
	{		
		this.streamID = streamID;
		parent = sm;
		int size = parent.config.getPositiveIntAttribute (XmlUtil.createXPath("/Public/InputStreamBuffersize"));
		buffer = new byte[size];

	}


    /** Get the timeout time for a read operation. */
    public int getSoTimeout() 
    {
        return soTimeout;
    }
    
    /** Set the timeout time for a read operation. */
    public void setSoTimeout(int i) 
    {
        if (i < 0)
            throw new IllegalArgumentException("called setSoTimeout with a negative argument!");

        soTimeout = i;
    }
    


	/** 
	 * Reads a byte from the BoundedBuffer. <p>
	 * Algorithm: <br>
	 *  1. Waits till the buffer is non-empty <br>
	 *  2. Copies first byte in the buffer <br>
	 *  3. Increments the Start pointer modulo buffer length <br>
	 *  4. Decreases size of the buffer <br>
	 *  5. Notifies any other process waiting <br>
	 *  6. Returns the copied element <br>
	 * 
	 * @return	the first byte in the buffer
	 */

	synchronized public int read() throws IOException
    {
		
		byte b;
		// 1. Waits till buffer is non-empty
        if (soTimeout == 0) 
        {
            try{
            	try 
				{
            		readersWaiting++;
            		while(NumberOfBytes == 0)
            		{
            			wait();
            		}
				}
            	finally 
				{
            		readersWaiting--;
				}
            }catch (InterruptedException ie) {}
        }
        else 
        {
            // Inaccurate, but we can live with that.
            if (NumberOfBytes == 0)
            {
              try
			  {	
                try 
                {
                    readersWaiting++;
                    wait(soTimeout);
                } 
                finally    
                {
                    readersWaiting--;
                }
			  }catch (InterruptedException ie) {}
                if (NumberOfBytes == 0)
                {
                    return -1;
                }
            }   
        }

		// 2. Copies first element in the buffer
		b = buffer[Start];

		// 3. Increments the Start pointer modulo buffer length
		Start = (Start + 1) % buffer.length;
		
		// 4. Decreases size of the buffer
		NumberOfBytes--;

		// 5. Notifies any writing elements waiting on this (only writing threads could be waiting.)
		notify();

		// 6. Returns 
		return b;
	}
	/** 
	 * Reads a byte from the BoundedBuffer. <p>
	 * Algorithm: <br>
	 *  1. Waits till the buffer is non-empty <br>
	 *  2. Copies first byte in the buffer <br>
	 *  3. Increments the Start pointer modulo buffer length <br>
	 *  4. Decreases size of the buffer <br>
	 *  5. Notifies any other process waiting <br>
	 *  6. Returns the copied element <br>
	 * 
	 * @return	the bytes in the buffer
	 */
	synchronized public int read( byte[] b,int off,int len)throws IOException 
	{
		if (debug)
			System.out.println("HCastOutputStream: read: reading data.");

		int value = -1;
		// 1. Waits till buffer contains bytes of length len
		if (soTimeout == 0) 
		{
			try{
				try 
				{
					readersWaiting++;

					if (debug)
						System.out.println("StreamManager: read: NumberOfBytes is "+NumberOfBytes);

					while(NumberOfBytes < len)
					{
						wait();

						if (debug)
							System.out.println("StreamManager: read: be notified, NumberOfBytes is "+NumberOfBytes);

					}
				}
				finally 
				{
					readersWaiting--;
				}
			}catch (InterruptedException ie) {}
		}
		else 
		{
			// Inaccurate, but we can live with that.
			if (NumberOfBytes == 0)
			{
			  try
			  {	
				try 
				{
					readersWaiting++;
					wait(soTimeout);
				} 
				finally    
				{
					readersWaiting--;
				}
			  }catch (InterruptedException ie) {}
				if (NumberOfBytes == 0)
				{
					return -1;
				}
			}   
		}

		// 2. Copies 
		if ((Start + len)> buffer.length) 
		{
			int remainbufLen = buffer.length - Start;
			System.arraycopy(b,off,buffer,Start,remainbufLen);
			System.arraycopy(b, off+remainbufLen, buffer, 0, len-remainbufLen);
			
		}
		else
		{
			System.arraycopy(buffer,Start,b,off,len);
		}



		// 3. Increments the Start pointer modulo buffer length
		Start = (Start + len) % buffer.length;
		
		// 4. Decreases size of the buffer
		NumberOfBytes-= len;
		
		value = len;
		// 5. Notifies any writing elements waiting on this (only writing threads could be waiting.)
		notify();

		if (debug)
			System.out.println("StreamManager: read: return value is "+value);

		// 6. Returns 
		return value;		
	}

	/** 
	 * Reads a byte from the BoundedBuffer. <p>
	 * 
	 * @return	the first bytes in the buffer
	 */
	public int read( byte[] b)throws IOException 
	{
		return read(b,0,b.length);
	}

	/** 
	 * Writes a message to the DropHeadMessageBuffer <P>
	 * Algorithm: <br>
	 *  1. checks if buffer is full <br>
	 *  2. Places New bytes at the end of the buffer <br>
	 *  3. Increase number of bytes <br>
	 *  4. Notifies all threads waiting <br>
	 * 
	 * @param	payload		byte array to be added into the buffer
	 */

	synchronized public void nonblockingWrite(byte[] payload, int off,int len)
	{
		/*
		 * If the received packet is larger than buffer size, drop it.
		 */
		if (len > buffer.length)
		{
			return;
		}

		if (debug)
			System.out.println("HCastInputStream: nonblockingWrite: copy data to buffer.");
		
		int index = (Start + NumberOfBytes) % buffer.length;
		
		if ((index + len)> buffer.length) 
		{
			int remainbufLen = buffer.length - index;
			System.arraycopy(payload,off,buffer,index,remainbufLen);
			System.arraycopy(payload, off+remainbufLen, buffer, 0, len-remainbufLen);
		}
		else
		{
			System.arraycopy(payload,off,buffer,index,len);
		}

		//Adjust available bytes in the buffer
		if ((NumberOfBytes + len)> buffer.length)
		{
			NumberOfBytes = buffer.length;
		}
		else
		{
			NumberOfBytes += len;
		}

		if (debug)
			System.out.println("HCastInputStream: nonblockingWrite: message length is "+len+", NumberOfBytes is "+NumberOfBytes);


        notify();
	}





	/** 
	 * Writes a message to the BoundedBuffer <P>
	 * Algorithm: <br>
	 *  1. checks if buffer is full <br>
	 *  2. Places New packet at the end of the buffer <br>
	 *  3. Increase number of elements <br>
	 *  4. Notifies all threads waiting <br>
	 * 
	 * @param	payload		byte array to be added into the buffer
	 */

	synchronized public void blockingWrite(byte[] payload) throws InterruptedException 
	{
		/*
		 * If the received packet is larger than buffer size, drop it.
		 */
		if (payload.length > buffer.length)
		{
			return;
		}

		// if the buffer is full, wait
		while (NumberOfBytes == buffer.length) 
        {
            try 
            {
                try 
                {
                    writersWaiting++;
                    wait();
                }
                finally 
                {
                    writersWaiting--;
                }
            }
            catch (InterruptedException ie) 
            {
                notify();
                throw ie;
            }
        }
        

		// Place new packet into the buffer
		int index = (Start + NumberOfBytes) % buffer.length;

		int len = payload.length;
		if ((index + len)> buffer.length) 
		{
			int remainbufLen = buffer.length - index;
			System.arraycopy(payload,0,buffer,index,remainbufLen);
			System.arraycopy(payload, 0+remainbufLen, buffer, 0, len-remainbufLen);
		}
		else
		{
			System.arraycopy(payload,0,buffer,index,len);
		}
		
		//Adjust available bytes in the buffer
		if ((NumberOfBytes + len)> buffer.length)
		{
			NumberOfBytes = buffer.length;
		}
		else
		{
			NumberOfBytes += len;
		}


		// Notifies a waiting read thread (Only a read thread could be wait()ing.)
        notify();
	}

	/**
	 * Closes this input stream and releases any system resources associated with the stream.
	 * 
	 */
	public void close()
	{
		parent.inputtable.remove(new String(streamID));
	}
	
	/** 
	 *  * @return	the number of bytes available that can be read without blocking
	 */

	synchronized public int available()throws IOException
	{
		return NumberOfBytes;
	}
 
	 /** 
	 *  * @  Tests if this input stream supports the mark and reset methods.
	 */
 	public boolean markSupported()
 	{
 		return false;
 	} 
 
 	/**
 	 * Skips over and discards n bytes of data from this input stream. If the
 	 * bytes of available data in the buffer is less than specified length,
 	 * clean up the buffer and return -1.
 	 */
	public long skip(long n)throws IOException
	{
		if(n > NumberOfBytes)
		{
			byte[] discarded = new byte[NumberOfBytes];
			read(discarded);
			return -1;
		}
		else
		{
			byte[] discarded = new byte[(new Long(n)).intValue()];
			read(discarded);
			return n;
		}
	}


	 /**
	  * @return stream id
	  * @deprecated use getStreamIdentifier
	  */
	public int getStreamID()
	{
		return ByteArrayUtility.toInteger(streamID,0);
		
	}

	 /**
	  * @return stream id*/
	public int getStreamIdentifier()
	{
		return ByteArrayUtility.toInteger(streamID,0);
			
	}

	/**
	 * @return true if the stream is assigned to the application*/
   public boolean isAssignedToApplication()
   {
	   return assigned;
		
   }
   
   /**
	* set the flag indicating if the stream has been assigned to the application*/
  public void setAssignedFlag(boolean a)
  {
	  assigned = a;
		
  }
 }
 
 
 
