/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.I_LogicalAddress;
import hypercast.I_OverlayMessage;
import hypercast.I_OverlaySocket;

import hypercast.I_ReceiveCallback;
import hypercast.NotificationHandler;

import hypercast.StreamManager;

import hypercast.StatsProcessor;

import hypercast.HyperCastConfig;

import hypercast.HyperCastConfigException;
import hypercast.HyperCastFatalRuntimeException;
import hypercast.HyperCastWarningRuntimeException;

import hypercast.events.E2EACK_RECEIVED;
import hypercast.events.E2EPARTIALACK_RECEIVED;
import hypercast.events.INTERNAL_EVENT;
//import hypercast.events.MSG_WITH_E2EACK_SENT;
import hypercast.events.NAMING_EVENT;
import hypercast.events.NEWSTREAM_ARRIVED_EVENT;
import hypercast.events.NODE_ISSTABLE;
import hypercast.events.NODE_LEAVEOVERLAY;
import hypercast.events.NODE_LOGICALADDRESSCHANGED;
import hypercast.events.NODE_NEIGHBORHOODCHANGED;
import hypercast.events.NOTIFICATION_EVENT;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.XPath;

import java.security.cert.Certificate;
import java.security.PrivateKey;

import java.io.PrintWriter;
import java.net.SocketException;

/**
 * The HyperCast Application Object is intended to be a simplified
 * interface to the HyperCast system. Instead of using the Berkeley
 * Sockets paradigm, an object-oriented approach is used. User code
 * simply extends this object to create a HyperCast application.
 *
 * @author HyperCast Team
 * @author Greg Mattes
 */
public abstract class HyperCastAppl
    implements I_OverlaySocket, I_ReceiveCallback
{
    /*
     * Constants used to configure action taken when exceptions occur
     *
     * Note: these constants could become a bit vector if other options are added.
     * Also, consider for Java 1.5 enumeration.
     */

    /** catch exception and output to log */
    public static final int EXCEPTION_LOG           = 0;

    /** re-throw exception to user */
    public static final int EXCEPTION_THROW         = 1;

    /** catch exception and output to log then re-throw exception to user */
    public static final int EXCEPTION_LOG_AND_THROW = 2;

    /** do nothing when exception is caught */
    public static final int EXCEPTION_IGNORE        = 3;

    private static final int WARNING_ACTION_DEFAULT = EXCEPTION_LOG;
    private static final int FATAL_ACTION_DEFAULT = EXCEPTION_LOG_AND_THROW;

    private int warningAction = WARNING_ACTION_DEFAULT;
    private int fatalAction = FATAL_ACTION_DEFAULT;

    private I_OverlaySocket socket = null;
    
    private HyperCastApplNotificationHandler notificationHandler;

    /*
     * This references the log stream embedded within the configuration object.
    */
    private PrintWriter log;

    private StatsProcessor statsProcessor = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);

    private final class HyperCastApplNotificationHandler
        extends NotificationHandler 
    {
        HyperCastApplNotificationHandler() { super(); }

        /*
         * The handler method is implemented as a method of the
         * HyperCastAppl object so that applications may extend it.
         */
        public void handler (final NOTIFICATION_EVENT event)
        {
            HyperCastAppl.this.handler (event);
        }

        /*
         * The application programmer is given access to the
         * <code>HyperCastAppl.handler()</code> method so that it can
         * be overridden with application specific code. Should the
         * programmer choose not to override that method it should
         * call the default handler. This method allows that method to
         * call the default notification handler.
         *
         * @see HyperCastAppl.handler()
         */
        public void invokeDefaultHandler (final NOTIFICATION_EVENT event)
        {
            super.handler (event);
        }

        public void handle_NODE_NEIGHBORHOODCHANGED   (NODE_NEIGHBORHOODCHANGED   event) { HyperCastAppl.this.handle_NODE_NEIGHBORHOODCHANGED   (event); }
        public void handle_NODE_LOGICALADDRESSCHANGED (NODE_LOGICALADDRESSCHANGED event) { HyperCastAppl.this.handle_NODE_LOGICALADDRESSCHANGED (event); }
        public void handle_NODE_LEAVEOVERLAY          (NODE_LEAVEOVERLAY          event) { HyperCastAppl.this.handle_NODE_LEAVEOVERLAY          (event); }
        public void handle_NEWSTREAM_ARRIVED_EVENT    (NEWSTREAM_ARRIVED_EVENT    event) { HyperCastAppl.this.handle_NEWSTREAM_ARRIVED_EVENT    (event); }
        //public void handle_MSG_WITH_E2EACK_SENT       (MSG_WITH_E2EACK_SENT       event) { HyperCastAppl.this.handle_MSG_WITH_E2EACK_SENT       (event); }
        public void handle_E2EPARTIALACK_RECEIVED     (E2EPARTIALACK_RECEIVED     event) { HyperCastAppl.this.handle_E2EPARTIALACK_RECEIVED     (event); }
        public void handle_E2EACK_RECEIVED            (E2EACK_RECEIVED            event) { HyperCastAppl.this.handle_E2EACK_RECEIVED            (event); }
        public void handle_NODE_ISSTABLE              (NODE_ISSTABLE              event) { HyperCastAppl.this.handle_NODE_ISSTABLE              (event); }
    }

    /**
     * Creates a HyperCast application object with defaults for
     * configuration filename and exception actions.
     * Relies on unified constructor.
     */
    public HyperCastAppl() {
        this (WARNING_ACTION_DEFAULT, FATAL_ACTION_DEFAULT);
    }

    /**
     * Creates a HyperCast application object with the specified
     * configuration filename and defaults for exception actions.
     * Calls unified constructor.
     */
    public HyperCastAppl (final String configurationFilename) {
        this (WARNING_ACTION_DEFAULT, FATAL_ACTION_DEFAULT, configurationFilename);
    }

    /**
     * Creates a HyperCast application object with the specified
     * values for exception actions and default configuration filename.
     * Calls unified constructor.
     */
    public HyperCastAppl (final int warningAction, final int fatalAction) {
        this (warningAction, fatalAction, "hypercast.xml");
    }

    /**
     * Creates a HyperCast application object with the specified
     * values for exception actions and the specified value for the
     * configuration filename.
     */
    public HyperCastAppl (final int warningAction, final int fatalAction, final String configurationFilename)
    {
        try
        {
            setWarningAction (warningAction);
            setFatalAction (fatalAction);

            HyperCastConfig config = HyperCastConfig.createConfig (configurationFilename);

            this.log = config.log;

            notificationHandler = new HyperCastApplNotificationHandler();

            socket = config.createOverlaySocket (this, notificationHandler);

            statsProcessor.addStatsElement ("Socket"/* name */, socket/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

            /*
             *  This call is *not* synchronous! It simply initiates the
             *  join operation. We rely on the notification mechanism to
             *  tells us when join is complete.
             */
            socket.joinOverlay();
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /*
     * Methods that block a thread until event occurs; the event is
     * returned. If multiple threads are blocked on the same type of
     * event, only one will be unblocked per event, which is unblocked
     * is not guarenteed to be deterministic.
     *
     * If a handler is defined for an event type and a thread is
     * blocked on the same event type the handler will be invoked
     * *and* the thread will be unblocked for the same event. (This is
     * because there is no way for the system to know if a thread is
     * blocked on an event so it must always perform the monitor
     * signal operation, similarly there is no way for the system to
     * know if a handler has been defined by the application so it
     * must always invoke the handler).
     *
     * A thread that is interested in consuming an event that occured
     * before the thread was blocked waiting on events of that type
     * can indicate this by passing the value of "true" to the
     * appropriate waitUntil_ method. For example if the stability
     * event only occurs once and the call to block a thread to wait
     * for it happens after the event, the thread will remained
     * blocked indefinitely even though the event for which it was
     * interested has already occurred.
     */

    public NODE_ISSTABLE              waitUntil_NODE_ISSTABLE()              { return notificationHandler.waitUntil_NODE_ISSTABLE();              }
    public E2EACK_RECEIVED            waitUntil_E2EACK_RECEIVED()            { return notificationHandler.waitUntil_E2EACK_RECEIVED();            }
    public E2EPARTIALACK_RECEIVED     waitUntil_E2EPARTIALACK_RECEIVED()     { return notificationHandler.waitUntil_E2EPARTIALACK_RECEIVED();     }
    //public MSG_WITH_E2EACK_SENT       waitUntil_MSG_WITH_E2EACK_SENT()       { return notificationHandler.waitUntil_MSG_WITH_E2EACK_SENT();       }
    public NEWSTREAM_ARRIVED_EVENT    waitUntil_NEWSTREAM_ARRIVED_EVENT()    { return notificationHandler.waitUntil_NEWSTREAM_ARRIVED_EVENT();    }
    public NODE_LEAVEOVERLAY          waitUntil_NODE_LEAVEOVERLAY()          { return notificationHandler.waitUntil_NODE_LEAVEOVERLAY();          }
    public NODE_LOGICALADDRESSCHANGED waitUntil_NODE_LOGICALADDRESSCHANGED() { return notificationHandler.waitUntil_NODE_LOGICALADDRESSCHANGED(); }
    public NODE_NEIGHBORHOODCHANGED   waitUntil_NODE_NEIGHBORHOODCHANGED()   { return notificationHandler.waitUntil_NODE_NEIGHBORHOODCHANGED();   }
    public NAMING_EVENT               waitUntil_NAMING_EVENT()               { return notificationHandler.waitUntil_NAMING_EVENT();               }

    public NODE_ISSTABLE              waitUntil_NODE_ISSTABLE              (final long timeout) { return notificationHandler.waitUntil_NODE_ISSTABLE              (timeout); }
    public E2EACK_RECEIVED            waitUntil_E2EACK_RECEIVED            (final long timeout) { return notificationHandler.waitUntil_E2EACK_RECEIVED            (timeout); }
    public E2EPARTIALACK_RECEIVED     waitUntil_E2EPARTIALACK_RECEIVED     (final long timeout) { return notificationHandler.waitUntil_E2EPARTIALACK_RECEIVED     (timeout); }
    //public MSG_WITH_E2EACK_SENT       waitUntil_MSG_WITH_E2EACK_SENT       (final long timeout) { return notificationHandler.waitUntil_MSG_WITH_E2EACK_SENT       (timeout); }
    public NEWSTREAM_ARRIVED_EVENT    waitUntil_NEWSTREAM_ARRIVED_EVENT    (final long timeout) { return notificationHandler.waitUntil_NEWSTREAM_ARRIVED_EVENT    (timeout); }
    public NODE_LEAVEOVERLAY          waitUntil_NODE_LEAVEOVERLAY          (final long timeout) { return notificationHandler.waitUntil_NODE_LEAVEOVERLAY          (timeout); }
    public NODE_LOGICALADDRESSCHANGED waitUntil_NODE_LOGICALADDRESSCHANGED (final long timeout) { return notificationHandler.waitUntil_NODE_LOGICALADDRESSCHANGED (timeout); }
    public NODE_NEIGHBORHOODCHANGED   waitUntil_NODE_NEIGHBORHOODCHANGED   (final long timeout) { return notificationHandler.waitUntil_NODE_NEIGHBORHOODCHANGED   (timeout); }
    public NAMING_EVENT               waitUntil_NAMING_EVENT               (final long timeout) { return notificationHandler.waitUntil_NAMING_EVENT               (timeout); }

    /*
     * Event handlers that are designed to be extended by user, called from HyperCastApplNotificationHandler.
     */
    public void handle_NODE_NEIGHBORHOODCHANGED   (final NODE_NEIGHBORHOODCHANGED   event) { }
    public void handle_NODE_LOGICALADDRESSCHANGED (final NODE_LOGICALADDRESSCHANGED event) { }
    public void handle_NODE_LEAVEOVERLAY          (final NODE_LEAVEOVERLAY          event) { }
    public void handle_NEWSTREAM_ARRIVED_EVENT    (final NEWSTREAM_ARRIVED_EVENT    event) { }
    //public void handle_MSG_WITH_E2EACK_SENT       (final MSG_WITH_E2EACK_SENT       event) { }
    public void handle_E2EPARTIALACK_RECEIVED     (final E2EPARTIALACK_RECEIVED     event) { }
    public void handle_E2EACK_RECEIVED            (final E2EACK_RECEIVED            event) { }
    public void handle_NODE_ISSTABLE              (final NODE_ISSTABLE              event) { }
    public void handle_NAMING_EVENT               (final NAMING_EVENT               event) { }

    /*
     * The handler method of application is not final by design.  An
     * application may extend it to do whatever processing it
     * pleases. If "regular" processing is needed, it simple calls
     * this code by calling "super.handler()".
     *
     * Note that this method is designed to be overridden, but it
     * should not normally be called. It is designed to be called by
     * the notification handler thread. No synchronization is done
     * since this method only delegates to user defined methods,
     * i.e. no state is changed.
     */
    protected void handler (final NOTIFICATION_EVENT event)
    {
        notificationHandler.invokeDefaultHandler (event);
    }

    /**
     * Application overrides this method to receive messages.
     */
    public void ReceiveCallback (final I_OverlayMessage message) { }

    /*
     * joinOverlay() does not have to be called explicitly. It is
     * included here for completeness because leaveOverlay() is also
     * included. It gives an application the option of leaving and
     * returning to an overlay.
     *
     * @see OL_Socket#joinOverlay()
     * @see OL_Socket#leaveOverlay()
     */
    public final void joinOverlay() { socket.joinOverlay(); }
    public final void leaveOverlay() { socket.leaveOverlay(); }

    /**
     * @see OL_Socket#createMessage(byte[] payload)
     */
    public final I_OverlayMessage createMessage (final byte[] payload)
    {
        I_OverlayMessage message = null;
        try
        {
            message = socket.createMessage (payload);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return message;
    }

    /**
     * @see OL_Socket#createMessage(byte[] payload, short servicenumber)
     */
    public final I_OverlayMessage createMessage (final byte[] payload, final short servicenumber)
    {
        I_OverlayMessage message = null;
        try
        {
            message = socket.createMessage (payload, servicenumber);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return message;
    }

    /**
     * @see OL_Socket#sendToNode(I_OverlayMessage message, I_LogicalAddress destination)
     */
    public final void sendToNode (final I_OverlayMessage message, final I_LogicalAddress destination)
    {
        try
        {
            socket.sendToNode (message, destination);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#sendToParent(I_OverlayMessage message, I_LogicalAddress root)
     */
    public final void sendToParent (final I_OverlayMessage message, final I_LogicalAddress root)
    {
        try
        {
            socket.sendToParent (message, root);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#sendToChildren(I_OverlayMessage message, I_LogicalAddress root)
     */
    public final void sendToChildren (final I_OverlayMessage message, final I_LogicalAddress root)
    {
        try
        {
            socket.sendToChildren (message, root);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#sendToAllNeighbors(I_OverlayMessage message)
     */
    public final void sendToAllNeighbors (final I_OverlayMessage message)
    {
        try
        {
            socket.sendToAllNeighbors (message);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#sendToAll(I_OverlayMessage message)
     */
    public final void sendToAll (final I_OverlayMessage message)
    {
        try
        {
            socket.sendToAll (message);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#sendFlood(I_OverlayMessage message)
     */
    public final void sendFlood (final I_OverlayMessage message)
    {
        try
        {
            socket.sendFlood (message);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#createLogicalAddress(byte[] laddr, int offset)
     */
    public final I_LogicalAddress createLogicalAddress (final byte[] laddr, final int offset)
    {
        try
        {
            return socket.createLogicalAddress (laddr, offset);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#createLogicalAddress(String laddr)
     */
    public final I_LogicalAddress createLogicalAddress (final String laddr)
    {
        try
        {
            return socket.createLogicalAddress (laddr);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#getLogicalAddress()
     */
    public final I_LogicalAddress getLogicalAddress()
    {
        try
        {
            return socket.getLogicalAddress();
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#setLogicalAddress(I_LogicalAddress address)
     */
    public final void setLogicalAddress (final I_LogicalAddress address)
    {
        try
        {
            socket.setLogicalAddress (address);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#getUniqueIdentifier()
     */
    public final byte [] getUniqueIdentifier()
    {
        try
        {
            return socket.getUniqueIdentifier();
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#setName(String)
     */
    public void setName (final String name)
    {
        try
        {
            socket.setName (name);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#setName(Certificate,PrivateKey)
     */
    public void setName (final Certificate certificate, final PrivateKey privateKey)
    {
        try
        {
            socket.setName (certificate, privateKey);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#setName(String,I_LogicalAddress)
     */
    public void setName (final String name, final I_LogicalAddress logicalAddress)
    {
        try
        {
            socket.setName (name, logicalAddress);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#setName(Certificate,PrivateKey,String,I_LogicalAddress)
     */
    public void setName (final Certificate certificate, final PrivateKey privateKey, final String name, final I_LogicalAddress logicalAddress)
    {
        try
        {
            socket.setName (certificate, privateKey, name, logicalAddress);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#unsetName(String)
     */
    public void unsetName (final String name)
    {
        try
        {
            socket.unsetName (name);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#unsetName(Certificate, PrivateKey)
     */
    public void unsetName (final Certificate certificate, final PrivateKey privateKey)
    {
        try
        {
            socket.unsetName (certificate, privateKey);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#getNames()
     */
    public String [] getNames()
    {
        try
        {
            return socket.getNames();
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#getNames(I_LogicalAddress)
     */
    public String [] getNames (final I_LogicalAddress peer)
    {
        try
        {
            return socket.getNames (peer);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#getNames(I_LogicalAddress, boolean, long, boolean, boolean, int, int, short, short)
     */
    public String [] getNames (final I_LogicalAddress peer,
                               final boolean          requireTrusted,
                               final long             timeout,
                               final boolean          requireSerialNumberMatch,
                               final boolean          requireAuthoritative,
                               final int              maxAge,
                               final int              maxResponses,
                               final short            minHopCount,
                               final short            maxHopCount)
    {
        try
        {
            return socket.getNames (peer, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#getNamesNonBlocking(I_LogicalAddress)
     */
    public int getNamesNonBlocking (final I_LogicalAddress peer)
    {
        try
        {
            return socket.getNamesNonBlocking (peer);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return 0;
    }

    /**
     * @see OL_Socket#getNamesNonBlocking(I_LogicalAddress, boolean, long, boolean, boolean, int, int, short, short)
     */
    public int getNamesNonBlocking (final I_LogicalAddress peer,
                                    final boolean          requireTrusted,
                                    final long             timeout,
                                    final boolean          requireSerialNumberMatch,
                                    final boolean          requireAuthoritative,
                                    final int              maxAge,
                                    final int              maxResponses,
                                    final short            minHopCount,
                                    final short            maxHopCount)
    {
        try
        {
            return socket.getNamesNonBlocking (peer, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return 0;
    }

    /**
     * @see OL_Socket#getLogicalAddressByName(String)
     */
    public I_LogicalAddress [] getLogicalAddressByName (final String name)
    {
        try
        {
            return socket.getLogicalAddressByName (name);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#getLogicalAddressByName(String, boolean, long, boolean, boolean, int, int, short, short)
     */
    public I_LogicalAddress [] getLogicalAddressByName (final String  name,
                                                        final boolean requireTrusted,
                                                        final long    timeout,
                                                        final boolean requireSerialNumberMatch,
                                                        final boolean requireAuthoritative,
                                                        final int     maxAge,
                                                        final int     maxResponses,
                                                        final short   minHopCount,
                                                        final short   maxHopCount)
    {
        try
        {
            return socket.getLogicalAddressByName (name, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#getLogicalAddressByNameNonBlocking(String)
     */
    public int getLogicalAddressByNameNonBlocking (final String name)
    {
        final int serialNumber;
        try
        {
            return socket.getLogicalAddressByNameNonBlocking (name);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return 0;
    }

    /**
     * @see OL_Socket#getLogicalAddressByNameNonBlocking(String, boolean, long, boolean, boolean, int, int, short, short)
     */
    public int getLogicalAddressByNameNonBlocking (final String  name,
                                                   final boolean requireTrusted,
                                                   final long    timeout,
                                                   final boolean requireSerialNumberMatch,
                                                   final boolean requireAuthoritative,
                                                   final int     maxAge,
                                                   final int     maxResponses,
                                                   final short   minHopCount,
                                                   final short   maxHopCount)
    {
        try
        {
            return socket.getLogicalAddressByNameNonBlocking (name, requireTrusted, timeout, requireSerialNumberMatch, requireAuthoritative, maxAge, maxResponses, minHopCount, maxHopCount);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return 0;
    }

    /**
     * @see OL_Socket#installTrustedNamingCertificate(Certificate)
     */
    public void installTrustedNamingCertificate (final Certificate certificate)
    {
        try
        {
            socket.installTrustedNamingCertificate (certificate);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#getParent(I_LogicalAddress root)
     */
    public final I_LogicalAddress getParent (final I_LogicalAddress root)
    {
        try
        {
            return socket.getParent (root);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#getChildren(I_LogicalAddress root)
     */
    public final I_LogicalAddress[] getChildren (final I_LogicalAddress root)
    {
        try
        {
            return socket.getChildren (root);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /**
     * @see OL_Socket#getNeighbors()
     */
    public final I_LogicalAddress [] getNeighbors()
    {
        try
        {
            return socket.getNeighbors();
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return null;
    }

    /*
     * Unsigned 16-bit value that specifies the maximum number of hops
     * that a message is permitted to make in the network.
     *
     * @see OL_Socket#setHopLimit(short hopLimit)
     */
    public final void setHopLimit (final short hopLimit)
    {
        try
        {
            socket.setHopLimit (hopLimit);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#getHopLimit()
     */
    public final short getHopLimit()
    {
        try
        {
            return socket.getHopLimit();
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return -1;
    }

    /**
     * @see OL_Socket#setSoTimeout(int timeout)
     */
    public final void setSoTimeout (final int timeout)
    {
        try
        {
            socket.setSoTimeout (timeout);
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
    }

    /**
     * @see OL_Socket#getSoTimeout()
     */
    public final int getSoTimeout()
    {
        int timeout = -1;
        try
        {
            timeout = socket.getSoTimeout();
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return timeout;
    }

    /**
     * @see OL_Socket#getStreamManager()
     */
    public final StreamManager getStreamManager()
    {
        StreamManager streamManager = null;
        try
        {
            streamManager = socket.getStreamManager();
        }
        catch (HyperCastFatalRuntimeException   fatal)   { handleFatal   (fatal);   }
        catch (HyperCastWarningRuntimeException warning) { handleWarning (warning); }
        catch (Throwable                        unknown) { handleOther   (unknown); }
        return streamManager;
    }

    /**
     * Sets the action taken when a HyperCastWarningRuntimeException
     * is occurs. The user may choose any exception action as defined
     * by the constants in this class.
     *
     * @see HyperCastAppl#EXCEPTION_LOG
     * @see HyperCastAppl#EXCEPTION_LOG_AND_THROW
     * @see HyperCastAppl#EXCEPTION_THROW
     * @see HyperCastAppl#EXCEPTION_IGNORE
     */
    public final void setWarningAction (final int action)
    {
        switch (action)
        {
            case EXCEPTION_LOG: /* fall-through */
            case EXCEPTION_LOG_AND_THROW: /* fall-through */
            case EXCEPTION_THROW: /* fall-through */
            case EXCEPTION_IGNORE: /* fall-through */
                warningAction = action;
                break;
            default:
                throw new HyperCastFatalRuntimeException ("Invalid warning action: " + action);
        }
    }

    /**
     * @return The action that will be taken when a HyperCastWarningRuntimeException occurs
     * @see HyperCastAppl#setWarningAction(int action)
     */
    public final int getWarningAction()
    {
        return warningAction;
    }

    /**
     * Sets the action taken when a HyperCastFatalRuntimeException
     * occurs. The user may choose any exception action as defined by
     * the constants in this class.
     *
     * @see HyperCastAppl#EXCEPTION_LOG
     * @see HyperCastAppl#EXCEPTION_LOG_AND_THROW
     * @see HyperCastAppl#EXCEPTION_THROW
     * @see HyperCastAppl#EXCEPTION_IGNORE
     */
    public final void setFatalAction (final int action)
    {
        switch (action)
        {
            case EXCEPTION_LOG: throw new HyperCastFatalRuntimeException ("Fatal exceptions must propagate to caller");
            case EXCEPTION_LOG_AND_THROW: /* fallthrough */
            case EXCEPTION_THROW: /* fallthrough */
                fatalAction = action;
                break;
            case EXCEPTION_IGNORE: throw new HyperCastFatalRuntimeException ("Cannot ignore fatal exceptions");
            default: throw new HyperCastFatalRuntimeException ("Invalid fatal action: " + action);
        }
    }

    /**
     * @return The action that will be taken when a HyperCastWarningRuntimeException occurs
     * @see HyperCastAppl#setFatalAction(int action)
     */
    public final int getFatalAction()
    {
        return fatalAction;
    }

    /**
     * Takes appropriate action when a HyperCastFatalRuntimeException occurs.
     * @see HyperCastAppl#setFatalAction(int action)
     */
    private final void handleFatal (final HyperCastFatalRuntimeException fatal)
    {
        String msg = "Encountered fatal exception: ";
        switch (fatalAction)
        {
            case EXCEPTION_LOG: log.println (msg + fatal); break;
            case EXCEPTION_LOG_AND_THROW: log.println (msg + fatal); /* fallthrough to "throw" case */
            case EXCEPTION_THROW: throw fatal;
            case EXCEPTION_IGNORE: /* fallthrough, i.e. default is to ignore */
            default: throw new HyperCastFatalRuntimeException ("fatalAction invalid state: " + fatalAction);
        }
    }

    /**
     * Takes appropriate action when a HyperCastWarningRuntimeException occurs.
     * @see HyperCastAppl#setWarningAction(int action)
     */
    private final void handleWarning (final HyperCastWarningRuntimeException warning)
    {
        switch (warningAction)
        {
            case EXCEPTION_LOG: log.println (warning); break;
            case EXCEPTION_LOG_AND_THROW: log.println (warning); /* fallthrough to "throw" case */
            case EXCEPTION_THROW: throw warning;
            case EXCEPTION_IGNORE: /* fallthrough, i.e. default is to ignore */
            default: throw new HyperCastFatalRuntimeException ("warningAction invalid state: " + warningAction);
        }
    }

    /**
     * Takes action when an exception occurs that is neither
     * HyperCastWarningRuntimeException nor
     * HyperCastFatalRuntimeException. The action is to create a
     * HyperCastFatalRuntimeException and throw it with the original
     * exception chained.
     * Can be overriden by the user.
     */
    private void handleOther (final Throwable throwable)
    {
        throw new HyperCastFatalRuntimeException ("Exception does not conform to HyperCast exception mechanism", throwable);
    }

    // Cannot make this private because that is more restrictive than previously declared.
    public final I_OverlayMessage receive() { throw new HyperCastFatalRuntimeException ("This should never be called"); }

    /** I_Stats#getStats */
    public Element[] getStats (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getStatsResult (doc, xpath);
    }

    /** @see I_Stats#setStats */
    public Element[] setStats (final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException
    {
        return statsProcessor.setStatsResult (doc, xpath, newValue);
    }

    /** I_Stats#getReadSchema */
    public Element[] getReadSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getReadSchemaResult (doc, xpath);
    }

    /** I_Stats#getWriteSchema */
    public Element[] getWriteSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException
    {
        return statsProcessor.getWriteSchemaResult (doc, xpath);
    }
    

//TODO: check the element name used in HyperCastAppl.
	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return null;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
	}

}
