/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.util.XmlUtil;
import hypercast.util.CurrentConfigurationSchemaGenerator;
import hypercast.util.SchemaDefaultValueCalculator;
import hypercast.util.createPrivatePropertyDocFromSchemaFile;

import java.io.*;
import java.net.*;
import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

import org.apache.xpath.domapi.XPathEvaluatorImpl;
import org.w3c.dom.xpath.XPathEvaluator;
import org.w3c.dom.xpath.XPathNSResolver;
import org.w3c.dom.xpath.XPathResult;
import org.apache.xpath.XPath;


import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.apache.crimson.tree.XmlDocument;


import org.apache.xpath.*;
import javax.xml.transform.TransformerException;

import hypercast.MonitorAndControl.I_Monitor;
import hypercast.MonitorAndControl.I_Portal;
import hypercast.MonitorAndControl.Monitor;
import hypercast.MonitorAndControl.Portal;

import java.text.ParseException;

/**	
 * This class defines the overlay socket configuration object which stores the 
 * attributes of the overlay. It contains a static method for creating an 
 * instance of overlay socket configuration object, and a set of methods for
 * operating on the attributes. It is also used to create an overlay socket or 
 * java.net.MulticastSocket interface with a socket matching its stored attributs.
 *
 * @author HyperCast Team
 * @author Dongwen Wang
 * @author Michael Nahas
 * @version 2.0, Feb. 26, 2001 
 * @author Jianping Wang
 * @version 3.0, Dec. 08, 2004 
 */

public class HyperCastConfig implements I_Stats
{
	/** The root attribute name in the configuration document. */
	public static final String ROOT_ATTRIBUTE_NAME = "/Public";

    /** Xpath string that identified the overlay ID attribute in the configuration file */
    public static final String CONFIG_ATTRIBUTE_OVERLAY_ID = ROOT_ATTRIBUTE_NAME+"/OverlayID";

	/** Xpath string that identifies the attribute which specifies the local attributes in the configuration file */
	public static final String LOCALATTRIBUTE_ATTRIBUTE_NAME = ROOT_ATTRIBUTE_NAME+"/LocalAttributes";

	/** Xpath string that identifies overlay server attribute in configuration file. */
	public final static String OVERLAY_SERVER_ATTRIBUTE_NAME = ROOT_ATTRIBUTE_NAME+"/OverlayServer";

	/** Xpath string that identifies overlay server address in configuration file. */
	public final static String OVERLAY_SERVER_URL_ATTRIBUTE_NAME = OVERLAY_SERVER_ATTRIBUTE_NAME+"/HTTPServer";

	/** Xpath string that identifies the attribute which specifies the attributes for calculating overlay hash. */
	public final static String OVERLAYHASH_ATTRIBUTE_NAME = ROOT_ATTRIBUTE_NAME+"/HashAttributes";

	/** Xpath string that identifies the log file name attribute. */
	public final static String LOGFILE_ATTRIBUTE_NAME = ROOT_ATTRIBUTE_NAME+"/LogFileName";

	/** Xpath string that identifies error log file name attribute. */
	public final static String ERRFILE_ATTRIBUTE_NAME = ROOT_ATTRIBUTE_NAME+"/ErrorFileName";

	/** The root attribute name in the configuration document. */
	public static final String CONFIGOBJ_ATTRIBUTE_NAME = "/Public/SocketConfig";

    /** Name used in properties file to indicate no logFileName or errFileName. */
    public final static String NO_FILE = "";
	
    /** Name of file to write log messages to.  If "stderr" or "stdout" then
     * messages are written to System.err or System.out respectively. */
    private String logFileName;
    
    /** The opened log file. */
    public PrintWriter log;

    /** Name of file to write error messages to.  If "stderr" or "stdout" then
     * messages are written to System.err or System.out respectively. */
    private String errFileName;

    /** The opened error file. */
	public PrintWriter err;

    /** A PrintWriter guaranteed not to go to a NullWriter (/dev/null). */
    public PrintWriter seriousErr;
    
    /** A hash of the OverlayID . */
    private int overlayHash;

	/** Dom Document instance which store the public attributes of the Overlay */
	Document propertyDoc = null;

	/**
	 * Local attributes. It defines the attributes (in XPath expression format)
	 * that are specified locally and not overrided by the attributes downloaded
	 * from overlay server.
	 */
	Vector localAttributes;


	/** Dom Document instance which store the private attributes of the Overlay */
	Document privatePropertyDoc = null;

	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	StatsProcessor statsPro;

	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;

	/**
	 * Debug flag.
	 */
	final static boolean debug = false;
	
	/**
	 * Construct an HyperCastConfig object from given configuration DOM object
	 * @param doc
	 */
	public HyperCastConfig(Document doc){
		/*
		 * Obtain Document instance from local configuration file.
		 */
		propertyDoc = doc;

		/*
		 * Obtain local attributes which are stored in the following Properties instance.
		 */
		localAttributes = this.getLocalAttributes(propertyDoc);

		/*
		 * Sets the private Properties instance.
		 */
	   this.setPrivateProperties();
	   
	   /*
		* Initialize this object.
		*/
	   this.Initialization();
	}
	/**
	 * Return the XML document of this configuration object
	 */
	public Document getDocument(){
		return propertyDoc;
	}
	/**
	 * Construct an HyperCastConfig object from given configuration file.
	 * 
	 * @param configurationFilename  configuration file name.
	 */
	public HyperCastConfig(String configurationFilename) 
	{
		/*
		 * Obtain Document instance from local configuration file.
		 */
		propertyDoc = this.createPropertiesFromConfigurationFile(configurationFilename);

		/*
		 * Obtain local attributes which are stored in the following Properties instance.
		 */
		localAttributes = this.getLocalAttributes(propertyDoc);

		/*
		 * Sets the private Properties instance.
		 */
	   this.setPrivateProperties();
	   
	   /*
		* Initialize this object.
		*/
	   this.Initialization();
	}

	/**
	 * Construct an HyperCastConfig object by downloading the configuration file from overlay server.
	 * 
	 * @param serverURL	URL instance representing the address of the overlay server.
	 * @param OverlayId string representing the overlay ID.
	 * 
	 */
	public HyperCastConfig(URL serverURL, String OverlayId) 
	{
		propertyDoc = this.DownloadConfig(serverURL, OverlayId);

		if (propertyDoc == null)
		{
			throw new HyperCastFatalRuntimeException ("Couldn't create property Document when constructing this HyperCastConfig instance!");
		}

		/*
		 * Obtain local attributes which are stored in the following Properties instance.
		 */
		localAttributes = this.getLocalAttributes(propertyDoc);

	   /*
		* Sets the private Properties instance.
		*/
	   this.setPrivateProperties();
	   
	   /*
		* Initialize this object.
		*/
	   this.Initialization();
	}


    /**  Returns a new OL_Socket object. 
     */
    public I_OverlaySocket createOverlaySocket (I_ReceiveCallback callback) 
    {
        return createOverlaySocket (callback, null, null);
    }
    
    /**  Returns a new OL_Socket object. 
     */
    public I_OverlaySocket createOverlaySocket(I_ReceiveCallback callback, NotificationHandler nh) 
    {
        return createOverlaySocket (callback, nh, null);
    }

    /**  Returns a new OL_Socket object. 
     */
    public I_OverlaySocket createOverlaySocket (I_ReceiveCallback callback, NotificationHandler nh, I_InterceptionCallback Icb) 
    {
            return new OL_Socket (this, callback, nh, Icb);
    }


	/**
	 * creates a I_Monitor instance
	 */
	public I_Monitor createMonitor(){
		return new Monitor(this);
	}

	/**
	 * Creates a I_Portal instance
	 */
	public I_Portal createPortal(final I_Stats stats, final String statsRoot){
		return new Portal(this, stats, statsRoot);
	}
	/**
	 * Creates a I_Portal instance
	 */
	public I_Portal createPortal(final I_Stats stats){
		return new Portal(this, stats, null);
	}
	


/** Creates a new object implementing Java.net.MulticastSocket that 
 * sends the messages over a new OL_Socket created using the createSocket function. 
 * @see HyperCastConfig#createOverlaySocket 
 * @see MulticastSocketMasquerade 
 */
    public MulticastSocket createJavaMulticastSocket() 
    {
        I_OverlaySocket sock = createOverlaySocket((I_ReceiveCallback) null);

        MulticastSocketMasquerade msock = null;
        try 
        {
            msock = new MulticastSocketMasquerade();
        }
        catch (Exception ignored) 
        {
        }

        msock.setOverlaySocket(sock);
        
        return msock;
    }

    /**
     * Creates an overlay configuration object based on the given filename.
     * It is called from application code.
     *
     * @param configurationFilename The name of the configuration file used to 
     * construct the property database,	cannot be null or empty.
     * 
     * @return A configuration object that contains the contents of the given configuration file.
     * 
     * @throws HyperCastConfigException when configuration filename is not valid or construction of
     * 	HyperCastConfig instance fails;
     */
    public static HyperCastConfig createConfig (String configurationFilename)
    {
        Exception filenameValidityException = null;
        if (configurationFilename == null)
        {
            filenameValidityException = new NullPointerException ("A non-null filename must be provided");
        }
        else if (configurationFilename.equals (""))
        {
            filenameValidityException = new IllegalArgumentException ("A non-empty filename must be provided");
        }

        if (filenameValidityException != null)
            throw new HyperCastConfigException (filenameValidityException);

		/*
		 * Creates an HyperCastConfig instance from local configuration file. 
		 */
		HyperCastConfig ConfObj = new HyperCastConfig(configurationFilename);
		
        try
        {
			String overlayId = ConfObj.getTextAttribute(new XPath(CONFIG_ATTRIBUTE_OVERLAY_ID, null, null, XPath.MATCH));

			String overlayServer = ConfObj.getTextAttribute(new XPath(OVERLAY_SERVER_ATTRIBUTE_NAME, null, null, XPath.MATCH));

			//System.out.println("overlayId="+overlayId+", overlayServer="+overlayServer+".");


			boolean existOverlayId = !((overlayId == null) || (overlayId.equals("")));
			boolean existOverlayServer = !((overlayServer == null) || (overlayServer.equals("")) || (overlayServer.equalsIgnoreCase("None")));

			/*
			  * if no overlay Id is provided in the configuration file and no overlay server
			  * is specified, this method creates a new overlay ID.
			  */ 
			if (!existOverlayId && !existOverlayServer)
			{
				//System.out.println("call ConfObj.createandsetOverlayId()");
				ConfObj.createandsetOverlayId();
			}

			//String newOverlayId = ConfObj.getTextAttribute(new XPath(CONFIG_ATTRIBUTE_OVERLAY_ID, null, null, XPath.MATCH));
			//System.out.println("new overlayID="+newOverlayId+".");


			if (existOverlayServer)
			{
				URL serverURL = ConfObj.getURLAttribute(new XPath(OVERLAY_SERVER_URL_ATTRIBUTE_NAME, null, null, XPath.MATCH));
		
				String OverlayId = overlayId;
				if ((!existOverlayId) || (!ConfObj.doesOverlayExist(serverURL, overlayId)))
				{
					if (debug)
						System.out.println("Overlay " + overlayId + " doesn't exist!");

					OverlayId = ConfObj.UploadConfig(serverURL);
					//System.out.println("overlayID after uploading is "+OverlayId+".");
				}

				Document DConfig = ConfObj.DownloadConfig(serverURL, OverlayId);
				
				ConfObj.MergeConfig(DConfig);

				ConfObj.Initialization();
			}
        }
        catch (Exception e)
        {
            System.err.println("Exception when creating overlay configuration object.");
			System.err.println("Exception: "+e.getMessage());
			System.err.println("Exiting.");
			e.printStackTrace(System.err);
			System.exit(1);
        }
        
        return ConfObj;
    }

	/**
	 * Create a new Document instance from XML configuration file.
	 * 
	 * @param configurationFilename	name of the file which defines the properties 
	 * 	for starting an overlay socket.
	 * 
	 * @return	a Document instance.
	 */
	private Document createPropertiesFromConfigurationFile(String configurationFilename)
	{
		if (debug)
			System.out.println ("Creating DOM Document using configuration file \"" + configurationFilename + "\"...");

		if (configurationFilename.endsWith (".xml")
			||
			configurationFilename.endsWith (".XML"))
		{
			Document property = null;
			try
			{
				property = XmlUtil.createDocument(configurationFilename);
			}
			catch (final IOException ioe)
			{
				throw new HyperCastConfigException("Cannot create property Document from file "+configurationFilename+".");
			}
			catch (final ParseException ioe)
			{
				throw new HyperCastConfigException("Failed in parsing file "+configurationFilename+".");
			}
		
			if (property == null)
			{
				throw new HyperCastConfigException ("Fails in creating Documnent instance from file: " + configurationFilename, null);
			}
			else
			{
				if (debug)
				{
					try
					{
						System.out.println("HyperCastConfig:createOLConfig: create property Doc from local configuration file is:");
						XmlUtil.writeXml(property, System.out);
					}
					catch (IOException ee)
					{
						System.err.println("HypercastConfiguration:createPropertiesFromConfigurationFile: IO exception when outputing property file.");
					}
				}

				return property;
			}
		}
		else
		{
			IllegalArgumentException iae = new IllegalArgumentException ("Unknown file extension: \"" + configurationFilename + "\"");
			throw new HyperCastConfigException (iae);
		}

	}

	/** 
	 * Initialize the variables in this object.
	 */
	private void Initialization()
	{
		if (log != null)
		{
			log.close();
		}

		if (err != null)
		{
			err.close();
		}

		/*
		 * Open err log file and log file.
		 */
		openErrAndLogFiles();

		/*
		 * Create overlay ID hash value.
		 */
		overlayHash = setOverlayHash();

		/*
		 * Initialize statistics structure in this object.
		 */
		InitStatisticsStructure();
		
	}

	/** 
	 * Sets the private property Document which is obtained from a private schema file.
	 */
	private void setPrivateProperties()
	{
		try
		{
			privatePropertyDoc = (new createPrivatePropertyDocFromSchemaFile()).getPrivateProperties();

			if (debug)
			{
				System.out.println("HyperCastConfig:createOLConfig: private properties is:");
				XmlUtil.writeXml(this.getPrivateConfiguration(), System.out);
			}
		}
		catch (IOException e)
		{
			throw new HyperCastConfigException("Cannot create private properties from private schema file:"+e);
		}
	}


	/** 
	 * Obtain local attributes from the property Document instance.
	 * 
	 * @param propertydoc	Document instance which represents the overlay properties.
	 * 
	 * @return a vector of local attribute xpath expressions.
	 */
	private Vector getLocalAttributes(Document propertydoc)
	{
		return getListOfAttributeNames(propertydoc, LOCALATTRIBUTE_ATTRIBUTE_NAME);
	}

	/** 
	 * Find all Nodes which match given xpath expression.
	 * 
	 * @param propertydoc Document instance to be queried.
	 * @param xpathStr Xpath expression.
	 * 
	 * @return vector of Nodes which match given xpath expression.
	 */
	static Vector XpathQuery(Document propertydoc, String xpathStr)
	{
		Vector result = new Vector();

		XPathEvaluator evaluator = new XPathEvaluatorImpl(propertydoc);
		XPathNSResolver resolver = evaluator.createNSResolver(propertydoc);
      
		XPathResult xp = (XPathResult)evaluator.evaluate(xpathStr, propertydoc, resolver, XPathResult.UNORDERED_NODE_ITERATOR_TYPE, null);

		Node n;
		while ((n = xp.iterateNext())!= null)
		{        
			result.addElement(n);
		}

		return result;
	}

	/** 
	 * Find the Node which matchs the given xpath expression. 
	 * If more than one Nodes are found, throw exception.
	 * 
	 * @param propertydoc Document instance to be queried.
	 * @param xpathStr Xpath expression.
	 * 
	 * @return The Node which matches given xpath expression.
	 */
	static Node singleNodeXpathQuery(Document propertydoc, String xpathStr)
	{
		Vector result = XpathQuery(propertydoc, xpathStr);

		if (debug)
			System.out.println("HypercastCastConfig:singleNodeXpathQuery: XPath string is "+xpathStr+".");

		if (debug)
			System.out.println("HypercastCastConfig:singleNodeXpathQuery: XPath string "+xpathStr+", result.size()="+result.size());

		if (result.size() == 0)
		{
			return null;
		}
		else if (result.size() > 1)
		{
			throw new IllegalStateException("Duplicated LocalAttribute properties!");
		}
		else
		{
			if (debug)
			{
				Node nn = (Node) result.elementAt(0);
				System.out.println("HypercastCastConfig:singleNodeXpathQuery: the name of found node is "+nn.getNodeName()+", node value is "+nn.getNodeValue());
			}

			return (Node) result.elementAt(0);
		}
	}


	/** Check if the overlay with given overlay ID exists at the server.
	 * 
	 * @param serverURL	URL instance representing the address of overlay server.
	 * @param overlayID	String representing the identifier of an overlay.
	 * 
	 * @throws IllegalStateException if overlay server property was not set or was set incorrectly.
	 */

	public boolean doesOverlayExist(URL serverURL, String overlayID)
	{
		if (overlayID == null || overlayID.equals("")) 
			throw new IllegalArgumentException("Overlay id cannot be null or \"\".");

		/*
		 * If overlay server is not used, return true.
		 */
		if (serverURL == null)
		{
			return true;
		}

		String serverURLString = serverURL.toString();

		if (debug)
			System.out.println("serverURL="+serverURLString);
		
		String reply = null;
		try
		{
			reply = HTTP_ServerUtility.query(serverURLString + "?cmd=test&"+ 
								"OverlayID=" +  HTTP_ServerUtility.encodeURLString(overlayID));
		}
		catch (MalformedURLException mue) 
		{
			throw new IllegalStateException("Malformed URL: "+serverURLString);
		}
		catch (IOException ioe) 
		{
			throw new HyperCastFatalRuntimeException ("Overlay Server is not started or there are other network problems!", ioe);
		}

		if (reply.startsWith("ERROR"))
		{
			throw new HyperCastFatalRuntimeException (reply);
		}
		else
		{
			return reply.startsWith("YES");
		}
	}


	/** 
	 * This method uploads the property XML file to the overlay server. If the uploaded property XMl document cintains
	 * overlay ID, the overlay ID is returned by the overlay server; if , the uploaded property XMl document
	 * doesn't contain an overlay ID, the overlay server will create one and return it.
	 * 
	 * @param serverLocation	URL instance representing the address of the overlay server.
	 * 
	 * @throws IllegalStateException if overlay server property was not set or was set incorrectly.
	 * @throws HyperCastFatalRuntimeException if overlay server cannot be connected or ERROR reply
	 * is returned by the overlay server.
	 */
	public String UploadConfig(URL serverLocation)
	{
		/*
		 * If overlay server is not specified, do nothing.
		 */
		if (serverLocation == null)
		{
			if (debug)
				System.out.println("HyperCastConfig: UploadConfig: overlay server is not specified!");
			
			return null;
		}

			String serverPrefix = serverLocation.toString();
			String reply = null;
			try
			{
				if (debug)
					System.out.println("Send upload query to overlay server "+serverPrefix);

				reply = HTTP_ServerUtility.query(serverPrefix + "?cmd=upload&"+ 
					  //"OverlayID=" + HTTP_ServerUtility.encodeURLString(overlayid) + "&" 
					  "configFile=" + HTTP_ServerUtility.encodeURLString(XmlUtil.getStringFromDocument(propertyDoc)));
			}
			catch (MalformedURLException mue) 
			{
				throw new IllegalStateException("Malformed URL: "+serverPrefix);
			}
			catch (IOException ioe) 
			{
				throw new HyperCastFatalRuntimeException ("Overlay Server is not started or there are other network problems!", ioe);
			}

			if (reply == null)
			{
				System.err.println("Upload configuration file failed!");
				return null;
			}
			else if (reply.startsWith("ERROR"))
			{
				if (reply.equals("ERROR: Overlay already exists!"))
				{
					System.err.println("Overlay already exists!");
					return null;
				}
				else if (reply.equals("ERROR: Property file is not a valid XML document!"))
				{
					System.err.println("Poor formated XML configuration file!");
					return null;
				}
				else 
				{
					throw new HyperCastFatalRuntimeException (reply);
				}
			}
			else 
			{
				int andSign = reply.indexOf("&");
				String overlayidSubStr = reply.substring(andSign+1);
				int equalSign = overlayidSubStr.indexOf("=");
				String configfileSubStr = HTTP_ServerUtility.decodeURLString(overlayidSubStr.substring(equalSign+1));

				return configfileSubStr;
			}
	}


	/** 
	 * This method downloads the property XML file specified by the given overlay ID from the overlay server.
	 * 
	 * @param serverLocation	URL instance representing the address of the overlay server.
	 * @param overlayID		the overlay ID which specifies the XML configuration file to download.
	 * 
	 * @throws IllegalStateException if overlay server property was not set or was set incorrectly.
	 * @throws HyperCastFatalRuntimeException if overlay server cannot be connected or ERROR reply
	 * is returned by the overlay server.
	 */
	public Document DownloadConfig(URL serverLocation, String overlayID)
	{
		/*
		 * If overlay server is not specified, do nothing.
		 */
		if (serverLocation == null)
		{
			if (debug)
				System.out.println("HyperCastConfig: UploadConfig: overlay server is not specified!");
			
			return null;
		}

			String serverPrefix = serverLocation.toString();
			String reply = null;
			try
			{
				if (debug)
					System.out.println("Send download query to overlay server "+serverPrefix);

				reply = HTTP_ServerUtility.query(serverPrefix + "?cmd=download&"+ 
					  "OverlayID=" + HTTP_ServerUtility.encodeURLString(overlayID));
			}
			catch (MalformedURLException mue) 
			{
				throw new IllegalStateException("Malformed URL: "+serverPrefix);
			}
			catch (IOException ioe) 
			{
				throw new HyperCastFatalRuntimeException ("Overlay Server is not started or there are other network problems!", ioe);
			}

			if (reply == null)
			{
				System.err.println("Download configuration file failed!");
				return null;
			}
			else if (reply.startsWith("ERROR"))
			{
				throw new HyperCastFatalRuntimeException (reply);
			}
			else 
			{
				int andSign = reply.indexOf("&");
				String confReplyString = reply.substring(andSign+1);
				int equalSign = confReplyString.indexOf("=");
				String configfileStr = confReplyString.substring(equalSign+1);
				
				String propertyDocStr = HTTP_ServerUtility.decodeURLString(configfileStr);
				
				Document doc = null;
				try
				{
					doc = XmlUtil.createDocumentFromString(propertyDocStr);
				}
				catch (final ParseException ioe)
				{
					throw new HyperCastConfigException("Cannot create XML Document from downloaded file.");
				}
								
				return doc;
			}
	}


	/** 
	 * Merges the downloaded configuration file with local property file. If overlay server is not used, 
	 * this method does nothing.
	 * 
	 */
	public void MergeConfig(Document fromDOM)
	{
		if (fromDOM == null)
		{
			System.err.println("The given Document is null!");
			return;
		}

		Node originalPropertyNode = singleNodeXpathQuery(propertyDoc, ROOT_ATTRIBUTE_NAME);
		Node downloadPropertyNode = singleNodeXpathQuery(fromDOM, ROOT_ATTRIBUTE_NAME);

		if ((originalPropertyNode == null) || (downloadPropertyNode == null))
		{
			System.err.println("The given Document or local property Document has wrong format!");
			return;
		}
				
		this.combineConfigDocuments(propertyDoc, originalPropertyNode, 
													fromDOM, downloadPropertyNode, "");
	}

	/** 
	 * Check if two Elements in two different Documents. If they are different, the second Element is used to
	 * replace or change the first Element.
	 * 
	 * @param firstDoc	Document to be changed.
	 * @param firstEle	Node in the first Document to be checked and changed.
	 * @param secondDoc	Document used to check and change the first Document.
	 * @param secondEle	Node used to check and change the first Node.
	 * @param xpathStr	Xpath string representing the current Element in the local property Document to be compared.
	 */
	private void combineConfigDocuments(Document firstDoc, Node firstEle, Document secondDoc, Node secondEle, String xpathStr)
	{
		if (debug)
		{
			System.out.println("OverlaySocketServer:combineConfigDocuments: name of two nodes are: "+firstEle.getNodeName()+" and "+secondEle.getNodeName());
			System.out.println("xpathStr="+xpathStr+", is localattribute? "+isLocalAttribute(xpathStr));
		}

		/*
		 * If "xpathStr" represnets a local attribute, do nothing.
		 */
		 if (isLocalAttribute(xpathStr))
		 {
		 	return;
		 }

		/*
		 * If two Nodes are both TEXT Nodes but have the same values, do nothing.
		 */
		if (((firstEle.getNodeType() == Node.TEXT_NODE) &&
			(secondEle.getNodeType() == Node.TEXT_NODE) &&
			(firstEle.getNodeValue().equals(secondEle.getNodeValue()))))
		{
			return;
		}
		/*
		 * If two Nodes are different types, or they are both TEXT Nodes but have different values,
		 * replace "firstEle" with "secondEle".
		 */
		else if ((secondEle.getNodeType() != firstEle.getNodeType()) ||
			((firstEle.getNodeType() == Node.TEXT_NODE) &&
			(secondEle.getNodeType() == Node.TEXT_NODE) &&
			(!firstEle.getNodeValue().equals(secondEle.getNodeValue()))))
		{
			Node parentNode = firstEle.getParentNode();
			Node tmpNode = firstDoc.importNode(secondEle, true);

			parentNode.replaceChild(tmpNode, firstEle);
			
			return;
		}

		/*
		 * If two Nodes are not TEXT Nodes, do the following things.
		 */
		
		/*
		 * If current Element is a local attribute, do nothing.
		 */
		String newXpathStr = xpathStr+"/"+firstEle.getNodeName();
		if (isLocalAttribute(newXpathStr))
		{
			return;
		}
		/*
		 * If "firstEle" represents a non-local attribute, two Elements are not TEXT Nodes and have different names, 
		 * replace the "firstEle" with "secondEle".
		 */
		else if (!firstEle.getNodeName().equals(secondEle.getNodeName()))
		{
			Node parentNode = firstEle.getParentNode();
			Node tmpNode = firstDoc.importNode(secondEle, true);
			
			parentNode.replaceChild(tmpNode, firstEle);
			return;
		}
		/*
		 * If two Elements are not TEXT Nodes and have the same name, compare their children.
		 */
		else
		{
			NodeList secondNL = secondEle.getChildNodes();
			NodeList firstNL = firstEle.getChildNodes();

			/*
			 * If both "firstEle" and "secondEle" has only one child seperately, compare their
			 * child.
			 */
			if ((secondNL.getLength() == 1) && (firstNL.getLength() == 1))
			{
				combineConfigDocuments(firstDoc, firstNL.item(0), secondDoc, secondNL.item(0), newXpathStr);
				return;
			}
			
			/*
			 * When one of the two Elements has multiple children, do the following things.
			 */
			/*
			 * Build a hash table. The key of the table is node name and value is
			 * the times this node occurs.
			 */
			HashMap childrenNameHash = new HashMap();
			for (int i=0; i<secondNL.getLength(); i++)
			{
				if (secondNL.item(i).getNodeType() == Node.TEXT_NODE)
				{
					continue;
				}

				String name = secondNL.item(i).getNodeName();
				
				if (childrenNameHash.containsKey(name))
				{
					int num = Integer.parseInt((String) childrenNameHash.get(name));
					childrenNameHash.put(name, "" + (num++));
				}
				else
				{
					childrenNameHash.put(name, "1");
				}
			}

			/*
			 * For each child with different name, do checking.
			 */
			for (Iterator it=childrenNameHash.entrySet().iterator(); it.hasNext(); )
			{ 
				Map.Entry entry = (Map.Entry)it.next(); 
				String nodeName = (String) entry.getKey(); 
				String key = (String) entry.getValue();

				int occurNum = Integer.parseInt(key);

				//System.out.println(key+"     " + occurNum); 
				//System.out.println("xpath string is "+xpathStr+"/"+nodeName); 

				if (isLocalAttribute(newXpathStr+"/"+nodeName))
				{
					continue;
				}

				/*
				 * If this node occurs one time in the second Element, check it with the first Element.
				 */
				if (occurNum == 1)
				{
					Node secondNode = singleNodeXpathQuery(secondDoc, newXpathStr+"/"+nodeName);

					Vector nodeVec = XpathQuery(firstDoc, newXpathStr+"/"+nodeName);
			
					/*
					 * If the first Element doesn't contain a child which has the name of second 
					 * Element's child, add a child to the first Element.
					 */
					if (nodeVec.size() == 0)
					{
						Node tmpNode = firstDoc.importNode(secondNode, true);
	
						firstEle.appendChild(tmpNode);
					}
					/*
					 * If the first Element has more than one childs with the same name of
					 * the second Element's child, delete all children of the first Element
					 * and add a new child to it.
					 */
					else if (nodeVec.size() > 1)
					{
						for (int i=0; i<nodeVec.size(); i++)
						{
							firstEle.removeChild((Node) nodeVec.elementAt(i));
						}
						Node tmpNode = firstDoc.importNode(secondNode, true);
	
						firstEle.appendChild(tmpNode);
					}
					/*
					 * Otherwise, both the first Element and second Element have one child.
					 * Continue the check and update on their children.
					 */
					else
					{
						combineConfigDocuments(firstDoc, (Node) nodeVec.elementAt(0), secondDoc, secondNode, newXpathStr);
					}
				}
				/*
				 * If node occurs multiple times in the second Element, check them with the first Element.
				 */
				else if (occurNum > 1)
				{
					Vector secondNodeVec = XpathQuery(secondDoc, newXpathStr+"/"+nodeName);

					Vector firstNodeVec = XpathQuery(firstDoc, newXpathStr+"/"+nodeName);
					
					/* For a child, assume its name is x. If the first Element has different
					 * number of children with name x from thee second Element, delete all children 
					 * with name x of the first Element and add a new child to it.
					*/
					if (secondNodeVec.size() != firstNodeVec.size())
					{
						for (int i=0; i<firstNodeVec.size(); i++)
						{
							firstEle.removeChild((Node) firstNodeVec.elementAt(i));
						}
						
						for (int i=0; i<secondNodeVec.size(); i++)
						{
							Node tmpNode = firstDoc.importNode((Node) secondNodeVec.elementAt(i), true);
	
							firstEle.appendChild(tmpNode);
						}
					}
					/*
					 * Otherwise, check each child with name x.
					 */
					else
					{
						for (int i=0; i<secondNodeVec.size(); i++)
						{
							combineConfigDocuments(firstDoc, (Element) firstNodeVec.elementAt(i), 
									secondDoc, (Element) secondNodeVec.elementAt(i), newXpathStr);
						}
					}
				}
				else
				{
					//It is impossible.
				}
			}
		}
	}

	/**
	 * Returns true if the given xpath string represents a local attribute. Returns false otherwise.
	 */
	private boolean isLocalAttribute(String xpathStr)
	{
		for (int i=0; i<localAttributes.size(); i++)
		{
			String localAttr = (String) localAttributes.elementAt(i);
			
			//if (xpathStr.trim().startsWith(localAttr))
			if (xpathStr.trim().equals(localAttr))
			{
				return true;
			}
		}
		return false;
	}

	/**
	 * Return the public configuration property document contained in this HyperCastConfig 
	 * instance.
	 */
	public Document getPublicConfiguration()
	{
		return this.propertyDoc;
	}

	/**
	 * Return the private configuration property document contained in this HyperCastConfig 
	 * instance.
	 */
	public Document getPrivateConfiguration()
	{
		return this.privatePropertyDoc;
	}

    
	/**
	 * Set the overlay ID of this HyperCastConfig instance to given value.
	 */
	public void setOverlayID(final String overlayid)
	{
		Node overlayidNode = singleNodeXpathQuery(propertyDoc, CONFIG_ATTRIBUTE_OVERLAY_ID);

		NodeList nl = overlayidNode.getChildNodes();

		if ((overlayidNode == null) || (nl.getLength() > 1))
		{
			System.err.println("overlay id is not correctly specified!");
			return;
		}

		if (nl.getLength() == 0)
		{
			overlayidNode.appendChild(propertyDoc.createTextNode(overlayid));
		}
		else
		{
			overlayidNode.replaceChild(propertyDoc.createTextNode(overlayid), nl.item(0));
		}

		/*
		 * Reset overlayHash
		 */		
		overlayHash = setOverlayHash();
	}

	/**
	 * @return the default value in string format of a property.
	 */
	private String queryDefaultAttribute(final String name)
	{
		return ConfigurationDefaults.getConfigurationDefaults().getProperty (name);
	}


	/**
	 * @return the value in string format of a property. The default value of the 
	 * property is not considered in this method.
	 */
	private String queryAttributeInConfigDoc(Document doc, final String name)
	{
		if (debug)
			System.out.println("HypercastCastConfig:queryAttributeInConfigDoc: attribute name is "+name+".");

		Node matchNode = singleNodeXpathQuery(doc, name);

		if (debug)
		{
			if (matchNode == null)
			{
				System.out.println("HypercastCastConfig:queryAttributeInConfigDoc: call singleNodeXpathQuery(propertyDoc, name) for "+name+" result is null.");
				return null;
			}
			else
				System.out.println("HypercastCastConfig:queryAttributeInConfigDoc: call singleNodeXpathQuery(propertyDoc, name) for "+name+" result is not null, num of children is "+matchNode.getChildNodes().getLength());


			NodeList nl = matchNode.getChildNodes();

			for (int i=0; i<nl.getLength(); i++)
			{
				Node node = nl.item(i);

				if (node.getNodeType() != Node.TEXT_NODE){
					System.out.println("Child element "+i+" has name --"+node.getNodeName()+"--.");
				}
				else
				{
					System.out.println("Child TEXT node "+i+" has name --"+node.getNodeValue()+"--.");
				}
				
				if (node.getNodeName().indexOf("\n")>=0)
				{
					System.out.println("This node contains return value.");
				}
			}
			
		}

		/*
		 * If no Node is found which matches xpath expression "name", return null.
		 */
		if ((matchNode == null) || (matchNode.getChildNodes().getLength() != 1))
		{

			return null;
		}
		else
		{
			Node resultNode = matchNode.getChildNodes().item(0);
			if (debug)
			{
				if (resultNode == null)
					System.out.println("HypercastCastConfig:queryAttributeInConfigDoc: resultTextNode is null.");
				else
				{
					System.out.println("HypercastCastConfig:queryAttributeInConfigDoc: resultTextNode has type "+resultNode.getNodeType());
				}
			}

			if (resultNode.getNodeType() != Node.TEXT_NODE)
			{
				//System.out.println("name of result node is "+resultTextNode.getNodeName());
				//throw new IllegalStateException("The attribute " + name + " is specified incorrectly!");
				//System.out.println(name + " is not a string attribute!");
				//return null;
				return resultNode.getNodeName();
			}
			else
			{
				return (String) resultNode.getNodeValue();
			}
		}
	}


//TODO: need implementation.
	/**
	 * Returns true if the given attribute is defined in the configuration Document. Returns
	 * false otherwise.
	 */
	public boolean validatePublicConfig (final String XSDfilename)
	{
		return true;
	}

	/**
	 * Returns true if the given attribute is defined in the configuration Document. Returns
	 * false otherwise.
	 */
	public boolean testIfAttributeIsDefined (final XPath attrXpath)
	{
	   final String attribute = attrXpath.getPatternString();

	   String result = queryAttributeInConfigDoc(propertyDoc, attribute);

	   if (result != null)
	   {
		   return true;
	   }
	   else
	   {
			return false;
	   }
	}


	/**
	 * @return the value of attribute in string format.
	 * 
	 * @deprecated use getTextAttribute(final XPath attrXpath).
	 */
	public String getProperty(final String name)
	{
		 return getStringProperty (name);
	}


//
//	String Attribute Accessors
//

	 /**
	  * @return a string property; a default value is returned if this property was not specified in the configuration file.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * It is possible that this method should be deprecated, but the
	  * jury is still out. Ideally the default value should be
	  * extracted from the HyperCast XML Schema file. However, there is
	  * some debate (Nov 2004) about what to do when an attribute value
	  * is not defined in a configuration file and that attribute has
	  * no "natural" default.
	  */
	 public String getStringProperty (final String attribute, final String defaultValue)
	 {
		 //return properties.getProperty (attribute, defaultValue);
		 return getStringProperty (attribute);
	 }

	 /**
	  * @return a string property; a default value is returned if this property was not specified in the configuration file.
	  * 
	  * @deprecated use getTextAttribute (final XPath attrXpath)
	  */
	 public String getStringProperty (final String attribute)
	 {
		 try
		 {
			 return getTextAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }


	 /**
	  * @return a string property; a default value is returned if this property was not specified in the configuration file.
	  */
	 public String getTextAttribute (final XPath attrXpath)
	 {
		final String attribute = attrXpath.getPatternString();

		String result = queryAttributeInConfigDoc(propertyDoc, attribute);

		if (debug)
			System.out.println("HypercastCastConfig:getTextAttribute: result ="+result+".");

		/*
		 * If the given xpath is defined in property Document, return the value in the property Document.
		 */
		if (result != null)
		{
			return result;
		}
		/*
		 * Otherwise, try to get the default value.
		 */
		else
		{
			String defaultValue = queryDefaultAttribute(attribute);

			/*
			 * If both property Document and Default value properties don't contain the given attribute,
			 * throws exception.
			 */
			if (defaultValue == null)
			{
				throw new HyperCastConfigException ("Attribute \"" + attribute + "\" doesn't exist!");
			}
			else
			{
				return defaultValue;
			}
		}
	 }


	  /**
	   * Sets the value of scalar string attribute.
	   */
	  public void setTextAttribute(final XPath xpath, final String value)
	  {
		  String name = xpath.getPatternString();

		  Node resultNode = singleNodeXpathQuery(propertyDoc, name);

		  if (resultNode == null)
		  {
			  throw new HyperCastConfigException ("Attribute \"" + name + "\" doesn't exist!");
		  }
		  else if (resultNode.getChildNodes().getLength() > 1)
		  {
			  throw new HyperCastConfigException ("Attribute \"" + name + "\" is not a scalar attribute!");
		  }
		  else if (resultNode.getChildNodes().getLength() == 0)
		  {
			  resultNode.appendChild(propertyDoc.createTextNode(value));
		  }
		  else
		  {
			  Node resultTextNode = resultNode.getChildNodes().item(0);
			  if (resultTextNode.getNodeType() != Node.TEXT_NODE)
			  {
				  throw new HyperCastConfigException ("Attribute \"" + name + "\" is not a writable attribute!");
			  }
			  else
			  {
				  resultNode.removeChild(resultTextNode);
				  resultNode.appendChild(propertyDoc.createTextNode(value));
			  }
		  }
		  
		/*
		 * If the set attribute is one of hash attributes, reset overlayHash
		 */		
		Vector hashAttributeVec = getListOfAttributeNames(this.propertyDoc, HyperCastConfig.OVERLAYHASH_ATTRIBUTE_NAME);
		for (int i=0; i<hashAttributeVec.size(); i++)
		{
			String attr = (String) hashAttributeVec.elementAt(i);
			if (name.equals(attr))
			{
				overlayHash = setOverlayHash();
			}
		}
		  
	  }

	/**
	 * @return a string private attribute. No default value is defined for a private attribute.
	 */
	public String getPrivateTextAttribute (final XPath attrXpath)
	{
		final String attribute = attrXpath.getPatternString();

		String result = queryAttributeInConfigDoc(privatePropertyDoc, attribute);

		if (debug)
			System.out.println("HypercastCastConfig:getTextAttribute: result ="+result+".");

		/*
		 * If the given xpath is defined in property Document, return the value in the property Document.
		 */
		if (result != null)
		{
			return result;
		}
		/*
		 * Otherwise, throws exception. A private attribute doesn't have default value.
		 */
		else
		{
			throw new HyperCastConfigException ("Private attribute \"" + attribute + "\" doesn't exist!");
		}
	}

	/**
	 * Sets the value of private scalar string attribute. If the given attribute (specified
	 * by the xpath) doesn't exist, create it.
	 */
	public void setPrivateTextAttribute(final XPath xpath, final String value)
	{
		String name = xpath.getPatternString();

		Node resultNode = singleNodeXpathQuery(privatePropertyDoc, name);

		if (resultNode == null)
		{
			Vector attrVec = this.getAttributeNamesInXpathString(name);

			Node parent = (Node) privatePropertyDoc;
			
			String searchStr = "";
			for (int i=attrVec.size()-1; i>=0; i--)
			{
				searchStr = searchStr + "/" + ((String) attrVec.elementAt(i));
				Node r = singleNodeXpathQuery(privatePropertyDoc, searchStr);
				
				if (r != null)
				{
					parent = r;
					continue;
				}
				else
				{
					if (i == 0)
					{
						Text textNode = privatePropertyDoc.createTextNode((String) attrVec.elementAt(i));
						parent.appendChild(textNode);
					}
					else
					{
						Element ele = privatePropertyDoc.createElement((String) attrVec.elementAt(i));
						parent.appendChild(ele);
					}
				}
			}
		}
		else if (resultNode.getChildNodes().getLength() > 1)
		{
			throw new HyperCastConfigException ("Private attribute \"" + name + "\" is not a scalar attribute!");
		}
		else if (resultNode.getChildNodes().getLength() == 0)
		{
			resultNode.appendChild(privatePropertyDoc.createTextNode(value));
		}
		else
		{
			Node resultTextNode = resultNode.getChildNodes().item(0);
			if (resultTextNode.getNodeType() != Node.TEXT_NODE)
			{
				throw new HyperCastConfigException ("Private attribute \"" + name + "\" is not a writable attribute!");
			}
			else
			{
				resultNode.removeChild(resultTextNode);
				resultNode.appendChild(privatePropertyDoc.createTextNode(value));
			}
		}
	}

	/**
	 * Gets the names of Elements carried in a Xpath string.
	 */
	private Vector getAttributeNamesInXpathString(String xpathStr)
	{
		Vector result = new Vector();
		
		StringTokenizer p = new StringTokenizer(xpathStr, "/");
		
		final StringTokenizer tokenizer = new StringTokenizer (xpathStr, "/");
		while (tokenizer.hasMoreTokens())
		{
			final String strValue = tokenizer.nextToken();
			result.addElement(strValue);
		}

		if (debug)
		{
			for (int i=0; i<result.size(); i++)
			{
				String ee = (String) result.elementAt(i);
				System.out.println("The "+i+"th element is "+ee+".");
			}
		}

		return result;
	}


//
//	Integer Attribute Accessors
//

	 /**
	  * Gets the value of an integer attribute. A caller provided
	  * default value is returned if no value for this attribute was
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an integer.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getIntProperty that does not accept a caller
	  * provided default value.
	  */
	 public int getIntProperty (final String attribute, final int defaultValue)
	 {
		 return getIntProperty(attribute);
	 }

	 /**
	  * Gets the value of an integer attribute. A default value from
	  * the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an integer.
	  * 
	  * @deprecated use getIntAttribute (final XPath attrXpath)
	  */
	 public int getIntProperty (final String attribute)
	 {
		 try
		 {
			 return getIntAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of an integer attribute. A default value from
	  * the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an integer.
	  */
	 public int getIntAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseInt (attribute, encodedValue, false, false);
	 }

    
	 /**
	  * Gets the value of a non-negative integer attribute. A caller
	  * provided default value is returned if no value for this
	  * attribute was specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative integer.
	  * @throws HyperCastFatalRuntimeException if the caller provided default value is not non-negative.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getNonNegativeIntProperty that does not accept a
	  * caller provided default value.
	  */
	 public int getNonNegativeIntProperty (final String attribute, final int defaultValue)
	 {
		 return getNonNegativeIntProperty(attribute);
	 }

	 /**
	  * Gets the value of a non-negative integer attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative integer.
	  * 
	  * @deprecated use getNonNegativeIntAttribute (final XPath attrXpath)
	  */
	 public int getNonNegativeIntProperty (final String attribute) 
	 {
		 try
		 {
			 return getNonNegativeIntAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of a non-negative integer attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative integer.
	  */
	 public int getNonNegativeIntAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseInt (attribute, encodedValue, true, false);
	 }


	 /**
	  * Gets the value of a positive integer attribute. A caller
	  * provided default value is returned if no value for this
	  * attribute was specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a postive integer.
	  * @throws HyperCastFatalRuntimeException if the caller provided default value is not positive.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getPositiveIntProperty that does not accept a
	  * caller provided default value.
	  */
	 public int getPositiveIntProperty (final String attribute, final int defaultValue)
	 {
		 return getPositiveIntProperty(attribute);
	 }

	 /**
	  * Gets the value of a positive integer attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a positive integer.
	  * 
	  * @deprecated use getPositiveIntAttribute (final XPath attrXpath)
	  */
	 public int getPositiveIntProperty (final String attribute) 
	 {
		 try
		 {
			 return getPositiveIntAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of a positive integer attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a positive integer.
	  */
	 public int getPositiveIntAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseInt (attribute, encodedValue, true, true);
	 }

	 /*
	  * Unified integer parsing routine.
	  */
	 private int parseInt (final String attribute, final String encodedValue, final boolean mustBeNonNegative, final boolean mustBePositive)
	 {
		 if (! mustBeNonNegative
			 &&
			 mustBePositive)
		 {
			 throw new HyperCastFatalRuntimeException ("Cannot be positive and not non-negative");
		 }

		 try 
		 {
			 final int result = Integer.parseInt (encodedValue);
			 if (mustBeNonNegative && result < 0
				 ||
				 mustBePositive && result <= 0)
			 {
				 final String badTypeString = mustBePositive ? "a non-positive" : "a negative";
				 throw new HyperCastConfigException ("Value \"" + encodedValue + "\" for attribute \"" + attribute + "\" has " + badTypeString + " integer value.");
			 }

			 return result;
		 }
		 catch (final NumberFormatException nfe) 
		 {
			 final String typeString = mustBePositive ? "a positive integer" : (mustBeNonNegative ? "a non-negative integer" : "an integer");
			 throw new HyperCastConfigException ("Value \"" + encodedValue + "\" for attribute \"" + attribute + "\" could not be parsed as " + typeString + " value.", nfe);
		 }
	 }
    
//
//	Short Attribute Accessors (Same code as Integer routines, Java 1.5 generics could remove duplicated code)
//

	 /**
	  * Gets the value of an short attribute. A caller provided
	  * default value is returned if no value for this attribute was
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an short.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getShortProperty that does not accept a caller
	  * provided default value.
	  */
	 public short getShortProperty (final String attribute, final short defaultValue)
	 {
		 return getShortProperty(attribute);
	 }

	 /**
	  * Gets the value of an short attribute. A default value from
	  * the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an short.
	  * 
	  * @deprecated use getShortAttribute (final XPath attrXpath)
	  */
	 public short getShortProperty (final String attribute)
	 {
		 try
		 {
			 return getShortAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of an short attribute. A default value from
	  * the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an short.
	  */
	 public short getShortAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseShort (attribute, encodedValue, false, false);
	 }
    
	 /**
	  * Gets the value of a non-negative short attribute. A caller
	  * provided default value is returned if no value for this
	  * attribute was specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative short.
	  * @throws HyperCastFatalRuntimeException if the caller provided default value is not non-negative.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getNonNegativeShortProperty that does not accept a
	  * caller provided default value.
	  */
	 public short getNonNegativeShortProperty (final String attribute, final short defaultValue)
	 {
		 return getNonNegativeShortProperty(attribute);
	 }

	 /**
	  * Gets the value of a non-negative short attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative short.
	  * 
	  * @deprecated use getNonNegativeShortAttribute (final XPath attrXpath)
	  */
	 public short getNonNegativeShortProperty (final String attribute) 
	 {
		 try
		 {
			 return getNonNegativeShortAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of a non-negative short attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative short.
	  */
	 public short getNonNegativeShortAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseShort (attribute, encodedValue, true, false);
	 }

	 /**
	  * Gets the value of a positive short attribute. A caller
	  * provided default value is returned if no value for this
	  * attribute was specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a postive short.
	  * @throws HyperCastFatalRuntimeException if the caller provided default value is not positive.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getPositiveShortProperty that does not accept a
	  * caller provided default value.
	  */
	 public short getPositiveShortProperty (final String attribute, final short defaultValue)
	 {
		 return getPositiveShortProperty(attribute);
	 }

	 /**
	  * Gets the value of a positive short attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a positive short.
	  * 
	  * @deprecated use getPositiveShortAttribute (final XPath attrXpath)
	  */
	 public short getPositiveShortProperty (final String attribute) 
	 {
		 try
		 {
			 return getPositiveShortAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of a positive short attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a positive short.
	  */
	 public short getPositiveShortAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseShort (attribute, encodedValue, true, true);
	 }

	 /*
	  * Unified short parsing routine.
	  */
	 private short parseShort (final String attribute, final String encodedValue, final boolean mustBeNonNegative, final boolean mustBePositive)
	 {
		 if (! mustBeNonNegative
			 &&
			 mustBePositive)
		 {
			 throw new HyperCastFatalRuntimeException ("Cannot be positive and not non-negative");
		 }

		 try 
		 {
			 final short result = Short.parseShort (encodedValue);
			 if (mustBeNonNegative && result < 0
				 ||
				 mustBePositive && result <= 0)
			 {
				 final String badTypeString = mustBePositive ? "a non-positive" : "a negative";
				 throw new HyperCastConfigException ("Value \"" + encodedValue + "\" for attribute \"" + attribute + "\" has " + badTypeString + " short value.");
			 }

			 return result;
		 }
		 catch (final NumberFormatException nfe) 
		 {
			 final String typeString = mustBePositive ? "a positive short" : (mustBeNonNegative ? "a non-negative short" : "an short");
			 throw new HyperCastConfigException ("Value \"" + encodedValue + "\" for attribute \"" + attribute + "\" could not be parsed as " + typeString + " value.", nfe);
		 }
	 }

//
//	Long Attribute Accessors (Same code as Integer routines, Java 1.5 generics could remove duplicated code)
//

	 /**
	  * Gets the value of an long attribute. A caller provided
	  * default value is returned if no value for this attribute was
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an long.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getLongProperty that does not accept a caller
	  * provided default value.
	  */
	 public long getLongProperty (final String attribute, final long defaultValue)
	 {
		 return getLongProperty(attribute);
	 }

	 /**
	  * Gets the value of an long attribute. A default value from
	  * the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an long.
	  * 
	  * @deprecated use getLongAttribute (final XPath attrXpath)
	  */
	 public long getLongProperty (final String attribute)
	 {
		 try
		 {
			 return getLongAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of an long attribute. A default value from
	  * the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as an long.
	  */
	 public long getLongAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseLong (attribute, encodedValue, false, false);
	 }

    
	 /**
	  * Gets the value of a non-negative long attribute. A caller
	  * provided default value is returned if no value for this
	  * attribute was specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative long.
	  * @throws HyperCastFatalRuntimeException if the caller provided default value is not non-negative.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getNonNegativeLongProperty that does not accept a
	  * caller provided default value.
	  */
	 public long getNonNegativeLongProperty (final String attribute, final long defaultValue)
	 {
		 return getNonNegativeLongProperty(attribute);
	 }

	 /**
	  * Gets the value of a non-negative long attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative long.
	  * 
	  * @deprecated use getNonNegativeLongAttribute (final XPath attrXpath)
	  */
	 public long getNonNegativeLongProperty (final String attribute) 
	 {
		 try
		 {
			 return getNonNegativeLongAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of a non-negative long attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a non-negative long.
	  */
	 public long getNonNegativeLongAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseLong (attribute, encodedValue, true, false);
	 }

	 /**
	  * Gets the value of a positive long attribute. A caller
	  * provided default value is returned if no value for this
	  * attribute was specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a postive long.
	  * @throws HyperCastFatalRuntimeException if the caller provided default value is not positive.
	  *
	  * @deprecated Default values are stored in the HyperCast XML
	  * Schema file which acts as a centralized repository for default
	  * values. By using the defaults stored in the XML Schema file,
	  * constants or "magic numbers" are not spread throughout the code.
	  * Use a version of getPositiveLongProperty that does not accept a
	  * caller provided default value.
	  */
	 public long getPositiveLongProperty (final String attribute, final long defaultValue)
	 {
		 return getPositiveLongProperty(attribute);
	 }

	 /**
	  * Gets the value of a positive long attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a positive long.
	  * 
	  * @deprecated use getPositiveLongAttribute (final XPath attrXpath)
	  */
	 public long getPositiveLongProperty (final String attribute) 
	 {
		 try
		 {
			 return getPositiveLongAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of a positive long attribute. A default value
	  * from the XML schema file is returned if this attribute was not
	  * specified in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a positive long.
	  */
	 public long getPositiveLongAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);

		 return parseLong (attribute, encodedValue, true, true);
	 }

	 /*
	  * Unified long parsing routine.
	  */
	 private long parseLong (final String attribute, final String encodedValue, final boolean mustBeNonNegative, final boolean mustBePositive)
	 {
		 if (! mustBeNonNegative
			 &&
			 mustBePositive)
		 {
			 throw new HyperCastFatalRuntimeException ("Cannot be positive and not non-negative");
		 }

		 try 
		 {
			 final long result = Long.parseLong (encodedValue);
			 if (mustBeNonNegative && result < 0
				 ||
				 mustBePositive && result <= 0)
			 {
				 final String badTypeString = mustBePositive ? "a non-positive" : "a negative";
				 throw new HyperCastConfigException ("Value \"" + encodedValue + "\" for attribute \"" + attribute + "\" has " + badTypeString + " long value.");
			 }

			 return result;
		 }
		 catch (final NumberFormatException nfe) 
		 {
			 final String typeString = mustBePositive ? "a positive long" : (mustBeNonNegative ? "a non-negative long" : "an long");
			 throw new HyperCastConfigException ("Value \"" + encodedValue + "\" for attribute \"" + attribute + "\" could not be parsed as " + typeString + " value.", nfe);
		 }
	 }

//
//	Boolean Attribute Accessors
//

	 /**
	  * Gets the value of a boolean attribute. A default value from the
	  * XML schema file is returned if this attribute was not specified
	  * in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a boolean.
	  *
	  * Note that <code>Boolean.valueOf</code> is not used here because
	  * it always returns either <code>true</code> or
	  * <code>false</code> regardless of the value of the string, for
	  * instance "foo" would result in <code>false</code>.
	  * 
	  * @deprecated use getBooleanAttribute (final XPath attrXpath)
	  */
	 public boolean getBooleanProperty (final String attribute)
	 {
		 try
		 {
			 return getBooleanAttribute (new XPath(attribute, null, null, XPath.MATCH));
		 }
		 catch (TransformerException te)
		 {
			 throw new HyperCastConfigException ("\""+attribute+"\" is not a valid XPath string.");
		 }
	 }

	 /**
	  * Gets the value of a boolean attribute. A default value from the
	  * XML schema file is returned if this attribute was not specified
	  * in the configuration file.
	  *
	  * @throws HyperCastConfigException if the configured value cannot be parsed as a boolean.
	  *
	  * Note that <code>Boolean.valueOf</code> is not used here because
	  * it always returns either <code>true</code> or
	  * <code>false</code> regardless of the value of the string, for
	  * instance "foo" would result in <code>false</code>.
	  */
	 public boolean getBooleanAttribute (final XPath attrXpath)
	 {
		 final String attribute = attrXpath.getPatternString();
		 final String encodedValue = this.getTextAttribute(attrXpath);


		 if (encodedValue == null)
		 {
			 final NullPointerException npe = new NullPointerException();
			 throw new HyperCastConfigException (npe);
		 }
		 else if (encodedValue.equalsIgnoreCase ("true"))
		 {
			 return true;
		 }
		 else if (encodedValue.equalsIgnoreCase ("false"))
		 {
			 return false;
		 }
		 else
		 {
			 final IllegalArgumentException iae = new IllegalArgumentException ("Value \"" + encodedValue + "\" for attribute \"" + attribute + "\" did not parse as a boolean.");
			 throw new HyperCastConfigException (iae);
		 }
	 }

    
	 /*
	  * @return <code>true</code> if the given string is "yes" (case
	  * insensitive).  <code>false</code> if the given string is "no"
	  * (case insensitive).
	  *
	  * @throws HyperCastConfigException if the string parameter is a
	  * <code>null</code> reference or if the string is not either
	  * "yes" or "no" (case insensitive).
	  */
	 public boolean booleanFromYesNoString (final String yesNoString)
	 {
		 if (yesNoString == null)
		 {
			 final NullPointerException npe = new NullPointerException();
			 throw new HyperCastConfigException (npe);
		 }
		 else if (yesNoString.equalsIgnoreCase ("yes"))
		 {
			 return true;
		 }
		 else if (yesNoString.equalsIgnoreCase ("no"))
		 {
			 return false;
		 }
		 else
		 {
			 final IllegalArgumentException iae = new IllegalArgumentException (yesNoString);
			 throw new HyperCastConfigException (iae);
		 }
	 }

    
	/**
	 * @return a URL property; a default value is returned if this property was not specified in the configuration file.
	 */
	public URL getURLAttribute (final XPath xpath)
	{
		String urlString = this.getTextAttribute(xpath);

		if (debug)
			System.out.println("urlString="+urlString);
		
		try
		{
			URL result = new URL(urlString);
			return result;
		}
		catch (MalformedURLException e)
		{
			throw new HyperCastConfigException ("Attribute \"" + xpath.getPatternString() + "\" didn't have valid URL string: "+e);
		}
	}


	/**
	 *  Creates a random overlay ID.
	 * 
	 * @deprecated
	 */
	private String createOverlayID() 
	{
		InetAddress localHost = null;
		try {
			localHost = InetAddress.getLocalHost();
		}
		catch (UnknownHostException uhe) 
		{
			throw new HyperCastFatalRuntimeException ("InetAddress.getLocalHost() threw UnknownHostException.");
		}

		long t = System.currentTimeMillis();
    
		return localHost.getHostAddress() + ":0" + "TS" + t;
	}

	/**
	 *  Creates a random overlay ID and sets it into the property Document.
	 */
	public void createandsetOverlayId() 
	{
		InetAddress localHost = null;
		try {
			localHost = InetAddress.getLocalHost();
		}
		catch (UnknownHostException uhe) 
		{
			throw new HyperCastFatalRuntimeException ("InetAddress.getLocalHost() threw UnknownHostException.");
		}

		long t = System.currentTimeMillis();
    
		String overlayID = localHost.getHostAddress() + ":0" + "TS" + t;
		
		try
		{
			this.setTextAttribute(new XPath(CONFIG_ATTRIBUTE_OVERLAY_ID, null, null, XPath.MATCH), overlayID);
		}
		catch (TransformerException te)
		{
			throw new HyperCastConfigException ("\""+CONFIG_ATTRIBUTE_OVERLAY_ID+"\" is not a valid XPath string.");
		}
		
	}


/** Open the log and err files.  If filename is "stdout" or "stderr" the 
 * error or log messages are redirected to System.out or System.err.
 */
    private void openErrAndLogFiles()
    {
        //open logFileName
		try
		{
			logFileName = this.getTextAttribute(new XPath(LOGFILE_ATTRIBUTE_NAME, null, null, XPath.MATCH));
		}
		catch (TransformerException te)
		{
			throw new HyperCastConfigException ("\""+LOGFILE_ATTRIBUTE_NAME+"\" is not a valid XPath string.");
		}
        
        if ((logFileName == null) ||(logFileName.equals(NO_FILE)))
        {
            log = new PrintWriter(new NullWriter());
        }
        else if (logFileName.equals("stdout")) 
        {
            log = new PrintWriter(System.out, true);
        }
        else if (logFileName.equals("stderr")) 
        {
            log = new PrintWriter(System.err, true);
        }
        else 
        {
            try {
                log = new PrintWriter(new FileWriter(logFileName), true);
            }
            catch (IOException e) {
                throw new HyperCastFatalRuntimeException ("Unable to open log file \"" + logFileName +"\"", e);
            }
        }
			
        //get ErrorFile
		try
		{
			errFileName = this.getTextAttribute(new XPath(ERRFILE_ATTRIBUTE_NAME, null, null, XPath.MATCH));;
		}
		catch (TransformerException te)
		{
			throw new HyperCastConfigException ("\""+ERRFILE_ATTRIBUTE_NAME+"\" is not a valid XPath string.");
		}

        if ((errFileName == null) || (errFileName.equals(NO_FILE)))
        {
            err = new PrintWriter(new NullWriter());
            seriousErr = new PrintWriter(System.out, true);
        }
        else if (errFileName.equals("stdout")) 
        {
            err = new PrintWriter(System.out, true);
            seriousErr = err;
        }
        else if (errFileName.equals("stderr")) 
        {
            err = new PrintWriter(System.err, true);
            seriousErr = err;
        }
        else if (!errFileName.equals(NO_FILE)){
            try {
                err = new PrintWriter(new FileWriter(errFileName), true);
                seriousErr = err;
            }
            catch (IOException e) {
                throw new HyperCastFatalRuntimeException ("Unable to open error file \"" + errFileName +"\"", e);
            }
        }
    }




	/** 
	 * Obtain the list of attribute name contained in the value of given attribute.
	 * 
	 * @param propertydoc	Document instance which represents the overlay properties.
	 * @param attribute		string representing the attribute whose value contains a list
	 * of attribute names.
	 * 
	 * @return a vector of the list of attribute names contained in the value of given attribute.
	 */
	private Vector getListOfAttributeNames(Document propertydoc, String attribute)
	{
		Vector p = new Vector();

		Node localAttrNode = singleNodeXpathQuery(propertydoc, attribute);

		if (localAttrNode == null)
		{
			//throw new IllegalStateException("The LocalAttribute property is missing!");
			return p;
		}

		String localAttrStr = null;

		NodeList localAttrContentNodes = localAttrNode.getChildNodes();

		if (localAttrContentNodes.getLength() == 0)
		{
			return p;
		}
		else if ((localAttrContentNodes.getLength() > 1) || (localAttrContentNodes.item(0).getNodeType() != Node.TEXT_NODE))
		{
			throw new IllegalStateException("The local attributes are specified incorrectly!");
		}

		localAttrStr = localAttrNode.getChildNodes().item(0).getNodeValue();


		if (localAttrStr == null)
			localAttrStr="";

		int start = 0;
		int end = localAttrStr.indexOf(',', start);
		if (end < 0)
		{
			end = localAttrStr.length();
		}
        
		while (end >= 0 && end-start > 0) 
		{
			String attr = localAttrStr.substring(start, end);

			p.addElement(attr.trim());    

			if (end == localAttrStr.length())
			{
				break;
			}

			start = end+1;
			end = localAttrStr.indexOf(',', start);
			if (end < 0)
				end = localAttrStr.length();
		}

		return p;
	}


			
/** Returns the hash calculated from overlay hash attributes.  This is a 
 * used to determine if two messages are from the same Overlay. 
 */
public int getOverlayHash() 
    {
        return overlayHash;
    }

/** Creates hash from overlay hash attributes.  This is a used to
 * determine if two messages are from the same Overlay. 
*/
public int setOverlayHash() 
	{
		try
		{
			Vector hashAttributeVec = getListOfAttributeNames(this.propertyDoc, HyperCastConfig.OVERLAYHASH_ATTRIBUTE_NAME);
	
			String s = "";
	
			for (int i=0; i<hashAttributeVec.size(); i++)
			{
				String value = getTextAttribute(new XPath((String) hashAttributeVec.elementAt(i), null, null, XPath.MATCH));
				
				if (value != null)
				{
					if (s.equals(""))
					{
						s = s + (String) hashAttributeVec.elementAt(i) + "=" + value;
					}
					else
					{
						s = s + "&" + (String) hashAttributeVec.elementAt(i) + "=" + value;
					}
				}
			}
	
			return hash( s.getBytes());
		}
		catch (TransformerException e)
		{
			throw new IllegalStateException("The overlay hash attributes are specified incorrectly!");
		}
	}

	/** Creates hash from overlay hash attributes.  This is a used to
	 * determine if two messages are from the same Overlay. 
	 * 
	 * @deprecated use setOverlayHash().
	 */
	private int generateOverlayHash() 
    {
    	return setOverlayHash();
    }

	/** Generates a hash of a byte array for the generateOverlayHash
	 * function.  I could just use Java's hashCode for a string object,
	 * but I wanted something that would be available to any other
	 * implementation language. 
	 */
	public static int hash(byte[] array)
    {
        int h = 0;
        for (int i = 0; i < array.length; i++) 
        {
            byte upperByte = (byte) ((h >>> 24) & 0xFF);
            int leftShiftValue = ((upperByte ^ array[i]) & 0x07) + 1;
            h = ((h << leftShiftValue) ^ ((upperByte ^ array[i]) & 0xFF));
        }
        
        return h;
    }

	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsProcessor instance is created which contains all statistics
	 * , with each represented by a StatsElement instance, supported in this object.
	 */
	private void InitStatisticsStructure()
	{
		/*
		 * Get the statistics name defined in the configuration file for this I_Stats instance.
		 */
		statisticsName = this.getTextAttribute (XmlUtil.createXPath(CONFIGOBJ_ATTRIBUTE_NAME+"/StatName"));

		/*
		 * Initial StatsProcessor instance.
		 */
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
		
		
		statsPro.addStatsElement("LogFileName"/* name */, new logFile()/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

		statsPro.addStatsElement("ErrorFileName", new errorFile(), 1, 1);

		statsPro.addStatsElement("Public", new PropertyStatsObj(this), 1, 1);
	}

	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}

	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("HyperCastConfig:getStats: xpath string is "+xpath.getPatternString());
		//System.out.println("HyperCastConfig:getStats: call statsPro.getStatsResult(doc, xpath)");
		return statsPro.getStatsResult(doc, xpath);
	}
	
	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		//System.out.println("HyperCastConfig:setStats: xpath string is "+xpath.getPatternString());
		//System.out.println("HyperCastConfig:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");
		return statsPro.setStatsResult(doc, xpath, newValue);
	}
	
	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("HyperCastConfig:getReadSchema: xpath string is "+xpath.getPatternString());
		//System.out.println("HyperCastConfig:getReadSchema: call statsPro.getReadSchemaResult(doc, xpath)");
		return statsPro.getReadSchemaResult(doc, xpath);
	}
	
	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
	

	/**
	 * This class manages simple statistic "LogFileName".
	 */
	class logFile extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return logFileName;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			if (!logFileName.equals(NO_FILE) && !logFileName.equals("stdout") && !logFileName.equals("stderr")) 
			{
				log.close();
			}

			logFileName = newValue;
			if (logFileName.equals(NO_FILE)) 
			{
				log = new PrintWriter(new NullWriter());
			}
			else if (logFileName.equals("stdout")) 
			{
				log = new PrintWriter(System.out, true);
			}
			else if (logFileName.equals("stderr")) 
			{
				log = new PrintWriter(System.err, true);
			}
			else 
			{
				try {
					log = new PrintWriter(new FileWriter(logFileName), true);
				}
				catch (IOException e) {
					logFileName = NO_FILE;
					log = new PrintWriter(new NullWriter());
					throw new HyperCastStatsException("Exception occured when opening file \"" + newValue + "\" : "+e.getMessage(), HyperCastStatsException.InappropriateValue);
				}
			}
			
			return newValue;
		}
		
		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		

		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}

	}


	/**
	 * This class manages simple statistic "ErrorFileName".
	 */
	class errorFile extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return errFileName;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			if (!errFileName.equals(NO_FILE) && !errFileName.equals("stdout") && !errFileName.equals("stderr"))
			{
				err.close();
			}
            

			errFileName = newValue;
			if (errFileName.equals(NO_FILE))
			{
				err = new PrintWriter(new NullWriter());
				seriousErr = new PrintWriter(System.out, true);
			}
			else if (errFileName.equals("stdout")) 
			{
				err = new PrintWriter(System.out, true);
				seriousErr = err;
			}
			else if (errFileName.equals("stderr")) 
			{
				err = new PrintWriter(System.err, true);
				seriousErr = err;
			}
			else 
			{
				try {
					err = new PrintWriter(new FileWriter(errFileName), true);
					seriousErr = err;
				}
				catch (IOException e) {
					errFileName = NO_FILE;
					err = new PrintWriter(new NullWriter());
					seriousErr = new PrintWriter(System.out, true);
					throw new HyperCastStatsException("Exception occured when opening file \"" + newValue + "\" : "+e.getMessage(), HyperCastStatsException.InappropriateValue);
				}
			}
			
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		

		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}

	}

}


/**
 * This class processes the statistics stored in the overlay property database.
 */

class PropertyStatsObj implements I_Stats
{
	/**
	 * Overlay configuration object.
	 */
	HyperCastConfig hcconfig;

	/**
	 * The statistics name for the object (as an I_Stats instance) which implements this class.
	 * 
	 * @see I_Stats interface.
	 */
	private String statisticsName = new String("Public");
	
	/**
	 * Constructor.
	 */
	public PropertyStatsObj(HyperCastConfig hcc)
	{
		hcconfig = hcc;
	}

	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		/*
		 * The setStatisticsName of a PropertyStatsObj instance is fixed to "Public" which
		 * is the top layer Element name of a configuration document.
		 */
		if (!name.equals(statisticsName))
		{
			hcconfig.err.println("Try to set a invalid statistic name to the configuration object (as an I_Stats object).");
		}
	}

	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("PropertyStatsObj:getStats: xpath string is "+xpath.getPatternString());
		//System.out.println("PropertyStatsObj:getStats: call statsPro.getStatsResult(doc, xpath)");

		Document propertydoc = hcconfig.getPublicConfiguration();

		String xpathStr = xpath.getPatternString();

		Vector matchedNodeVec = HyperCastConfig.XpathQuery(propertydoc, xpathStr);

		if (matchedNodeVec.size() == 0)
		{
			return new Element[0];
		}

		Element[] valueElement = new Element[matchedNodeVec.size()];

		for (int i = 0; i < matchedNodeVec.size(); i++)
		{ 
			Element matchNode = (Element) matchedNodeVec.elementAt (i);
			valueElement[i] = (Element) doc.importNode(matchNode, true);
		}

		return valueElement;
	}
	
	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		//System.out.println("PropertyStatsObj:setStats: xpath string is "+xpath.getPatternString());
		//System.out.println("PropertyStatsObj:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");

		//throw new HyperCastStatsException(xpath.getPatternString(), HyperCastStatsException.WriteToAReadOnlyValue);

		Document propertydoc = hcconfig.getPublicConfiguration();

		String xpathStr = xpath.getPatternString();

		Vector matchedNodeVec = HyperCastConfig.XpathQuery(propertydoc, xpathStr);

		String newValueName = newValue.getNodeName();

		if (!StatsProcessor.allElementsInVectorMatchGivenName(matchedNodeVec, newValueName))
			
		{
			return new Element[0];
		}

		if (matchedNodeVec.size() == 0)
		{
			//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.StatisticNotFound);
			return new Element[0];
		}

		Element[] newValueElement = new Element[matchedNodeVec.size()];

		for (int j = 0; j < matchedNodeVec.size(); j++)
		{
			Element oldElement = (Element) matchedNodeVec.elementAt(j);
				
			Node parent = oldElement.getParentNode();
			
			if (parent == null)
			{
				//throw new HyperCastStatsException(xpathStr, HyperCastStatsException.IncorrectSetValue);
				return new Element[0];
			}

			parent.removeChild(oldElement);
			parent.appendChild(propertydoc.importNode(newValue, true));
			
			/*
			 * Create a new Element for each setting action.
			 */
			newValueElement[j] = (Element) doc.importNode(newValue, true);
		}
		
		return newValueElement;
	}
	
	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:complexType"/* type */, 
		//									null/* restriction base */, null/* pattern value */);
		
		final String xpathStr = xpath.getPatternString();
		if (!xpathStr.equals("/Public"))
		{
			return new Element[0];
		}
		else
		{
			try
			{
				Element[] eleArray = new Element[1];
				eleArray[0] = (new CurrentConfigurationSchemaGenerator()).cerateSchemaTree(doc, hcconfig.getPublicConfiguration());

				return eleArray;
			}
			catch (Exception e)
			{
				return new Element[0];
			}
		}
	}
	
	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{		//return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:complexType"/* type */, 
		//									null/* restriction base */, null/* pattern value */);
		
		return getReadSchema(doc, xpath);
	}

}

