/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.security.cert.Certificate;
import java.security.PrivateKey;

import java.net.*;
import java.io.*;

/**
 * This is the interface for a OL_socket. It defines all visible functions
 * in the OL_Socket that can be used by an application to send and receive 
 * data on the overlay. 
 *  
 * @author HyperCast Team
 * @author Michael Nahas
 * @author Jianping Wang
 * @author Guimin Zhang
 * @author Greg Mattes
 * @version     3.0, Spring 2005
 */

public interface I_OverlaySocket extends I_Stats
{
    /** Join the overlay multicast group.  NOTE: It may take a while
     * for the overlay to stabilize; right after joining some sends
     * may not reach all members of the group.  
     */
    public void joinOverlay();

    /**
     * Leave the overlay multicast group. 
     */
    public void leaveOverlay();

    /** Create a message for sending on the overlay socket. */
    public I_OverlayMessage createMessage(byte[] payload);

    /** Create a message with specified service number. */
    public I_OverlayMessage createMessage(byte[] payload, short servicenumber);

    /** Creates a logical address from a byte array.  This enables applications to
     * send logical addresses in messages.   The number of bytes read is equal to 
     * I_OverlaySocket.getLogicalAddress().getSize().
     */
    public I_LogicalAddress createLogicalAddress(byte[] laddr, int offset);

    /** Creates a logical address object from a String.
     */
    public I_LogicalAddress createLogicalAddress(String laStr);

    /** Sets the logical address to specified one.
     */
    public void setLogicalAddress(I_LogicalAddress la);

    /** Returns the address of this socket on the overlay. */
    public I_LogicalAddress getLogicalAddress();

    /** Returns a byte array containing a unique identifier of this socket. */
    public byte[] getUniqueIdentifier();

    /**
     * <p>Binds the specified name to the logical address of this
     * overlay socket.  This operation might cause the new binding to
     * be "pushed" to other peers in the network, depending upon the
     * policy configuration of the socket. The global uniqueness of
     * the names of overlay sockets is neither enforced nor
     * required. The set of all bindings with a particular name
     * defines a group.</p>
     *
     * @param name The string that will be set as the name of this overlay socket.
     *
     * @throws HyperCastWarningRuntimeException if the argument is <code>null</code>
     */
    public void setName (String name);

    /**
     * <p>Assigns the specified name to the specified logical
     * address. The logical address can be any address in the current
     * logical address space. This operation allows bindings to be
     * created for the logical addresses of sockets other than the
     * local socket.
     *
     * @param name The string that will be used as the name of a new binding.
     * @param logicalAddress The logical address that will be used as the logical address of a new binding.
     */
    public void setName (String name, I_LogicalAddress logicalAddress);

    /**
     * <p>This method is like setName(String) but the subject common
     * name of the given certificate is used as the name string
     * instead of a Java String object. Also, a private key,
     * cooresponding to the public key in the certificate, is given so
     * that naming message that are sent to peers can be
     * cryptographically signed.</p>
     */
    public void setName (Certificate certificate, PrivateKey privateKey);

    /**
     * <p>Assigns the specified name to the specified logical address
     * using the specified private key (that matches the public key in
     * the specified certificate) to create a digital signature for
     * the new binding. The subject common name of the specified
     * certificate is used as the signer name of the binding. The
     * logical address can be any address in the current logical
     * address space. This operation allows bindings to be created for
     * the logical addresses of sockets other than the local socket.
     *
     * @param name The string that will be used as the name of a new binding.
     * @param logicalAddress The logical address that will be used as the logical address of a new binding.
     * @param certificate Certificate that contains the name that will be the signer of the new binding.
     * @param privateKey Key used to create a digital signature for a new binding
     */
    public void setName (Certificate certificate, PrivateKey privateKey, String name, I_LogicalAddress logicalAddress);

    /**
     * <p>Unassigns the given name from this overlay socket.  This
     * operation might cause an invalidation operation for the name to
     * be "pushed" to other nodes in the network, depending upon the
     * policy configuration of the socket.</p>
     *
     * @param name The string that will be unset as the name of this overlay socket.
     *
     * @throws HyperCastWarningRuntimeException if the argument is <code>null</code>
     */
    public void unsetName (String name);

    /**
     * <p>This method is like unsetName(String) but uses a certificate
     * instead of a String, it is the inverse of
     * setName(Certificate,PrivateKey).  </p>
     */
    public void unsetName (Certificate certificate, PrivateKey privateKey);

    /**
     * <p>Queries the name of the local overlay socket.</p>
     *
     * @return The name of this overlay socket, never null since the lookup is local.
     */
    public String [] getNames();

    /**
     * @see #getNames(I_LogicalAddress, boolean, long, boolean, boolean, int, int, short, short)
     * @param peer The logical address of the node for which a name is requested.
     */
    public String [] getNames (final I_LogicalAddress peer);

    /**
     * Queries the name of the overlay socket with logical address
     * <code>peer</code>. The name returned can be the name of the
     * local socket, if the logical address is the logical address of
     * the local socket. This implements a "reverse lookup" function
     * for name resolution. If the value returned is a zero length
     * array it does not imply that no such logical address exists on
     * the network, only that the request could not be satisfied,
     * i.e. this call does not guarentee that it will find a name if
     * one exists for a given logical address.</p>
     *
     * @param peer The logical address of the node for which a name is requested.
     * @param requireTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requireSerialNumberMatch Each query has a unique serial number assigned to it. If this parameter is true then
     * the serial number of a valid response must match the serial number of a query.
     * @param requireAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     * @param maxHopCount The maximum number of hops that a query can make before it is not propagated further.
     *
     * @return A array of the names of the overlay sockets having the
     * given logical address
     *
     * @throws HyperCastWarningRuntimeException if maxResponses is <code>Naming.NO_RESPONSE_LIMIT</code> and timeout <code>Naming.WAIT_INDEFINITELY</code>
     */
    public String [] getNames (final I_LogicalAddress peer,
                               final boolean          requireTrusted,
                               final long             timeout,
                               final boolean          requireSerialNumberMatch,
                               final boolean          requireAuthoritative,
                               final int              maxAge,
                               final int              maxResponses,
                               final short            minHopCount,
                               final short            maxHopCount);

    /**
     * <p>Same as <code>getNames(I_LogicalAddress)</code> but returns
     * immediately. Application must handle related
     * <code>NAMING_EVENT</code> notifications containing responses to
     * this call.</p>
     *
     * @see #getNames(I_LogicalAddress, boolean, long, boolean, boolean, int, int, short, short)
     */
    public int getNamesNonBlocking (I_LogicalAddress address);

    /**
     * @see #getNamesNonBlocking(I_LogicalAddress)
     *
     * @param peer The logical address of the node for which a name is requested.
     * @param requireTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requireSerialNumberMatch Each query has a unique serial number assigned to it. If this parameter is true then
     * the serial number of a valid response must match the serial number of a query.
     * @param requireAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     * @param maxHopCount The maximum number of hops that a query can make before it is not propagated further.
     *
     * @return serial number of query
     */
    public int getNamesNonBlocking (final I_LogicalAddress peer,
                                    final boolean          requireTrusted,
                                    final long             timeout,
                                    final boolean          requireSerialNumberMatch,
                                    final boolean          requireAuthoritative,
                                    final int              maxAge,
                                    final int              maxResponses,
                                    final short            minHopCount,
                                    final short            maxHopCount);

    /**
     * @see I_LogicalAddress [] getLogicalAddressByName(String,boolean,long,boolean,int,int,short)
     * @param name The name of a socket for which the logical address is requested.
     */
    public I_LogicalAddress [] getLogicalAddressByName (final String name);

    /**
     * <p>Queries the logical address of the overlay socket having the
     * specified name. If no such socket can be found, a zero length
     * array is returned. A result of a zero length array does not
     * imply that no such socket exists, only that one could not be
     * found, i.e. this call does not guarentee that it will find a
     * logical address if one exists for a given name.</p>
     *
     * @param name The name for which a  logical address lookup is being performed
     * @param requireTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requireSerialNumberMatch Each query has a unique serial number assigned to it. If this parameter is true then
     * the serial number of a valid response must match the serial number of a query.
     * @param requireAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     * @param maxHopCount The maximum number of hops that a query can make before it is not propagated further.
     *
     * @return A array of the logical addresses of the sockets with the given name.
     *
     * @throws HyperCastWarningRuntimeException if maxResponses is <code>Naming.NO_RESPONSE_LIMIT</code> and timeout <code>Naming.WAIT_INDEFINITELY</code>
     */
    public I_LogicalAddress [] getLogicalAddressByName (final String  name,
                                                        final boolean requireTrusted,
                                                        final long    timeout,
                                                        final boolean requireSerialNumberMatch,
                                                        final boolean requireAuthoritative,
                                                        final int     maxAge,
                                                        final int     maxResponses,
                                                        final short   minHopCount,
                                                        final short   maxHopCount);


    /**
     * <p>Same as <code>getLogicalAddress(String)</code> but returns
     * immediately. Application must handle related
     * <code>NAMING_EVENT</code> notifications containing responses to
     * this call.</p>
     *
     * @see #getLogicalAddressByName(String, boolean, long, boolean, boolean, int, int, short, short)
     */
    public int getLogicalAddressByNameNonBlocking (String name);

    /**
     * @see #getLogicalAddressByNameNonBlocking(String)
     *
     * @param name The name for which a  logical address lookup is being performed
     * @param requireTrusted If true only trusted bindings will be accepted
     * @param timeout Time in milliseconds before this query will be ended. Can be Naming.WAIT_INDEFINITELY.
     * @param requireSerialNumberMatch Each query has a unique serial number assigned to it. If this parameter is true then
     * the serial number of a valid response must match the serial number of a query.
     * @param requireAuthoritative If true only bindings received from the authoritative peer will be accepted
     * @param maxAge The maximum age in milliseconds of an acceptable
     * query. Calculated from the time of the query for already cached
     * bindings; calculated from time of receipt for bindings received
     * after the query is made.
     * @param maxResponses The query will end when this number of responses are received. Can be Naming.NO_RESPONSE_LIMIT.
     * @param minHopCount The minimum number of hops that a query must make before it is not propagated further.
     * @param maxHopCount The maximum number of hops that a query can make before it is not propagated further.
     *
     * @return serial number of query
     */
    public int getLogicalAddressByNameNonBlocking (final String  name,
                                                   final boolean requireTrusted,
                                                   final long    timeout,
                                                   final boolean requireSerialNumberMatch,
                                                   final boolean requireAuthoritative,
                                                   final int     maxAge,
                                                   final int     maxResponses,
                                                   final short   minHopCount,
                                                   final short   maxHopCount);


    /**
     * Installs a trust anchor for use in building chains of
     * trust. Trust is established for naming bindings by verifying
     * the signature of the binding using the certificate that created
     * the signature. That certificate becomes trusted when a chain of
     * trust can be established from it to an installed trust anchor.
     */
    public void installTrustedNamingCertificate (final Certificate certificate);

    /** Returns the addresspair of the next hop for a message routed by this socket
     * towards the root.  The next hop would be the node's parent in the spanning tree
     * rooted at <code>root</code>.
     * @throws IllegalStateException if socket has not joined group.
     */
    public I_LogicalAddress getParent(I_LogicalAddress root);

    /** Returns the socket's children's physical/logical address pairs, with respect to
     * the spanning tree rooted at <code>root</code> in the overlay.
     * @throws IllegalStateException if socket has not joined group.
     */
    public I_LogicalAddress[] getChildren(I_LogicalAddress root);

    /** Returns the socket's neighbors' physical/logical address pairs.
     * @throws IllegalStateException if socket has not joined group.
     */
    public I_LogicalAddress[] getNeighbors();

    public static final short HOP_LIMIT_MAXIMUM = (short) 0xFFFF;

    /**
     * Sets the default hop limit for all future messages created
     * by <code>createMessage</code>.  It can be overriden on a
     * per-message basis by <code>I_OverlayMessage.setHopLimit</code>.
     *
     * @see I_OverlayMessage#setHopLimit(short) for an explanation of the interpretation of the field value.
     */
    public void setHopLimit (short hopLimit);

    /**
     * Gets the default hop limit for all future messages created
     * by <code>createMessage</code>.
     */
    public short getHopLimit();

    /**
     * Sends the message to all members of the overlay topology.
     * (Including self!)
     */
    public void sendToAll(I_OverlayMessage m);

    /**
     * Sends the message one hop to all the children of this socket
     * in the spanning tree rooted at <code>root</code>.  Also
     * see <code>getChildren</code>.
     */
    public void sendToChildren(I_OverlayMessage m, I_LogicalAddress root);

    /** Sends the message one hop to all neighbors of this socket in the overlay. */
    public void sendToAllNeighbors(I_OverlayMessage m);

    /**
     * Sends this message one hop to the parent of this socket in
     * the spanning tree rooted at <code>root</code>.
     */
    public void sendToParent(I_OverlayMessage m, I_LogicalAddress root);

    /**
     * Sends the message unicast over the overlay to the destination.
     */
    public void sendToNode(I_OverlayMessage m, I_LogicalAddress destla);

    /**
     * Sends the message to all nodes using a "flood" routing
     * mechanism.  Some nodes may receive the message multiple times.
     */
    public void sendFlood(I_OverlayMessage m);

    /**
     * Receive a message.  This can only be used if no object
     * implementing I_ReceiveCallback was passed to the constructor 
     * of the socket.  (e.g., <code> config.createSocket(null)</code>).
     * The thread doing this call can be interrupted.
     * @exception throws IllegalStateException if a callback has been specified.
     */
    public I_OverlayMessage receive();

    /**
     * Set the timeout for any thread calling receive().  See
     * java.net.DatagramSocket.setSoTimeout().
     */
    public void setSoTimeout(int timeout);

    /**
     * Get the timeout for any thread calling receive().  See
     * java.net.DatagramSocket.getSoTimeout().
     */
    public int getSoTimeout();

    public StreamManager getStreamManager();
}
