/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.XPath;

/**
 * This is the interface implemented by objects that have statistics
 * that are available to applications. In particular, this is designed
 * to be an external interface. Hypercast uses the I_StatsInternal
 * interface internally.
 *
 * @author HyperCast Team
 * @author Greg Mattes
 * @author Mike Nahas
 * @author Dongwen Wang
 * @author Neelima Putrevu
 * @version 3.0 November 2004
 */
public interface I_Stats
{
    /**
     * Gets statistics information from an object which implements
     * this interface. Statistics are arranged in a tree
     * structure. This method has the ability to query an arbitrary
     * subtree.
     *
     * @param xpath Specifies the statistic subtree to query.
     * 
     * @param doc The document is used as a factory to create XML
     * objects such as Nodes and Elements as needed. TODO: describe
     * how the XML document in the doc is used.
     *
     * @throws HyperCastStatsException If the xpath does not specify a valid subtree.
     */
    public Element[] getStats (final Document doc, final XPath xpath)
        throws HyperCastStatsException;

    /**
     * Sets statistics information for an object which implements
     * this interface. Statistics are arranged in a tree
     * structure. This method has the ability to set an arbitrary
     * subtree of statistics.
     *
     * @param xpath Specifies the statistic subtree to use when updating values.
     * 
     * @param doc The document is used as a factory to create XML
     * objects such as Nodes and Elements as needed. TODO: describe
     * how the XML document in the doc is used.
     *
     * @throws HyperCastStatsException If the xpath does not specify a
     * valid subtree or if any part of the the subtree is read-only,
     * or if the <code>newValue</code> does not conform with the
     * expected format.
     */
    public Element[] setStats (final Document doc, final XPath xpath, final Element newValue)
        throws HyperCastStatsException;

    /**
     * Dynamically generates a subtree of the Readable Schema rooted
     * at the path specified by the xpath. The Readable Schema is
     * distrubuted among the componenets of hypercast. Each component
     * knows how to generate its portion of the Readable Schema. If a
     * component has sub-components, it calls getReadSchema
     * recursively.
     *
     * Note that the Readable Schema is disjoint from the Writable
     * Schema. The Readable Schema is not modeled as a subset of the
     * Writable Schema.
     *
     * @throws HyperCastStatsException If some part of the specified
     * subtree does not correctly implement this interface method.
     */
    public Element[] getReadSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException;

    /**
     * Dynamically generates a subtree of the Writable Schema rooted
     * at the path specified by the xpath. The Writable Schema is
     * distrubuted among the componenets of hypercast. Each component
     * knows how to generate its portion of the Writable Schema. If a
     * component has sub-components, it calls getWriteSchema
     * recursively.
     *
     * Note that the Readable Schema is disjoint from the Writable
     * Schema. The Readable Schema is not modeled as a subset of the
     * Writable Schema.
     *
     * @throws HyperCastStatsException If some part of the specified
     * subtree does not correctly implement this interface method.
     */
    public Element[] getWriteSchema (final Document doc, final XPath xpath)
        throws HyperCastStatsException;
        
        
	/**
	 * Gets the element name of this I_Stats instance. The element name
	 * is originally assigned from the attribute StstsName in the 
	 * configuration file, and can be changed by this method.
	 */
	public String getStatsName ();

    /**
     * Assigns a name to this I_Stats instance. The assigned name becomes
     * the tag of the element associated to this I_Stats instance on the 
     * statistics tree.
     * 
     * @param name the name assigned to the I_Stats instance.
     */
	public void setStatsName (final String name);
    
}
