/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import java.io.FileNotFoundException;

import hypercast.I_Stats;
import hypercast.I_LogicalAddress;
import hypercast.HyperCastWarningRuntimeException;

/**
 * The <code>Monitor</code> class provides the API used to access
 * remote <code>Portals</code>. This class can be seen as the
 * combination of <code>PortalListManager</code> and
 * <code>MCSocket</code> with a few extra services.
 * 
 * Remote <code>Portals</code> are addressed using integer
 * indices. Once an index is assigned to a <code>Portal</code>, that
 * index is bound to that <code>Portal</code> for the duration of the
 * existence of the <code>Monitor</code>, i.e. indices are never
 * reassigned (unless <code>compactPortalList()</code> or
 * <code>clearPortalList()</code> is called).
 * 
 * Responsible for: - ensuring uniqueness of message ids - registering
 * callbacks for message receipt and timeouts - delegating calls to
 * <code>MCSocket</code> and <code>PortalListManager</code>
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 * @author Greg Mattes, January 2005
 */
public interface I_Monitor
    extends I_Stats
{
    /**
     * Get an integer array containing the indices of active portals.
     * 
     * @return An integer array containing the indices of active
     * portals, can be zero length array, but never
     * <code>null</code>. The length of this array is <i>at most</i>
     * the number of Portals.
     */
    public int [] getActivePortalIndices();
    
    /**
     * Get an integer array containing the indices of inactive portals.
     * 
     * @return An integer array containing the indices of inactive
     * portals, can be zero length array, but never
     * <code>null</code>. The length of this array is <i>at most</i>
     * the number of Portals.
     */
    public int[] getInactivePortalIndices();
    /**
     * Get An integer array containing the indexes of all.
     * 
     * @return An integer array containing the indexes of all
     * portals, can be zero length array, but never
     * <code>null</code>. The length of this array is the number
     * of Portals.
     */
    public int [] getAllPortalIndices();

    /**
     * Check if a portal is still active
     * 
     * @param portalIndex	the index of the portal
     * @return active status of specified Portal
     * @throws NoSuchPortalException	if this portal doesn't exist.
     */
    public boolean isPortalActive (final int portalIndex)
        throws NoSuchPortalException;
    
    /**
     * Clear the portal list. Remove all portal entires. Does not
     * cause Portals to stop. This is a local operation on Monitor
     * data structures.
     */
    public void clearPortalList();
    
    /**
     * Compact the portal list Remove all inactive portals from
     * Monitor data structures.
     */
    public void compactPortalList();
    
    /**
     * Sends an <code>MonMessage</code> to a single
     * <code>Portal</code>. A reply message from the
     * <code>Portal</code> having the same message ID as the request
     * message sent from the <code>Monitor</code> will be the
     * parameter of the <code>receiveCallback</code> for handling by
     * the application. The application need not provide a
     * <code>receiveCallback</code> (can be null) if it is
     * uninterested in the response. If no reply is received during
     * the timeout time (configured in the config file), the
     * <code>timeoutCallback</code> will be invoked with the request
     * message as a parameter (a copy of the request message is made
     * by the Monitor to ensure that the original request message is
     * given to the caller in the <code>timeoutCallback</code>, this
     * is because a caller is free to use a message object many times,
     * possibly changing the contents between uses). A
     * <code>timeoutCallback</code> need not be specified if the
     * application is not interested in timeouts (in this case no copy
     * of the original message is made).
     *
     * Note that a caller may use the same message object again and
     * again as a parameter to this method and each invocation will be
     * interpreted as a new request <I>so long as the message id field
     * is cleared before each call of this method</I>
     *
     * In the case of a <code>timeoutCallback</code> invocation, an
     * application that wants to retransmit a request can call
     * <code>sendTo()</code> again with the message reference that is
     * passed via the <code>timeoutCallback</code> (this reference
     * points to a copy of the original message that was made when it
     * was sent). Doing this invokes a retransmission using the same
     * message id as the original transmission. <I>Do not clear the
     * message id before calling <code>sendTo()</code> for
     * retransmission</I>.The subsequent call to <code>sendTo()</code>
     * will detect that the message was aready sent and not assign a
     * new message id to the request.
     *
     * Be careful not to change the state of the message after calling
     * <code>sendTo()</code> for a retransmission because in that
     * situation the Monitor implementation and the application both
     * have a pointer to the same data structure. Remember the the
     * Monitor made a copy of the message on the first send so that it
     * could give a pointer to the copy to the application on a
     * timeout. By calling <code>sendTo()</code> with this reference
     * you are simply passing a pointer to an object that should be
     * considered internal state of the Monitor.
     * 
     * @param portalIndex     index of destination portal
     * @param message         message to send, cannot be null
     * @param receiveCallback callback object handling the reply message, can be null
     * @param timeoutCallback callback to handle the timeout event, can be null
     *
     * @throws NullPointerException if message is null
     */
    public void sendTo (final int               portalIndex,
                        final MonMessage         message,
                        final I_ReceiveCallback receiveCallback,
                        final I_TimeoutCallback timeoutCallback)
        throws NoSuchPortalException;
    
    /**
     * Send a MC message to all portals by multicast.
     * A reply message with the same message ID as the message sent
     * will be sent to the receiveCallback for handling.
     * 
     * @param message				message to send, cannot be null
     * @param receiveCallback	callback to handle reply message
     * @param timeoutCallback callback to handle situation when no portal replies
     *
     * @throws NullPointerException if message is null
     */
    public void sendToAll (final MonMessage         message,
                           final I_ReceiveCallback receiveCallback,
                           final I_TimeoutCallback timeoutCallback);

    /**
     * Get the index of a portal by its logical address.
     * 
     * @param portalLogicalAddress	the logical address of the portal 
     * @return		the index of the portal
     * @throws NoSuchPortalException	if no portal exists with that logical address
     */
    public int indexOfPortal (final I_LogicalAddress portalLogicalAddress)
        throws NoSuchPortalException;

    /**
     * Get the logical address of a Portal
     * 
     * @param portalIndex The index of the Portal.
     * @return	The logical address of the Portal.
     * @throws NoSuchPortalException If the portal doesn't exist.
     */
    public I_LogicalAddress getPortalAddress (final int portalIndex)
        throws NoSuchPortalException;
    
    /**
     * create a GetRequest message
     * 
     * "GetRequest" message is used to query for one or more statistics
     * on a Portal.  After the mesage is created, it can be
     * manipulated to add or remove statistics of interest.
     * 
     * @return	a MonMessage of "GetRequest" type
     */
    public MonMessage createGetRequest();
    

    /**
     * create a SetRequest message
     * 
     * "SetRequest" message is used to set the values of a list of statistics 
     * on a Portal. 
     * After thie mesage is created, it can be manipulated to add, remove, or 
     * remove XPath expression to represent the statistics of interest, and the
     * new XML value of the statistics to be set.
     * 
     * @return	a MonMessage of "SetRequest" type
     */
    public MonMessage createSetRequest();
    
    
    /**
     * create a ReadSchema request
     * 
     * "ReadSchema" message is used to query the schemas of a list of the readable
     * statistics on a Portal. 
     * After thie mesage is created, it can be manipulated to add, remove, or 
     * remove XPath expression to represent the statistics of interest.
     * 
     * @return a MonMessage of "ReadSchema" type
     */
    public MonMessage createReadSchemaRequest();
   
    
    /**
     * create a WriteSchema request
     * 
     * "WriteSchema" message is used to query the schemas of a list of the writable
     * statistics on a Portal. 
     * After thie mesage is created, it can be manipulated to add, remove, or 
     * remove XPath expression to represent the statistics of interest.
     * 
     * @return	a MonMessage of "WriteSchema" type
     */
    public MonMessage createWriteSchemaRequest();

    
    /** 
     * Causes <code>MCMessages</code> to be recorded in a transcript
     * file. Depending on the options present in the configuration
     * file either messages that are sent, received, or both will be
     * recorded.
     *
     * @param filename The name of the transcript file.
     *
     * @throws HyperCastWarningRuntimeException If the header cannot be written to the transcript file.
     * @throws HyperCastWarningRuntimeException If the configuration
     * specifies that neither sent nor received messsages should be
     * recorded.
     * @throws FileNotFoundException If the specified file cannot be accessed
     * @throws IllegalStateException If a transcript is already started
     */
    public void startTranscript (final String filename)
        throws FileNotFoundException, IllegalStateException, HyperCastWarningRuntimeException;
    
    /**
     * Stops transcripting.
     *
     * @throws IllegalStateException If no transcript is active.
     */
    public void stopTranscript()
        throws IllegalStateException, HyperCastWarningRuntimeException;
    
    /**
     * Explicitly open monitor.
     *  
     * A Monitor cannot be used until it's opened. 
     */
    public void openMonitor();
    
    /**
     * explicitly close monitor
     */
    public void closeMonitor();

    /**
     * Causes a Monitor to cease sending MonitorAdvertise messages to Portals
     */
    public void setSilentOn();

    /**
     * Causes a Monitor to begin sending MonitorAdvertise messages to Portals
     */
    public void setSilentOff();
}
