/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.util.Vector;

import java.security.Signature;
import java.security.cert.Certificate;
import java.security.PublicKey;
import java.security.PrivateKey;
import java.security.SignatureException;
import java.security.InvalidKeyException;

/**
 * A <code>NamingBinding</code> object binds a single name to a single logical address. It has a timestamp which records
 * the creation time of the binding (NB: not necessarily the creation time of the <code>NamingBinding</code> object). It
 * also has a metric that reflects the number of times that the socket that created the binding has changed its logical
 * address. Finally it has an optional digitial signature that ensures the integrity and the authenticity of the binding
 * information.
 *
 * All of the naming binding fields mentioned thusfar are immutable - fixed at the time of the creation of the
 * binding. Binding information received from peers can never be modified so a <code>NamingBinding</code> object that
 * represents such data has fixed fields. A <code>NamingBinding</code> object that represents a local binding has its
 * fields (except name) updated on demand to reflect current values. This on-demand update is the responsibility of some
 * container that holds binding objects. This container calls the various "setter" methods when it wants an updated
 * binding. In the current (Spring 2005) implementation of the Naming Service this container is the
 * <code>PeerBindingCache</code> of the Naming Service Message Store FSM.
 *
 * Bindings also have the notion of being authoritative or not authoritative. This refers to whether or not a binding
 * was received from the peer that created the binding. This field of the binding is not covered by the digital
 * signature because if it was, it could never change.
 *
 * As an example, consider three peers: A, B, and C. A creates a binding and sends it to B where it is cached. Next, B
 * sends the cached binding to C. B received an authoritative binding because it received the binding from A and A
 * created the binding. C did not received an authoritative binding; C received a cached copy. Note that the
 * authoritiative binding and the cached binding both have a digital signature created by A. In particular B does not
 * resign the binding, therefore it cannot alter any fields of the binding. If the authoritative field was included in
 * the digital signature computation B could not change the authoritative field to mean "not authoritative" when
 * transmitting the binding to C.
 *
 * A binding can be created locally or it can be received from a peer. An argument can be made that this class [sc]ould
 * be factored into two subclasses: <code>LocalBinding</code> and <code>PeerBinding</code>, instead an instance
 * variable, <code>local</code>, is used to control the behavior of particular objects.
 *
 * There is no synchronization provided by this class. It is assumed that the caller ensures synchronization. For
 * example, computing a digital signature using the data from an object of this class while simultaneously altering that
 * data in another thread is almost certainly a Bad Thing.
 *
 * @author HyperCast Team
 * @author Greg Mattes
 * @version Spring 2005
 */
public class NamingBinding
    implements I_NamingBinding, I_LRUObject
{
    /** Controls general debugging output */
    protected static final boolean debug = false;

    /** Controls debugging output related to (un)marshalling/(de)serializing. */
    protected static final boolean debugRestore = false;

    /** Fixed for both local and peer bindings */
    protected final String name;

    /** Fixed for peer bindings, updated for local bindings on demand. */
    private String logicalAddressString;

    /** Subject common name of certificate needed to verify the digital signature of the binding (digital signature optional) */
    public String signerName;

    /**
     * Time of creation of this binding (not of the binding object, but of the binding). Measured in milliseconds since
     * epoch. Fixed for peer bindings, updated for local bindings each time the binding is needed.
     */
    private long timestamp;

    /** Fixed if binding information is received from a peer; set dynamically by local binding. */
    private int logicalAddressChangeCounter;

    /** true if this object represents a binding created at the local socket, false if from a peer. */
    private final boolean local;

    /**
     * True if peer binding and received from authoritative source. Always true of bindings created locally. Only false
     * if peer binding received from the cache of another peer.
     */
    private final boolean authoritative;

    /**
     * Signature computed over fields of the binding. Fixed for bindings received from a peer; updated on demand for
     * local bindings.
     *
     * @see #binaryForm
     */
    private byte [] digitalSignature;

    /**
     * true if it has been decided that this binding can be trusted. Bindings for which there is no signature are not
     * considered to be trusted. All local bindings are considered to be trusted.
     */
    private boolean trusted;

    /**
     * True if the signature of the binding has been verified to have been created using the private key corresponding
     * to the public key contained in the certificate that matches the name of the binding.
     */
    private boolean verified;

    /** The delegate bit controls whether or the logical address
     * string field of a locally created binding can change. Usually,
     * locally created bindings use the logical address of the local
     * socket and the binding is updated when the logical address
     * changes.  Some locally created bindings use a logical address
     * other than the address of the local socket. In these cases, the
     * address used in the binding should not be changed when the
     * address of the socket changes.  */
    private boolean delegate;

    /**
     * Everything except the digital signature data - the length of signature (last two bytes of binary form) are set to
     * "(byte) 0" for signing and verifying. Note that this means that these two bytes are set to zero upon receipt of a
     * binding from a peer and must be set to reflect the size of the signature when sent to other peers. The binary
     * form of a local binding is computed on demand since at least the timestamp changes.
     */
    private byte [] binaryForm;

    /**
     * Pointer to a structure that controls timeouts. A pointer is stored here to avoid linear search through the
     * timeout queue when resetting the timeout.
     */
    private LRUTimeoutQueue.Entry toqEntry;

    /**
     * Used to construct a peer binding.
     *
     * @param name Bound name
     * @param logicalAddressString Bound logical address
     * @param signerName Subject common name of certificate needed to verify the digital signature of the binding (digital signature optional) 
     * @param timestamp Time of binding creation
     * @param logicalAddressChangeCounter Number of times the logical address of the socket that created the binding had
     * changed when the binding was created
     * @param digitalSignature Signature computed by the creator of the binding that ensure authenticity and integrity of binding data
     * @param binaryForm Marshalled binding data, it is this data that is sent to other peers from the local cache -
     * it's already computed when received from a peer so save it rather than recompute
     * @param authoritative True if binding is received from the peer that created the binding
     */
    protected NamingBinding (final String     name,
                             final String     logicalAddressString,
                             final String     signerName,
                             final long       timestamp,
                             final int        logicalAddressChangeCounter,
                             final byte    [] digitalSignature,
                             final byte    [] binaryForm,
                             final boolean    authoritative)
    {
        this.local = false;

        this.name = name;
        this.logicalAddressString = logicalAddressString;
        this.signerName = signerName;

        this.timestamp = timestamp;
        this.logicalAddressChangeCounter = logicalAddressChangeCounter;

        this.digitalSignature = digitalSignature;
        this.binaryForm = binaryForm;

        this.verified = false;
        this.trusted = false;

        this.authoritative = authoritative;
    }

    /**
     * Base class constructor called when bindings are created locally.
     *
     * @param name Bound name
     * @param logicalAddressString Bound logical address
     * @param signerName Subject common name of certificate needed to verify the digital signature of the binding (digital signature optional) 
     */
    public NamingBinding (final String name, final String logicalAddressString, final String signerName)
    {
        this.local = true;

        this.name = name;
        this.logicalAddressString = logicalAddressString;
        this.delegate = logicalAddressString != null;
        this.signerName = signerName;

        this.timestamp = 0;
        this.logicalAddressChangeCounter = 0;

        this.digitalSignature = null;
        this.binaryForm = null;

        this.verified = true;
        this.trusted = true;

        this.authoritative = true;
    }

    /**
     * Returns the name of the binding. The name field of objects of this class can only be set by a constructor.
     */
    public String getName() {
        return name;
    }

    public String getSignerName() {
        return signerName;
    }

    /**
     * Returns the logical address of the binding.
     */
    public String getLogicalAddressString() {
        return logicalAddressString;
    }

    /**
     * Sets the logical address of the local binding.
     */
    public void setLogicalAddressString (final String logicalAddressString)
    {
        if (! local)
            throw new HyperCastFatalRuntimeException ("Can set logical address for local bindings only: <" + name + " <-> " + logicalAddressString + ">");

        if (delegate)
            return;

        this.logicalAddressString = logicalAddressString;
    }

    /**
     * Returns the number of times that the logical address of the socket that made this binding changed at the time the
     * binding was created.
     */
    public int getLogicalAddressChangeCounter() {
        return logicalAddressChangeCounter;
    }

    /**
     * Sets the value of the logical address change counter for a local binding.
     */
    public void setLogicalAddressChangeCounter (final int logicalAddressChangeCounter)
    {
        if (! local)
            throw new HyperCastFatalRuntimeException ("Can set logical address change counter for local bindings only: <" + name + " <-> " + logicalAddressString + ">");

        this.logicalAddressChangeCounter = logicalAddressChangeCounter;
    }

    /**
     * Returns the time the binding was created.
     */
    public long getTimestamp() {
        return timestamp;
    }

    /**
     * Sets the values of the timestamp for a local binding.
     */
    public void setTimestamp (final long timestamp)
    {
        if (! local)
            throw new HyperCastFatalRuntimeException ("Can set timestamp for local bindings only: <" + name + " <-> " + logicalAddressString + ">");

        this.timestamp = timestamp;
    }

    /**
     * Returns the age of the binding in milleseconds relative to the time parameter given.
     */
    public long getAge (final long now) {
        return local ? 0 : now - getTimestamp();
    }

    /**
     * Returns whether or not this binding was created by the local socket.
     */
    public boolean isLocal() {
        return local;
    }

    /**
     * Returns whether or not this binding was received from an authoritative source (local implies authoritative). The
     * authoritive aspect of an object of this class is determined at construction time and stays fixed, i.e. there is
     * no "setAuthoritative" method, even for local bindings. When a binding is serialized for sending to a peer the
     * authoritative field in the serialized representation is set appropriately, i.e. authoritiative if the binding was
     * created locally, non-authoritative if the binding was received from a peer.
     *
     * @see #toByteArray(Signature,PrivateKey)
     */
    public boolean isAuthoritative() {
        return authoritative;
    }

    /**
     * This method is part of the binding interface that is not used here. To establish trust in this implemenation a
     * digital signature verifier and certificate cache entry must be provided.
     */
    public boolean isTrusted () { throw new UnsupportedOperationException(); }

    /**
     * Decides if this binding is trusted using certificates. Local bindings are considered to be trusted by definition.
     *
     * @param signer Security object used to sign the binding information.
     * @param certificateCacheEntry Entry of certificate cache that points the certificate needed for the crypto operations of this method.
     */
    public boolean isTrusted (final Signature signer, final CertificateCache.Entry certificateCacheEntry)
    {
        if (debug)
            System.err.println ("NamingBinding.isTrusted: <" + name + " <-> " + logicalAddressString + "> " +
                                "local (" + local + ") verified (" + verified + ") trusted (" + trusted + ")");

        final boolean result;

        if (local)
        {
            result = true;
        }
        else if (signer == null)
        {
            result = false;
        }
        else if (trusted)
        {
            if (! verified)
                throw new HyperCastFatalRuntimeException ("NamingBinding is marked trusted but not verified: logic error: " + this);

            result = true;
        }
        else if (digitalSignature == null
                 ||
                 digitalSignature.length == 0
                 ||
                 certificateCacheEntry == null)
        {
            result = false;
        }
        else
        {
            if (! verified)
            {
                verify (signer, certificateCacheEntry.getCertificate().getPublicKey());

                if (debug)
                    System.err.println ("NamingBinding.isTrusted: after verify() - verified (" + verified + ") trusted (" + trusted + ")");
            }

            if (verified
                &&
                ! trusted)
            {
                trusted = certificateCacheEntry.isTrusted();

                if (debug)
                    System.err.println ("NamingBinding.isTrusted: after certificateCacheEntry.isTrusted() - verified (" + verified + ") trusted (" + trusted + ")");
            }

            result = trusted;
        }
        return result;
    }

    /** Used in situation where trusted state should not be set permanently - like in experiments */
    public void clearTrusted() {
        verified = false;
        trusted = false;
    }

    /**
     * Returns a pointer to an entry in a timeout queue for this binding.
     */
    public LRUTimeoutQueue.Entry getTimeoutQueueEntry() {
        return toqEntry;
    }

    /**
     * Sets the timeout queue entry for a binding.
     */
    public void setTimeoutQueueEntry (final LRUTimeoutQueue.Entry toqEntry) {
        this.toqEntry = toqEntry;
    }

    /**
     * This method passes the data that will be signed to the signer object.
     *
     * @param signer Security object used to sign the binding information.
     */
    protected void addSignatureData (final Signature signer)
        throws SignatureException
    {
        if (binaryForm == null)
            throw new NullPointerException ("NamingBinding: binaryForm == null");

        signer.update (binaryForm);
    }

    /**
     * A local binding is signed when it is released to a peer, not when the binding is created. In this way the current
     * logical address, current timestamp, etc. are signed. Bindings received from peers are never signed; only the
     * creator of a binding may sign it.
     *
     * @param signer Security object used to sign the binding information.
     * @param privateKey The private key needed to create the signature of this binding.
     */
    private void sign (final Signature signer, final PrivateKey privateKey)
    {
        if (! local)
            throw new HyperCastFatalRuntimeException ("Cannot sign a non-local binding: <" + name + " <-> " + logicalAddressString + ">");

        try
        {
            if (privateKey == null)
                throw new HyperCastWarningRuntimeException ("Private Key cannot be null", new NullPointerException ("privateKey == null"));

            signer.initSign (privateKey);
            addSignatureData (signer);
            digitalSignature = signer.sign();
        }
        catch (final SignatureException se)
        {
            se.printStackTrace();
            throw new HyperCastWarningRuntimeException ("SignatureException encountered when signing binding: ", se);
        }
        catch (final InvalidKeyException ike)
        {
            throw new HyperCastWarningRuntimeException ("InvalidKeyException encountered when signing binding: ", ike);
        }
    }

    /**
     * Verifies the digital signature of a binding. A binding can be verified but not trusted. This can happen if it can
     * be confirmed that the binding was signed by the holder of the private key for the given certificate, but the
     * certificate itself is not (yet) considered trusted.
     *
     * @param verifier Security object used to verify the signature of the binding
     * @param publicKey The public key needed to verify the signature of this binding.
     */
    private void verify (final Signature verifier, final PublicKey publicKey)
    {
        if (local)
            throw new HyperCastFatalRuntimeException ("NamingBinding.verify: Local bindings are never verified");

        if (verified)
            return;

        try
        {
            verifier.initVerify (publicKey);
            addSignatureData (verifier);
            this.verified = verifier.verify (digitalSignature);
        }
        catch (final SignatureException se)
        {
            throw new HyperCastWarningRuntimeException ("SignatureException encountered when verifying binding: ", se);
        }
        catch (final InvalidKeyException ike)
        {
            throw new HyperCastWarningRuntimeException ("InvalidKeyException encountered when verifying binding: ", ike);
        }
    }

    /**
     * Creates a string with information about the binding object.
     */
    public String toString() {
        final long now = System.currentTimeMillis();
        return
            "<"                                                                                  +
            "name"          + " -> " + name                                                      + ", "         +
            "LA"            + " -> " + logicalAddressString                                      + ", "         +
            "local"         + " -> " + local                                                     + ", "         +
            "authoritative" + " -> " + authoritative                                             + ", "         +
            "timestamp"     + " -> " + (local ? 0 : ((now - getTimestamp())/1000))               + " sec old, " +
            "sig"           + " -> " + (digitalSignature != null && digitalSignature.length > 0) + ", "         +
            "signer"        + " -> " + signerName                                                + ", "         +
            "trusted"       + " -> " + trusted                                                   + ", "         +
            "verified"      + " -> " + verified                                                  + ", "         +
            "LA changes"    + " -> " + logicalAddressChangeCounter                               +
            ">";
    }

    /**
     * Serializes all local binding information, note that binary form does not include the digital signature. Rather
     * the digital signature is computed over the binary form. The value of the binary form is stored in an instance
     * variable.
     */
    private void calculateBinaryForm()
    {
        if (! local)
            throw new HyperCastFatalRuntimeException ("Can calculate binary form of local binding only: <" + name + " <-> " + logicalAddressString + ">");
            
        final int nameLength = name.length() & 0xFFFF;
        final int logicalAddressStringLength = logicalAddressString.length() & 0xFFFF;
        final int signerNameLength = signerName.length() & 0xFFFF;

        final int binaryFormLength = 
            NAME_LENGTH_BYTES + nameLength +
            LOGICAL_ADDRESS_STRING_LENGTH_BYTES + logicalAddressStringLength +
            NAME_LENGTH_BYTES + signerNameLength +
            TIMESTAMP_SIZE +
            LOGICALADDRESSCHANGECOUNTER_SIZE +
            DIGITAL_SIGNATURE_LENGTH_BYTES;

        this.binaryForm = new byte [binaryFormLength];

        int index = 0;

        // Name String Length
        this.binaryForm [index + 0] = (byte) ((nameLength >>> (8*1)) & 0xFF);
        this.binaryForm [index + 1] = (byte) ((nameLength >>> (8*0)) & 0xFF);
        index += NAME_LENGTH_BYTES;

        // Name String
        System.arraycopy (name.getBytes(), 0, this.binaryForm, index, nameLength);
        index += nameLength;

        // Logical Address String Length
        this.binaryForm [index + 0] = (byte) ((logicalAddressStringLength >>> (8*0)) & 0xFF);
        index += LOGICAL_ADDRESS_STRING_LENGTH_BYTES;

        // Logical Address String
        System.arraycopy (logicalAddressString.getBytes(), 0, this.binaryForm, index, logicalAddressStringLength);
        index += logicalAddressStringLength;

        // Signer Name String Length
        this.binaryForm [index + 0] = (byte) ((signerNameLength >>> (8*1)) & 0xFF);
        this.binaryForm [index + 1] = (byte) ((signerNameLength >>> (8*0)) & 0xFF);
        index += NAME_LENGTH_BYTES;

        // Signer Name String
        System.arraycopy (signerName.getBytes(), 0, this.binaryForm, index, signerNameLength);
        index += signerNameLength;

        // Timestamp
        this.binaryForm [index + 0] = (byte) ((timestamp >>> (8*7)) & 0xFF);
        this.binaryForm [index + 1] = (byte) ((timestamp >>> (8*6)) & 0xFF);
        this.binaryForm [index + 2] = (byte) ((timestamp >>> (8*5)) & 0xFF);
        this.binaryForm [index + 3] = (byte) ((timestamp >>> (8*4)) & 0xFF);
        this.binaryForm [index + 4] = (byte) ((timestamp >>> (8*3)) & 0xFF);
        this.binaryForm [index + 5] = (byte) ((timestamp >>> (8*2)) & 0xFF);
        this.binaryForm [index + 6] = (byte) ((timestamp >>> (8*1)) & 0xFF);
        this.binaryForm [index + 7] = (byte) ((timestamp >>> (8*0)) & 0xFF);
        index += TIMESTAMP_SIZE;

        // LogicalAddressChangeCounter
        this.binaryForm [index + 0] = (byte) ((logicalAddressChangeCounter >>> (8*3)) & 0xFF);
        this.binaryForm [index + 1] = (byte) ((logicalAddressChangeCounter >>> (8*2)) & 0xFF);
        this.binaryForm [index + 2] = (byte) ((logicalAddressChangeCounter >>> (8*1)) & 0xFF);
        this.binaryForm [index + 3] = (byte) ((logicalAddressChangeCounter >>> (8*0)) & 0xFF);
        index += LOGICALADDRESSCHANGECOUNTER_SIZE;

        if (debugRestore)
            System.err.println ("NamingBinding: calculateBinaryForm() index: " + index +
                                " binaryForm.length: " + this.binaryForm.length +
                                " diff (should be 2): " + (this.binaryForm.length - index));

        // make sure that digital sig length bytes are zero
        this.binaryForm [index + 0] = (byte) 0;
        this.binaryForm [index + 1] = (byte) 0;
        index += DIGITAL_SIGNATURE_LENGTH_BYTES;
    }
    
    /**
     * Serializes and signs a binding.
     *
     * @param signer Security object used to create the signature of the binding
     * @param privateKey The private key needed to verify the signature of this binding.
     */
    public synchronized byte [] toByteArray (final Signature signer, final PrivateKey privateKey)
    {
        final byte [] result;

        if (! local)
        {
            final int digitalSignatureLength = digitalSignature.length & 0xFFFF;

            if (debugRestore)
                System.err.println ("NamingBinding.toByteArray: digitalSignatureLength: " + digitalSignatureLength);

            final int totalLength = AUTHORITATIVE_FLAG_LENGTH + binaryForm.length + digitalSignature.length;

            result = new byte [totalLength];
            int index = 0;

            // If we are sending a binding that we received from a peer we are not the authoritative source of the binding.
            result [0] = (byte) 0;
            index += AUTHORITATIVE_FLAG_LENGTH;

            System.arraycopy (binaryForm, 0, result, index, binaryForm.length);
            index += binaryForm.length;

            System.arraycopy (digitalSignature, 0, result, index, digitalSignature.length);
            index += digitalSignature.length;

            result [AUTHORITATIVE_FLAG_LENGTH + (binaryForm.length - 2)] = (byte) ((digitalSignatureLength >>> (8*1)) & 0xFF);
            result [AUTHORITATIVE_FLAG_LENGTH + (binaryForm.length - 1)] = (byte) ((digitalSignatureLength >>> (8*0)) & 0xFF);

            if (debugRestore)
                System.err.println ("non-local: result [binaryForm.length - 2]: " + result [binaryForm.length - 2] + " result [binaryForm.length - 1]: " + result [binaryForm.length - 1]);

            return result;
        }
        else
        {
            calculateBinaryForm();

            if (signer != null && privateKey != null)
                sign (signer, privateKey);
            else
                this.digitalSignature = new byte [0];

            if (debugRestore)
                System.err.println ("NamingBinding.toByteArray: digitalSignature " + ((digitalSignature == null) ? "" : "not") + " null");

            if (debugRestore)
            {
                System.err.println ("NamingBinding.toByteArray: digitalSignature.length: " + digitalSignature.length);
                System.err.println ("NamingBinding.toByteArray: binaryForm.length: " + binaryForm.length);
            }

            final int digitalSignatureLength = digitalSignature.length & 0xFFFF;

            final int totalLength = AUTHORITATIVE_FLAG_LENGTH + binaryForm.length + digitalSignature.length;

            result = new byte [totalLength];
            int index = 0;

            // If we are sending a binding that we created we are the authoritative source of the binding.
            result [0] = (byte) 1;
            index += AUTHORITATIVE_FLAG_LENGTH;

            System.arraycopy (binaryForm, 0, result, index, binaryForm.length);
            index += binaryForm.length;

            System.arraycopy (digitalSignature, 0, result, index, digitalSignature.length);
            index += digitalSignature.length;

            // Fill in the digital signature length field in the binary form
            result [AUTHORITATIVE_FLAG_LENGTH + (binaryForm.length - 2)] = (byte) ((digitalSignatureLength >>> (8*1)) & 0xFF);
            result [AUTHORITATIVE_FLAG_LENGTH + (binaryForm.length - 1)] = (byte) ((digitalSignatureLength >>> (8*0)) & 0xFF);

            // These are always generated on demand for a local binding. Stop referring to them so they can be
            // garbage collected.

            this.digitalSignature = null;
            this.binaryForm = null;
        }

        return result;
    }

    // one byte is used to indicate authoritative response. The data in this byte is not signed.
    private static final int AUTHORITATIVE_FLAG_LENGTH = 1;

    // units for all values is bytes
    private static final int NAME_LENGTH_BYTES                   = 2;
    private static final int LOGICAL_ADDRESS_STRING_LENGTH_BYTES = 1;
    private static final int DIGITAL_SIGNATURE_LENGTH_BYTES      = 2;
    private static final int TIMESTAMP_SIZE                      = 8;
    private static final int LOGICALADDRESSCHANGECOUNTER_SIZE    = 4;

    /**
     * Creates invalidation naming binding object given an array of one or more serialized bindingss.
     */
    public static NamingBinding [] restoreInvalidateNamingBindings (final byte [] data, int offset) {
        return restoreNamingBindings (data, offset, true);
    }

    /**
     * Creates naming binding object given an array of one or more serialized bindingss.
     */
    public static NamingBinding [] restoreNamingBindings (final byte [] data, int offset) {
        return restoreNamingBindings (data, offset, false);
    }

    /**
     * Given an array of serialized binding objects, parses array and instanciates objects.
     */
    public static NamingBinding [] restoreNamingBindings (final byte [] data, int index, final boolean makeInvalidate)
    {
        final Vector result = new Vector();
        try
        {
            if (debugRestore) System.err.println ("restoreNamingBindings data: (" + ((data == null) ? -1 : (data.length - index)) + ")");

            int count = 0; // for debugging
            do
            {
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + ") index:" + index);

                final boolean authoritative = data [index] == (byte) 1;
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): authoritative: " + authoritative);
                index += AUTHORITATIVE_FLAG_LENGTH;

                final int startIndex = index;

                final int nameLength = ByteArrayUtility.toUnsignedInteger (data, index, NAME_LENGTH_BYTES);
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): nameLength: " + nameLength);
                index += NAME_LENGTH_BYTES;

                final String name = new String (data, index, nameLength);
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): name: " + name);
                index += nameLength;

                final int logicalAddressStringLength = ByteArrayUtility.toUnsignedInteger (data, index, LOGICAL_ADDRESS_STRING_LENGTH_BYTES);
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): logicalAddressStringLength: " + logicalAddressStringLength);
                index += LOGICAL_ADDRESS_STRING_LENGTH_BYTES;
            
                final String logicalAddressString = new String (data, index, logicalAddressStringLength);
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): logicalAddressString: " + logicalAddressString);
                index += logicalAddressStringLength;

                final int signerNameLength = ByteArrayUtility.toUnsignedInteger (data, index, NAME_LENGTH_BYTES);
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): nameLength: " + signerNameLength);
                index += NAME_LENGTH_BYTES;

                final String signerName = new String (data, index, signerNameLength);
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): name: " + signerName);
                index += signerNameLength;

                final long timestamp = ByteArrayUtility.toLong (data, index);
                if (debugRestore)
                {
                    final long now = System.currentTimeMillis();
                    System.err.println ("restoreNamingBindings(" + count + "): timestamp: " + ((now - timestamp)/1000) + " secs ago");
                }
                index += TIMESTAMP_SIZE;

                final int logicalAddressChangeCounter = ByteArrayUtility.toInteger (data, index);
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): logicalAddressChangeCounter: " + logicalAddressChangeCounter);
                index += LOGICALADDRESSCHANGECOUNTER_SIZE;

                final int digitalSignatureLength = ByteArrayUtility.toUnsignedInteger (data, index, DIGITAL_SIGNATURE_LENGTH_BYTES);
                if (debugRestore) System.err.println ("restoreNamingBindings(" + count + "): digitalSignatureLength: " + digitalSignatureLength);

                index += DIGITAL_SIGNATURE_LENGTH_BYTES;

                final byte [] binaryForm = new byte [index - startIndex];
                System.arraycopy (data, startIndex, binaryForm, 0, index - startIndex);

                // signature is not computed with value of sig length because sig length is unknown before sig is
                // computed. zero out the bytes that hold the length so that the data can be verified with the
                // sig. length of sig is known by taking the length of the byte array containing the sig, restore this
                // value to the byte array when propagating this binding.
                for (int i = 0; i < DIGITAL_SIGNATURE_LENGTH_BYTES; ++i)
                    binaryForm [((index - startIndex) - DIGITAL_SIGNATURE_LENGTH_BYTES) + i] = (byte) 0;

                final byte [] digitalSignature = new byte [digitalSignatureLength];
                System.arraycopy (data, index, digitalSignature, 0, digitalSignatureLength);
                index += digitalSignatureLength;

                final NamingBinding binding;
                if (! makeInvalidate)
                    binding = new NamingBinding (name, logicalAddressString, signerName, timestamp, logicalAddressChangeCounter, digitalSignature, binaryForm, authoritative);
                else {
                    // Only invalidations from an authoritative source are considered - otherwise any peer could replay an invalidation.
                    binding = authoritative
                        ? new BindingInvalidation (name, logicalAddressString, signerName, timestamp, logicalAddressChangeCounter, digitalSignature, binaryForm)
                        : null;
                }

                if (binding != null)
                    result.add (binding);

                if (debugRestore) ++count;
            }
            while (index < data.length);

            return (NamingBinding []) result.toArray (new NamingBinding [0]);
        }
        catch (final ArrayIndexOutOfBoundsException aioobe)
        {
            System.err.println ("Badly formatted Naming Message, data length: " + ((data == null) ? -1 : data.length) + " : " + aioobe);
            aioobe.printStackTrace();
            return (NamingBinding []) result.toArray (new NamingBinding [0]);
        }
    }

    /**
     * Concatenates byte representations of all bindings into single byte array. For each binding, there may or may not
     * be a certificate. For each binding with a corresponding certificate there may or may not be a private key.
     *
     * @see #restoreNamingBindings(byte[], int, boolean)
     *
     * Note: this method could be very inefficient regarding the handling of byte arrays. Instead of making a small byte
     * array for each binding and then allocating a large array that holds all of the smaller array (with a copy for each of
     * the smaller arrays), perhaps a large array could be created at the beginning of the method (the size of this array
     * would be a guess). Each particular binding would be packed into the large array as long as there is room to do so. The
     * large array would be expanded as needed. The "toByteArray" method of NamingBinding would take two extra params: the
     * large array and an offset into the large array. The final valid byte of the large array would have to be communicated
     * back to the caller.
     */
    public static byte [] bindingsToByteArray (final NamingBinding [] bindings, final Signature signer, final CertificateCache certificateCache)
    {
        final Vector bindingBytes = new Vector();
        int length = 0;
        for (int i = 0; i < bindings.length; ++i)
        {
            final CertificateCache.Entry certificateCacheEntry = certificateCache.get (bindings[i].getSignerName());

            if (debugRestore)
                System.err.println ("NamingBinding.bindingsToByteArray: certificate cache entry: " + certificateCacheEntry);

            final PrivateKey privateKey = (certificateCacheEntry == null) ? null : certificateCacheEntry.getPrivateKey();
            final byte [] bytes = bindings[i].toByteArray (signer, privateKey);

            if (debugRestore)
                if (bytes == null)
                    System.err.println ("NamingBinding.bindingsToByteArray: bytes is null");
                else
                    System.err.println ("NamingBinding.bindingsToByteArray: bytes length: " + bytes.length);

            length += bytes.length;
            bindingBytes.add (bytes);
        }

        final byte [] result = new byte [MessageStoreFSM_Naming.SERIAL_NUMBER_LENGTH + length];
        int index = MessageStoreFSM_Naming.SERIAL_NUMBER_LENGTH;

        for (int i = 0; i < bindingBytes.size(); ++i)
        {
            final byte [] currentBytes = (byte []) bindingBytes.elementAt (i);
            final int currentLength = currentBytes.length;
            System.arraycopy (currentBytes, 0, result, index, currentLength);
            index += currentLength;
        }

        return result;
    }
}

/**
 * A binding used to invalidate another binding.
 */
final class BindingInvalidation
    extends NamingBinding
{
    /**
     * Added to the hash of an invalidation binding so that the hash is different than that of a regular binding,
     * e.g. pull or push. In this way, a hash computed for a binding cannot be used by an attacker to fabricate an
     * invalidate message.
     */
    private String invalidateString = "INVALIDATE";

    /**
     * Called for local bindings.
     */
    public BindingInvalidation (final String name, final String logicalAddressString, final long timestamp)
    {
        super (name, logicalAddressString, null);
        setTimestamp (timestamp);
    }

    /**
     * Called when restoring a binding invalidation received from peer. All invalidations are authoritative, i.e. they
     * are not cached and forwarded.
     */
    public BindingInvalidation (final String    name,
                                final String    logicalAddressString,
                                final String    signerName,
                                final long      timestamp,
                                final int       logicalAddressChangeCounter,
                                final byte   [] digitalSignature,
                                final byte   [] binaryForm)
    {
        super (name, logicalAddressString, signerName, timestamp, logicalAddressChangeCounter, digitalSignature, binaryForm, true);
    }

    /**
     * Adds all data for the signer to sign including the string particular to invalidations that makes invalidation
     * digests different from regular digests.
     */
    protected void addSignatureData (final Signature signer)
        throws SignatureException
    {
        super.addSignatureData (signer);
        signer.update (invalidateString.getBytes());
    }
}
