/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.lang.Thread;
import java.util.Vector;

import hypercast.events.*;

/**
 * This class implements a notification mechanism for applications so
 * that applications can capture asynchronous events that happen in
 * the system.  To receive notification of events the application
 * extends the NotificationHandler class. When an interesting event
 * occurs, the application is notified.
 *
 * As events occur they are placed into a queue and sent to the
 * application in FIFO order. Applications may block a thread on an
 * event. Each event type has a lock object for this
 * purpose. Application may also choose to extend the default
 * asynchronous handler for an event type. If an application does
 * both, it will receive an event twice.
 *
 * When blocking a thread on an event type, a boolean value can be
 * specified that determines whether or not an event that occurred
 * before the blocking call took place will cause the blocking thread
 * to unblock. For example if a node generates an NODE_ISSTABLE event
 * only once, but the application calls waitUntil_NODE_ISSTABLE
 * *after* this event occurs, it will never see the event unless it
 * asks to see any prior, unconsumed event. Note that only one prior
 * event is accessible, in general events do not queue if nobody is
 * listening for them.
 *
 * A third way in which applications may receive notification of event
 * is by overriding the method of this class that dispatches
 * individual event handlers: <code>handler</code>.
 * 
 * To add a new event search this file for TEMPLATE and copy and paste
 * templates doing a query/replace for the event type name.
 * Also, update the HyperCastAppl.java file.
 *
 * Although this class is designed to be extended, it is not
 * abstract. This is because certain mechanisms, such as the naming
 * system, need to use the notification mechanism even when the
 * application does not.
 *
 * @author HyperCast Team
 * @author Greg Mattes
 * @author Jianping Wang
 * @version 3.0, November 2004 
 */
public class NotificationHandler
{
    /**
     * A queue of pending events. A new event is put into the queue by
     * the components in <code>OL_Socket</code>; the queue is served
     * in this class.  Each event causes an invocation of a callback.
     */
    private Vector eventQueue = new Vector();
    
    /**
     * A list of notification handler that would listen to the same events.
     * Any event that sent to this handler will be sent to all the handlers
     * in this list as well. By doing this multiple notification handlers
     * can share a single event source. 
     */
    private Vector handlerChain = new Vector();

    /**
     * Construction of <code>NotificationHandler</code> causes a new
     * thread to be created and started. This thread serves
     * notification events that appear in the
     * <code>eventQueue</code>. Having a separate thread ensures that
     * the "internal" HyperCast threads (i.e. the adapter threads)
     * will not block due to application code.
     */
    public NotificationHandler()
    {
        this (null);
    }

    public NotificationHandler (final String threadName)
    {
        final String defaultThreadName = "Notification Handler";
        final String name = defaultThreadName + ((threadName != null) ? (": " + threadName) : "");
        new Thread (new InternalThread(), name).start();
    }
    /**
     * Add notification handler to share the event source. Be careful not to
     * make a loop of handlers, otherwise the program will run into deadlock
     * when some event happens. A check is made here to make sure a handler is
     * not added more than twice.
     * @param nh
     */
    public final void addNotificationHandler(final NotificationHandler nh){
    	if(nh==null)
    		return;
    	
    	for(int i=0;i< handlerChain.size(); i++){
    		if(handlerChain.elementAt(i) == nh)
    			return;
    	}
    	handlerChain.addElement(nh);
    }
    /**
     * Remove a handler from the chain
     * @see #addNotificationHandler(NotificationHandler)
     * @param nh
     */
    public final void removeNotificationHandler(final NotificationHandler nh){
    	for(int i=0;i< handlerChain.size(); i++){
    		if(handlerChain.elementAt(i) == nh){
    			handlerChain.removeElementAt(i);
    			return;
    		}
    	}
    }
    /**
     * Called by the components which support notification function, method 
     * is final so that user cannot extend this method and write code that 
     * blocks indefinitely when an event occurs.
     */
    public final void eventOccurred (final NOTIFICATION_EVENT event)
    {
        synchronized (eventQueue)
        {
            System.err.println ("eventOccurred: " + event.getClass().getName());

            eventQueue.addElement (event);
            eventQueue.notifyAll();
            
            /* send to all other handlers */
            for(int i=0;i<handlerChain.size();i++){
            	NotificationHandler nh = (NotificationHandler)handlerChain.elementAt(i);
				nh.eventOccurred(event);
            }
        }
    }
    
    /*
     * The NotificationHandler class is designed to be extended by the
     * user, however by design the user cannot change this method when
     * this class is extended. If the user could change this method,
     * the notification system would fail to work properly.
     *
     * The <code>run()</code> method of <code>Runnable</code> cannot
     * be private so a private class is created so that the
     * application programmer cannot call <code>run()</code>.
     */
    private class InternalThread
        implements Runnable
    {
        public final void run()
        {
            while (true)
            {
                NOTIFICATION_EVENT event = null;
                synchronized (eventQueue)
                {
                    while (eventQueue.size() == 0)
                    {
                        try { eventQueue.wait(); } catch (final InterruptedException ignored) { }
                    }
                    event = (NOTIFICATION_EVENT) eventQueue.elementAt(0);
                    eventQueue.removeElementAt(0);
                }

                if (event == null)
                    throw new HyperCastFatalRuntimeException ("event should not be null at this point");

                if (event instanceof INTERNAL_EVENT)
                {
                    internalHandler ((INTERNAL_EVENT) event);
                }
                else
                {
                    handler (event);
                }
            }
        }
    }

    /**
     * This method is not designed to be extended (otherwise it would not be internal).
     */
    private void internalHandler (final INTERNAL_EVENT event)
    {
/*
        if (event instanceof <SOME_INTERNAL_EVENT_TYPE>)
        {
            unblock_<SOME_INTERNAL_EVENT_TYPE> ((<SOME_INTERNAL_EVENT_TYPE>) event);
        }
        else
*/
        {
            IllegalArgumentException iae = new IllegalArgumentException ("Internal event of unknown type occurred");
            throw new HyperCastFatalRuntimeException (iae);
        }
    }

    /**
     * Determines the type of notification and dispatches the proper
     * handler. Note that each type of event could have a method
     * called "handle" and it could be called here as in:
     *
     * <code>
     * public void handler (final NOTIFICATION_EVENT event)
     * {
     *     event.handle (this);
     * }
     * </code>
     *
     * Each event could then call the appropriate method of the
     * notification handler from within its handle method. For example
     * for event type FOO:
     *
     * <code>
     * public void handle (NotificationHandler notificationHandler)
     * {
     *    notificationHandler.handle_FOO (this);
     * }
     * </code>
     *
     * In this way, Java's virtual method dispatch would do the
     * dispatching.
     *
     * This method is designed to be overriden by application code
     * that wants to have more control over event handling.
     */
    public void handler (final NOTIFICATION_EVENT event)
    {
        /*
         * Check event type and execute corresponding event handler.
         *
         * Expensive - frequently occurring events should appear
         * before infrequently occurring events.
         */
        if (event instanceof NAMING_EVENT)
        {
            unblock_NAMING_EVENT ((NAMING_EVENT) event);
            handle_NAMING_EVENT ((NAMING_EVENT) event);
        }
        else if (event instanceof NODE_ISSTABLE)
        {
            unblock_NODE_ISSTABLE ((NODE_ISSTABLE) event);
            handle_NODE_ISSTABLE ((NODE_ISSTABLE) event);
        }
        else if (event instanceof E2EACK_RECEIVED)
        {
            unblock_E2EACK_RECEIVED ((E2EACK_RECEIVED) event);
            handle_E2EACK_RECEIVED ((E2EACK_RECEIVED) event);
        }
        else if (event instanceof E2EPARTIALACK_RECEIVED)
        {
            unblock_E2EPARTIALACK_RECEIVED ((E2EPARTIALACK_RECEIVED) event);
            handle_E2EPARTIALACK_RECEIVED ((E2EPARTIALACK_RECEIVED) event);
        }
        else if (event instanceof MSG_WITH_E2EACK_SENT)
        {
            unblock_MSG_WITH_E2EACK_SENT ((MSG_WITH_E2EACK_SENT) event);
            handle_MSG_WITH_E2EACK_SENT ((MSG_WITH_E2EACK_SENT) event);
        }
        else if (event instanceof NEWSTREAM_ARRIVED_EVENT)
        {
            unblock_NEWSTREAM_ARRIVED_EVENT ((NEWSTREAM_ARRIVED_EVENT) event);
            handle_NEWSTREAM_ARRIVED_EVENT ((NEWSTREAM_ARRIVED_EVENT) event);
        }
        else if (event instanceof NODE_LEAVEOVERLAY)
        {
            unblock_NODE_LEAVEGROUP ((NODE_LEAVEOVERLAY) event);
            handle_NODE_LEAVEGROUP ((NODE_LEAVEOVERLAY) event);
        }
        else if (event instanceof NODE_LEAVEOVERLAY)
        {
            unblock_NODE_LEAVEOVERLAY ((NODE_LEAVEOVERLAY) event);
            handle_NODE_LEAVEOVERLAY ((NODE_LEAVEOVERLAY) event);
        }
        else if (event instanceof NODE_LOGICALADDRESSCHANGED)
        {
            unblock_NODE_LOGICALADDRESSCHANGED ((NODE_LOGICALADDRESSCHANGED) event);
            handle_NODE_LOGICALADDRESSCHANGED ((NODE_LOGICALADDRESSCHANGED) event);
        }
        else if (event instanceof NODE_NEIGHBORHOODCHANGED)
        {
            unblock_NODE_NEIGHBORHOODCHANGED ((NODE_NEIGHBORHOODCHANGED) event);
            handle_NODE_NEIGHBORHOODCHANGED ((NODE_NEIGHBORHOODCHANGED) event);
        }
        else if (event instanceof NOTIFICATION_EVENT)
        {
            final IllegalArgumentException iae = new IllegalArgumentException ("Abstract event type NOTIFICATION_EVENT seen in handler - concrete subtype should be handled instead.");
            throw new HyperCastFatalRuntimeException (iae);
        }
/*
TEMPLATE
        else if (event instanceof <TYPE>)
        {
            unblock_<TYPE> ((<TYPE>) event);
            handle_<TYPE> ((<TYPE>) event);
        }
*/
        else
        {
            final IllegalArgumentException iae = new IllegalArgumentException ("Event of unknown type occurred");
            throw new HyperCastFatalRuntimeException (iae);
        }
    }

    // NAMING_EVENT
    // NODE_ISSTABLE
    // E2EACK_RECEIVED
    // E2EPARTIALACK_RECEIVED
    // MSG_WITH_E2EACK_SENT
    // NEWSTREAM_ARRIVED_EVENT
    // NODE_LEAVEOVERLAY
    // NODE_LEAVEOVERLAY
    // NODE_LOGICALADDRESSCHANGED
    // NODE_NEIGHBORHOODCHANGED
    // NODE_SPT_CHILDRENCHANGED
    // NODE_SPT_PARENTCHANGED
    // NODE_SPT_ROOTCHANGED

    // NAMING_EVENT

    private final Object lock_NAMING_EVENT = new Object();
    private NAMING_EVENT event_NAMING_EVENT = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_NAMING_EVENT = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_NAMING_EVENT (final NAMING_EVENT event) { }

    private final void unblock_NAMING_EVENT (final NAMING_EVENT event)
    {
        synchronized (lock_NAMING_EVENT)
        {
            if (waitersFor_NAMING_EVENT == 0)
                return;

            event_NAMING_EVENT = event;
            lock_NAMING_EVENT.notifyAll();

            while (waitersFor_NAMING_EVENT > 0)
            {
                try { lock_NAMING_EVENT.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final NAMING_EVENT waitUntil_NAMING_EVENT()
    {
        return waitUntil_NAMING_EVENT (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final NAMING_EVENT waitUntil_NAMING_EVENT (final long timeout)
    {
        NAMING_EVENT result = null;
        synchronized (lock_NAMING_EVENT)
        {
            while (event_NAMING_EVENT != null)
            {
                try { lock_NAMING_EVENT.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_NAMING_EVENT == null)
            {
                ++waitersFor_NAMING_EVENT;
                while (event_NAMING_EVENT == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_NAMING_EVENT.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_NAMING_EVENT;
            }

            result = event_NAMING_EVENT;
            if (waitersFor_NAMING_EVENT == 0)
            {
                event_NAMING_EVENT = null;
                lock_NAMING_EVENT.notifyAll();
            }
        }
        return result;
    }

    // NODE_ISSTABLE

    private final Object lock_NODE_ISSTABLE = new Object();
    private NODE_ISSTABLE event_NODE_ISSTABLE = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_NODE_ISSTABLE = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_NODE_ISSTABLE (final NODE_ISSTABLE event) { }

    private final void unblock_NODE_ISSTABLE (final NODE_ISSTABLE event)
    {
        synchronized (lock_NODE_ISSTABLE)
        {
            if (waitersFor_NODE_ISSTABLE == 0)
                return;

            event_NODE_ISSTABLE = event;
            lock_NODE_ISSTABLE.notifyAll();

            while (waitersFor_NODE_ISSTABLE > 0)
            {
                try { lock_NODE_ISSTABLE.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final NODE_ISSTABLE waitUntil_NODE_ISSTABLE()
    {
        return waitUntil_NODE_ISSTABLE (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final NODE_ISSTABLE waitUntil_NODE_ISSTABLE (final long timeout)
    {
        NODE_ISSTABLE result = null;
        synchronized (lock_NODE_ISSTABLE)
        {
            while (event_NODE_ISSTABLE != null)
            {
                try { lock_NODE_ISSTABLE.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_NODE_ISSTABLE == null)
            {
                ++waitersFor_NODE_ISSTABLE;
                while (event_NODE_ISSTABLE == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_NODE_ISSTABLE.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_NODE_ISSTABLE;
            }

            result = event_NODE_ISSTABLE;
            if (waitersFor_NODE_ISSTABLE == 0)
            {
                event_NODE_ISSTABLE = null;
                lock_NODE_ISSTABLE.notifyAll();
            }
        }
        return result;
    }
    
    // E2EACK_RECEIVED

    private final Object lock_E2EACK_RECEIVED = new Object();
    private E2EACK_RECEIVED event_E2EACK_RECEIVED = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_E2EACK_RECEIVED = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_E2EACK_RECEIVED (final E2EACK_RECEIVED event) { }

    private final void unblock_E2EACK_RECEIVED (final E2EACK_RECEIVED event)
    {
        synchronized (lock_E2EACK_RECEIVED)
        {
            if (waitersFor_E2EACK_RECEIVED == 0)
                return;

            event_E2EACK_RECEIVED = event;
            lock_E2EACK_RECEIVED.notifyAll();

            while (waitersFor_E2EACK_RECEIVED > 0)
            {
                try { lock_E2EACK_RECEIVED.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final E2EACK_RECEIVED waitUntil_E2EACK_RECEIVED()
    {
        return waitUntil_E2EACK_RECEIVED (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final E2EACK_RECEIVED waitUntil_E2EACK_RECEIVED (final long timeout)
    {
        E2EACK_RECEIVED result = null;
        synchronized (lock_E2EACK_RECEIVED)
        {
            while (event_E2EACK_RECEIVED != null)
            {
                try { lock_E2EACK_RECEIVED.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_E2EACK_RECEIVED == null)
            {
                ++waitersFor_E2EACK_RECEIVED;
                while (event_E2EACK_RECEIVED == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_E2EACK_RECEIVED.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_E2EACK_RECEIVED;
            }

            result = event_E2EACK_RECEIVED;
            if (waitersFor_E2EACK_RECEIVED == 0)
            {
                event_E2EACK_RECEIVED = null;
                lock_E2EACK_RECEIVED.notifyAll();
            }
        }
        return result;
    }

    // E2EPARTIALACK_RECEIVED

    private final Object lock_E2EPARTIALACK_RECEIVED = new Object();
    private E2EPARTIALACK_RECEIVED event_E2EPARTIALACK_RECEIVED = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_E2EPARTIALACK_RECEIVED = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_E2EPARTIALACK_RECEIVED (final E2EPARTIALACK_RECEIVED event) { }

    private final void unblock_E2EPARTIALACK_RECEIVED (final E2EPARTIALACK_RECEIVED event)
    {
        synchronized (lock_E2EPARTIALACK_RECEIVED)
        {
            if (waitersFor_E2EPARTIALACK_RECEIVED == 0)
                return;

            event_E2EPARTIALACK_RECEIVED = event;
            lock_E2EPARTIALACK_RECEIVED.notifyAll();

            while (waitersFor_E2EPARTIALACK_RECEIVED > 0)
            {
                try { lock_E2EPARTIALACK_RECEIVED.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final E2EPARTIALACK_RECEIVED waitUntil_E2EPARTIALACK_RECEIVED()
    {
        return waitUntil_E2EPARTIALACK_RECEIVED (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final E2EPARTIALACK_RECEIVED waitUntil_E2EPARTIALACK_RECEIVED (final long timeout)
    {
        E2EPARTIALACK_RECEIVED result = null;
        synchronized (lock_E2EPARTIALACK_RECEIVED)
        {
            while (event_E2EPARTIALACK_RECEIVED != null)
            {
                try { lock_E2EPARTIALACK_RECEIVED.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_E2EPARTIALACK_RECEIVED == null)
            {
                ++waitersFor_E2EPARTIALACK_RECEIVED;
                while (event_E2EPARTIALACK_RECEIVED == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_E2EPARTIALACK_RECEIVED.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_E2EPARTIALACK_RECEIVED;
            }

            result = event_E2EPARTIALACK_RECEIVED;
            if (waitersFor_E2EPARTIALACK_RECEIVED == 0)
            {
                event_E2EPARTIALACK_RECEIVED = null;
                lock_E2EPARTIALACK_RECEIVED.notifyAll();
            }
        }
        return result;
    }

    // MSG_WITH_E2EACK_SENT

    private final Object lock_MSG_WITH_E2EACK_SENT = new Object();
    private MSG_WITH_E2EACK_SENT event_MSG_WITH_E2EACK_SENT = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_MSG_WITH_E2EACK_SENT = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_MSG_WITH_E2EACK_SENT (final MSG_WITH_E2EACK_SENT event) { }

    private final void unblock_MSG_WITH_E2EACK_SENT (final MSG_WITH_E2EACK_SENT event)
    {
        synchronized (lock_MSG_WITH_E2EACK_SENT)
        {
            if (waitersFor_MSG_WITH_E2EACK_SENT == 0)
                return;

            event_MSG_WITH_E2EACK_SENT = event;
            lock_MSG_WITH_E2EACK_SENT.notifyAll();

            while (waitersFor_MSG_WITH_E2EACK_SENT > 0)
            {
                try { lock_MSG_WITH_E2EACK_SENT.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final MSG_WITH_E2EACK_SENT waitUntil_MSG_WITH_E2EACK_SENT()
    {
        return waitUntil_MSG_WITH_E2EACK_SENT (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final MSG_WITH_E2EACK_SENT waitUntil_MSG_WITH_E2EACK_SENT (final long timeout)
    {
        MSG_WITH_E2EACK_SENT result = null;
        synchronized (lock_MSG_WITH_E2EACK_SENT)
        {
            while (event_MSG_WITH_E2EACK_SENT != null)
            {
                try { lock_MSG_WITH_E2EACK_SENT.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_MSG_WITH_E2EACK_SENT == null)
            {
                ++waitersFor_MSG_WITH_E2EACK_SENT;
                while (event_MSG_WITH_E2EACK_SENT == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_MSG_WITH_E2EACK_SENT.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_MSG_WITH_E2EACK_SENT;
            }

            result = event_MSG_WITH_E2EACK_SENT;
            if (waitersFor_MSG_WITH_E2EACK_SENT == 0)
            {
                event_MSG_WITH_E2EACK_SENT = null;
                lock_MSG_WITH_E2EACK_SENT.notifyAll();
            }
        }
        return result;
    }

    // NEWSTREAM_ARRIVED_EVENT

    private final Object lock_NEWSTREAM_ARRIVED_EVENT = new Object();
    private NEWSTREAM_ARRIVED_EVENT event_NEWSTREAM_ARRIVED_EVENT = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_NEWSTREAM_ARRIVED_EVENT = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_NEWSTREAM_ARRIVED_EVENT (final NEWSTREAM_ARRIVED_EVENT event) { }

    private final void unblock_NEWSTREAM_ARRIVED_EVENT (final NEWSTREAM_ARRIVED_EVENT event)
    {
        synchronized (lock_NEWSTREAM_ARRIVED_EVENT)
        {
            if (waitersFor_NEWSTREAM_ARRIVED_EVENT == 0)
                return;

            event_NEWSTREAM_ARRIVED_EVENT = event;
            lock_NEWSTREAM_ARRIVED_EVENT.notifyAll();

            while (waitersFor_NEWSTREAM_ARRIVED_EVENT > 0)
            {
                try { lock_NEWSTREAM_ARRIVED_EVENT.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final NEWSTREAM_ARRIVED_EVENT waitUntil_NEWSTREAM_ARRIVED_EVENT()
    {
        return waitUntil_NEWSTREAM_ARRIVED_EVENT (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final NEWSTREAM_ARRIVED_EVENT waitUntil_NEWSTREAM_ARRIVED_EVENT (final long timeout)
    {
        NEWSTREAM_ARRIVED_EVENT result = null;
        synchronized (lock_NEWSTREAM_ARRIVED_EVENT)
        {
            while (event_NEWSTREAM_ARRIVED_EVENT != null)
            {
                try { lock_NEWSTREAM_ARRIVED_EVENT.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_NEWSTREAM_ARRIVED_EVENT == null)
            {
                ++waitersFor_NEWSTREAM_ARRIVED_EVENT;
                while (event_NEWSTREAM_ARRIVED_EVENT == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_NEWSTREAM_ARRIVED_EVENT.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_NEWSTREAM_ARRIVED_EVENT;
            }

            result = event_NEWSTREAM_ARRIVED_EVENT;
            if (waitersFor_NEWSTREAM_ARRIVED_EVENT == 0)
            {
                event_NEWSTREAM_ARRIVED_EVENT = null;
                lock_NEWSTREAM_ARRIVED_EVENT.notifyAll();
            }
        }
        return result;
    }

    // NODE_LEAVEOVERLAY

    private final Object lock_NODE_LEAVEGROUP = new Object();
    private NODE_LEAVEOVERLAY event_NODE_LEAVEGROUP = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_NODE_LEAVEGROUP = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_NODE_LEAVEGROUP (final NODE_LEAVEOVERLAY event) { }

    private final void unblock_NODE_LEAVEGROUP (final NODE_LEAVEOVERLAY event)
    {
        synchronized (lock_NODE_LEAVEGROUP)
        {
            if (waitersFor_NODE_LEAVEGROUP == 0)
                return;

            event_NODE_LEAVEGROUP = event;
            lock_NODE_LEAVEGROUP.notifyAll();

            while (waitersFor_NODE_LEAVEGROUP > 0)
            {
                try { lock_NODE_LEAVEGROUP.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final NODE_LEAVEOVERLAY waitUntil_NODE_LEAVEGROUP()
    {
        return waitUntil_NODE_LEAVEGROUP (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final NODE_LEAVEOVERLAY waitUntil_NODE_LEAVEGROUP (final long timeout)
    {
        NODE_LEAVEOVERLAY result = null;
        synchronized (lock_NODE_LEAVEGROUP)
        {
            while (event_NODE_LEAVEGROUP != null)
            {
                try { lock_NODE_LEAVEGROUP.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_NODE_LEAVEGROUP == null)
            {
                ++waitersFor_NODE_LEAVEGROUP;
                while (event_NODE_LEAVEGROUP == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_NODE_LEAVEGROUP.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_NODE_LEAVEGROUP;
            }

            result = event_NODE_LEAVEGROUP;
            if (waitersFor_NODE_LEAVEGROUP == 0)
            {
                event_NODE_LEAVEGROUP = null;
                lock_NODE_LEAVEGROUP.notifyAll();
            }
        }
        return result;
    }

    // NODE_LEAVEOVERLAY

    private final Object lock_NODE_LEAVEOVERLAY = new Object();
    private NODE_LEAVEOVERLAY event_NODE_LEAVEOVERLAY = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_NODE_LEAVEOVERLAY = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_NODE_LEAVEOVERLAY (final NODE_LEAVEOVERLAY event) { }

    private final void unblock_NODE_LEAVEOVERLAY (final NODE_LEAVEOVERLAY event)
    {
        synchronized (lock_NODE_LEAVEOVERLAY)
        {
            if (waitersFor_NODE_LEAVEOVERLAY == 0)
                return;

            event_NODE_LEAVEOVERLAY = event;
            lock_NODE_LEAVEOVERLAY.notifyAll();

            while (waitersFor_NODE_LEAVEOVERLAY > 0)
            {
                try { lock_NODE_LEAVEOVERLAY.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final NODE_LEAVEOVERLAY waitUntil_NODE_LEAVEOVERLAY()
    {
        return waitUntil_NODE_LEAVEOVERLAY (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final NODE_LEAVEOVERLAY waitUntil_NODE_LEAVEOVERLAY (final long timeout)
    {
        NODE_LEAVEOVERLAY result = null;
        synchronized (lock_NODE_LEAVEOVERLAY)
        {
            while (event_NODE_LEAVEOVERLAY != null)
            {
                try { lock_NODE_LEAVEOVERLAY.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_NODE_LEAVEOVERLAY == null)
            {
                ++waitersFor_NODE_LEAVEOVERLAY;
                while (event_NODE_LEAVEOVERLAY == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_NODE_LEAVEOVERLAY.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_NODE_LEAVEOVERLAY;
            }

            result = event_NODE_LEAVEOVERLAY;
            if (waitersFor_NODE_LEAVEOVERLAY == 0)
            {
                event_NODE_LEAVEOVERLAY = null;
                lock_NODE_LEAVEOVERLAY.notifyAll();
            }
        }
        return result;
    }

    // NODE_LOGICALADDRESSCHANGED

    private final Object lock_NODE_LOGICALADDRESSCHANGED = new Object();
    private NODE_LOGICALADDRESSCHANGED event_NODE_LOGICALADDRESSCHANGED = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_NODE_LOGICALADDRESSCHANGED = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_NODE_LOGICALADDRESSCHANGED (final NODE_LOGICALADDRESSCHANGED event) { }

    private final void unblock_NODE_LOGICALADDRESSCHANGED (final NODE_LOGICALADDRESSCHANGED event)
    {
        synchronized (lock_NODE_LOGICALADDRESSCHANGED)
        {
            if (waitersFor_NODE_LOGICALADDRESSCHANGED == 0)
                return;

            event_NODE_LOGICALADDRESSCHANGED = event;
            lock_NODE_LOGICALADDRESSCHANGED.notifyAll();

            while (waitersFor_NODE_LOGICALADDRESSCHANGED > 0)
            {
                try { lock_NODE_LOGICALADDRESSCHANGED.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final NODE_LOGICALADDRESSCHANGED waitUntil_NODE_LOGICALADDRESSCHANGED()
    {
        return waitUntil_NODE_LOGICALADDRESSCHANGED (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final NODE_LOGICALADDRESSCHANGED waitUntil_NODE_LOGICALADDRESSCHANGED (final long timeout)
    {
        NODE_LOGICALADDRESSCHANGED result = null;
        synchronized (lock_NODE_LOGICALADDRESSCHANGED)
        {
            while (event_NODE_LOGICALADDRESSCHANGED != null)
            {
                try { lock_NODE_LOGICALADDRESSCHANGED.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_NODE_LOGICALADDRESSCHANGED == null)
            {
                ++waitersFor_NODE_LOGICALADDRESSCHANGED;
                while (event_NODE_LOGICALADDRESSCHANGED == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_NODE_LOGICALADDRESSCHANGED.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_NODE_LOGICALADDRESSCHANGED;
            }

            result = event_NODE_LOGICALADDRESSCHANGED;
            if (waitersFor_NODE_LOGICALADDRESSCHANGED == 0)
            {
                event_NODE_LOGICALADDRESSCHANGED = null;
                lock_NODE_LOGICALADDRESSCHANGED.notifyAll();
            }
        }
        return result;
    }

    // NODE_NEIGHBORHOODCHANGED

    private final Object lock_NODE_NEIGHBORHOODCHANGED = new Object();
    private NODE_NEIGHBORHOODCHANGED event_NODE_NEIGHBORHOODCHANGED = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_NODE_NEIGHBORHOODCHANGED = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_NODE_NEIGHBORHOODCHANGED (final NODE_NEIGHBORHOODCHANGED event) { }

    private final void unblock_NODE_NEIGHBORHOODCHANGED (final NODE_NEIGHBORHOODCHANGED event)
    {
        synchronized (lock_NODE_NEIGHBORHOODCHANGED)
        {
            if (waitersFor_NODE_NEIGHBORHOODCHANGED == 0)
                return;

            event_NODE_NEIGHBORHOODCHANGED = event;
            lock_NODE_NEIGHBORHOODCHANGED.notifyAll();

            while (waitersFor_NODE_NEIGHBORHOODCHANGED > 0)
            {
                try { lock_NODE_NEIGHBORHOODCHANGED.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final NODE_NEIGHBORHOODCHANGED waitUntil_NODE_NEIGHBORHOODCHANGED()
    {
        return waitUntil_NODE_NEIGHBORHOODCHANGED (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final NODE_NEIGHBORHOODCHANGED waitUntil_NODE_NEIGHBORHOODCHANGED (final long timeout)
    {
        NODE_NEIGHBORHOODCHANGED result = null;
        synchronized (lock_NODE_NEIGHBORHOODCHANGED)
        {
            while (event_NODE_NEIGHBORHOODCHANGED != null)
            {
                try { lock_NODE_NEIGHBORHOODCHANGED.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_NODE_NEIGHBORHOODCHANGED == null)
            {
                ++waitersFor_NODE_NEIGHBORHOODCHANGED;
                while (event_NODE_NEIGHBORHOODCHANGED == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_NODE_NEIGHBORHOODCHANGED.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_NODE_NEIGHBORHOODCHANGED;
            }

            result = event_NODE_NEIGHBORHOODCHANGED;
            if (waitersFor_NODE_NEIGHBORHOODCHANGED == 0)
            {
                event_NODE_NEIGHBORHOODCHANGED = null;
                lock_NODE_NEIGHBORHOODCHANGED.notifyAll();
            }
        }
        return result;
    }

   
    /*
     * TEMPLATE
     */

// NB: Java 1.5 has generics that would make this much easier...

/*
    // <TYPE>

    private final Object lock_<TYPE> = new Object();
    private <TYPE> event_<TYPE> = null; // blocking queue of length 1, could extend this to arbitrary length queue
    private int waitersFor_<TYPE> = 0;

    // Default event handler does nothing; designed to be overridden by users.
    public void handle_<TYPE> (final <TYPE> event) { }

    private final void unblock_<TYPE> (final <TYPE> event)
    {
        synchronized (lock_<TYPE>)
        {
            if (waitersFor_<TYPE> == 0)
                return;

            event_<TYPE> = event;
            lock_<TYPE>.notifyAll();

            while (waitersFor_<TYPE> > 0)
            {
                try { lock_<TYPE>.wait(); } catch (final InterruptedException ignored) {}
            }
        }
    }

    public final <TYPE> waitUntil_<TYPE>()
    {
        return waitUntil_<TYPE> (0); // block indefinitely
    }

    // @param timeout milliseconds
    public final <TYPE> waitUntil_<TYPE> (final long timeout)
    {
        <TYPE> result = null;
        synchronized (lock_<TYPE>)
        {
            while (event_<TYPE> != null)
            {
                try { lock_<TYPE>.wait(); } catch (final InterruptedException ignored) {}
            }

            final long start = System.currentTimeMillis();
            long elapsed = 0;

            if (event_<TYPE> == null)
            {
                ++waitersFor_<TYPE>;
                while (event_<TYPE> == null
                       &&
                       (timeout == 0 || elapsed < timeout))
                {
                    try { lock_<TYPE>.wait (timeout - elapsed); } catch (final InterruptedException ignored) {}
                    final long now = System.currentTimeMillis();
                    elapsed = now - start;
                }
                --waitersFor_<TYPE>;
            }

            result = event_<TYPE>;
            if (waitersFor_<TYPE> == 0)
            {
                event_<TYPE> = null;
                lock_<TYPE>.notifyAll();
            }
        }
        return result;
    }
*/
}
