/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.util.*;

/**
 * This class defines an object pool which stores predefined certain
 * types object instances. When a new object is needed, it can be 
 * obtained from this object pool instead of creating a new object on 
 * demand. The type of an object in the pool is identified by an object ID.
 * By default, there is no limit on the amount of object with certain 
 * type that can be stored in the pool.
 * 
 * @author HyperCast Team
 * @author Jianping Wang
 * @version Apr. 5, 2005
 */
public class ObjectPool
{
	/**
	 * Debug flag
	 */
	protected boolean debug = false;

	/**
	 * The hashtable that stores the maximum number of 
	 * instances of a certain type object in the pool.
     * (String (Obj id) -> Maximum number of Object instances with respect to the Obj id)
	 */
	private Hashtable objMaxNum = new Hashtable();

    /**
     * Hashtable that stores the Object instances in the pool
     * for reuse. It uses ObjId as hashcode.
     * (String (Obj id) -> Vector of Object instances with respect to the Obj id)
     */
    private Hashtable objPool = new Hashtable();

	/**
	 * Constructor.
	 */
	public ObjectPool()
	{
	}

	/**
	 * Add a vector of object instances with specified Object id to the pool.
	 */
	public void addObjVector(final Vector v, final short objid)
	{
		objPool.put(new Short(objid), v);
	}

	/**
	 * Get the vector of object instances with specified object id in the pool.
	 */
	public Vector getObjVector(final short objid)
	{
		return (Vector) objPool.get(new Short(objid));
	}
	
	/**
	 * Add a Object instance with specified object id to the pool.
	 */
	public void addObjInstance(final Object objInstance, final short objid)
	{
		Vector v = (Vector) objPool.get(new Short(objid));

		if (v == null)
		{
			//create a new vector to store the FSM instances with specified type.
			v = new Vector();
		}
		else
		{
			int maxNum = this.getMaxNumOfObjInstances(objid);
			if ((maxNum != -1) && (v.size() >= maxNum))
			{
				//The pool is full, the new instance is not added into the pool.
				return; 	
			}
		}
		
		v.addElement(objInstance);
		this.addObjVector(v, objid);
	}

	/**
	 * Return a Object instances with specified object id from the pool.
	 * If no required object instance is found, return null.
	 */
	public Object getObjInstance(final short objid)
	{
		Vector v = (Vector) objPool.get(new Short(objid));
		if ((v == null) || (v.size() == 0))
		{
			return null;
		}
		else
		{
			Object objInstance = (Object) v.elementAt(v.size()-1);
			v.removeElementAt(v.size()-1);
			this.addObjVector(v, objid);
			
			return objInstance;
		}
	}
	
	/**
	 * Set the maximum number of a certain type object instances that the pool
	 * can store.
	 */
	public void setMaxNumOfObjInstances(final short objid, final int num)
	{
		objMaxNum.put(new Short(objid), new Integer(num));
	}

	/**
	 * Get the maximum number of a certain type object instances that the pool
	 * can store.
	 */
	public int getMaxNumOfObjInstances(final short objid)
	{
		Integer val = (Integer) objMaxNum.get(new Short(objid));
		if (val == null)
			return -1;
		else
			return val.intValue();
	}
}
