/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.Pastry;

import hypercast.I_LogicalAddress;
import hypercast.I_Address;

import hypercast.Pastry.rice.pastry.NodeId;

import java.net.*;
import java.util.*;

/**
 *  A logical address for the Pastry protocol. This class implements
 *  the object adapter pattern, delegating Pastry_LogicalAddress
 *  method calls to an instance of NodeId from Pastry.
 * 
 *  @author HyperCast Team
 *  @author Greg Mattes
 *  @version May 2004 
 */
final public class Pastry_LogicalAddress
    implements I_LogicalAddress
{
    private NodeId nodeId;
    
    /** Constructs a Pastry NodeId object from information in a byte array */
    public Pastry_LogicalAddress (final byte [] bytes) {
        this.nodeId = new NodeId (bytes);
    }

    /** @see #Pastry_LogicalAddress(byte[]) */
    public Pastry_LogicalAddress (final NodeId nodeId) {
        this (nodeId.copy());
    }

    /**
     * Strings are of the form: 0x12345678, <0x12345678>, <12345678>, or 12345678
     *
     * The number of bytes is the number of digits of the string
     * divided by two (since each digit of the string encode a
     * nibble). The first two character, "0x," are not used.
     *
     * @see #Pastry_LogicalAddress(byte[])
     */
    public Pastry_LogicalAddress (final String nodeIdStringParam)
        throws NumberFormatException
    {
        final String nodeIdString;
        if (nodeIdStringParam.startsWith ("<0x"))
            nodeIdString = nodeIdStringParam.substring (3, nodeIdStringParam.length() - 1); // skip '>' at end
        else if (nodeIdStringParam.startsWith ("<"))
            nodeIdString = nodeIdStringParam.substring (1, nodeIdStringParam.length() - 1); // skip '>' at end
        else if (nodeIdStringParam.startsWith ("0x"))
            nodeIdString = nodeIdStringParam.substring (2);
        else
            nodeIdString = nodeIdStringParam;

        final byte [] bytes = new byte [nodeIdString.length() / 2]; // each char is a nibble
        for (int i = 0; i < bytes.length; ++i)
        {
            final int index = i * 2;
            final String highNibble = nodeIdString.substring (index, index + 1);
            final String lowNibble = nodeIdString.substring (index + 1, index + 2);
            bytes [bytes.length - i - 1] = (byte) (Byte.parseByte (highNibble, 16) << 4);
            bytes [bytes.length - i - 1] |= Byte.parseByte (lowNibble, 16);
        } 
        this.nodeId = new NodeId (bytes);
    }

    /** @see #Pastry_LogicalAddress(byte[]) */
    public Pastry_LogicalAddress (final Pastry_LogicalAddress other) {
        this (other.getNodeId().copy());
    }

    /** @return The underlying NodeId object. */
    public NodeId getNodeId() {
        return nodeId;
    }

    /** Delegates equality calculation to Pastry implementation. */
    public boolean equals (final Object object)
    {
        if (! (object instanceof Pastry_LogicalAddress))
            throw new IllegalArgumentException ("Object is not of type Pastry_LogicalAddress");

        final Pastry_LogicalAddress other = (Pastry_LogicalAddress) object;
        return this.nodeId.equals (other.getNodeId());
    }

    /** Delegates greaterThan calculation to Pastry implementation. */
    public boolean greaterThan (final I_Address address)
    {
        if (! (address instanceof Pastry_LogicalAddress))
            throw new IllegalArgumentException ("Object not of type Pastry_LogicalAddress");

        final Pastry_LogicalAddress other = (Pastry_LogicalAddress) address;
        return this.nodeId.compareTo (other.getNodeId()) > 0;
    }

    /** Delegates to Pastry implemenation */
    public static int getSizeStatic() {
        return NodeId.nodeIdBitLength / 8;
    }

    /** Delegates to Pastry implemenation */
    public int getSize() {
        return getSizeStatic();
    }

    /** Delegates to Pastry implemenation */
    public String toString () {
        return nodeId.toString();
    }

    /** Delegates to Pastry implemenation */
    public byte[] toByteArray() {
        return nodeId.copy();
    }

    /** Delegates to Pastry implemenation */
    public int hashCode() {
        return nodeId.hashCode();
    }
}
