/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.util.*;

/**
 * This class defines the route record of an OL_Message. It stores a sequence
 * of logical addresses of the nodes this message passed through.
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, Feb 27, 2004
 */
public class RouteRecord extends Extension
{

	/** the route record */
	Vector records;

	/** Hash table which stores the same sequence of logical addresses. 
	 * It is used to speed up the searching speed */
	Hashtable hash;
	 
	/**
	 * Constructs a empty route record.
	 */
	public RouteRecord()
	{
		records=new Vector();
		hash=new Hashtable();
	}

	/**
	 * Constructs a route record by cloning the given route record instance.
	 */
	public RouteRecord(RouteRecord old_rr)
	{
		this.records=new Vector();
		for (int i=0; i<old_rr.records.size(); i++)
		{
			this.records.addElement(old_rr.records.elementAt(i));
		}
		
		this.hash=new Hashtable();
		Enumeration keys= old_rr.hash.keys();
		while(keys.hasMoreElements())
		{
			Object key=keys.nextElement();
			this.hash.put(key, old_rr.hash.get(key));
		}
	}


	/**
	 * Constructs the route record from an byte array.
	 * @param contents		the overlay message
	 * @param node		the node used to construct logical address
	 */
	public RouteRecord(byte[] contents, I_Node node)
	{

		records=new Vector();
		hash=new Hashtable();

		/*
		 * Reconstruct the sequence of logical address from contents (byte array).
		 */
		int index = 0;
		while(index<contents.length){
			I_LogicalAddress la=node.createLogicalAddress(contents, index);
			records.addElement(la);
			hash.put(la, la);
			index+=la.getSize();
		}
		
	}

	/**
	 * Return a new RouteRecord extension instance by cloning this extension.
	 */
	public Extension Clone()
	{
		return new RouteRecord(this);
	}

	/**
	 * Returns the type of extension.
	 */
	public byte getExtensionType() {
		return Extension.ROUTE_RECORD;
	}


	/*
	 * This method provides a way for all extensions to get the length of 
	 * the byte array of an extension without actually executing toByteArray().
	 */
	/**
	 * Returns the length of the byte array of this extension.
	 */
	public int getSize(){
		if (records.size() == 0)
		{
			return 0;
		}
		
		I_LogicalAddress la = (I_LogicalAddress)records.elementAt(0);
		return records.size()*la.getSize();
	}


	/**
	 * Inserts a logical address to the route record
	 * @param la	the address to be inserted
         * @deprecated use insertLogicalAddress
	 */
	public void insertLA(I_LogicalAddress la, int max_size)
    {
        insertLogicalAddress (la, max_size);
    }


	/**
	 * Inserts a logical address to the route record
	 * @param la	the address to be inserted
	 */
	public void insertLogicalAddress(I_LogicalAddress la, int max_size){
		if(la!=null)
		{
			records.addElement(la);
			hash.put(la,la);

			/*
			 * If the size of route record is greater than the 
			 * limit of route record size, remove the logical address
			 * at position 0, which is the logical address existing for
			 * the longest time, until the size of the vector is less
			 * than or equal to the limit.
			 */
			while (records.size() > max_size)
			{
				
				I_LogicalAddress lao = (I_LogicalAddress)records.elementAt(0);
				records.removeElementAt(0);
				hash.remove(lao);
			}
		}
	}

	
	/**
	 * Returns the byte array of the extension header.
	 */
	public byte[] toByteArray(){
		byte []payload;
		int pl_length=0;
		
		for(int i=0; i<records.size(); i++){
			I_LogicalAddress la=(I_LogicalAddress)records.elementAt(i);
			pl_length += la.getSize();
		}
		
		payload = new byte[pl_length];
		int index=0;
		
		for(int i=0; i<records.size(); i++){
			I_LogicalAddress la=(I_LogicalAddress)records.elementAt(i);
			System.arraycopy(la.toByteArray(), 0, payload, index, la.getSize());
			index += la.getSize();
		}
		
		return payload;
			
	}

	/**
	 * Checks if the specified logical address is contained in
	 * this RouteRecord.
	 *
	 * @param query 	The logical address for which the search is being done.
	 * @return 			true if the logical address is contained in the record,
	 * 					false otherwise.
	 */

	public boolean contains (I_LogicalAddress query)
	{
		if (query == null)
		{
			throw new HyperCastFatalRuntimeException ("Cannot call with a null Logical Address"); 
		}
		 
		I_LogicalAddress result = null;
		if (records.size () > 0)
		{
			result = (I_LogicalAddress) hash.get (query);
		}
		
		return ((result != null) && (result.equals (query)));
	}
}
