/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.SPT;

import hypercast.*;
import hypercast.events.*;
import hypercast.util.XmlUtil;

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.apache.xpath.*;

/** This class is a Spanning Tree node.  
 * 
 * @author HyperCast Team
 * @author Guangyu Dong
 * @version 0.1, May 25, 2003 
 */

public class SPT_Node implements I_AdapterCallback, I_Node
{
	/** It may be used to generate the node's LogicalAddress. */
	protected static Random rand = new Random();

	/** The list of neighbors */
	protected SPT_Neighborhood neighbors;
	
	/** Routing Table */
	protected SPT_RoutingTable routing_table;
	
	/** Source of constants & error/log message display */
	protected HyperCastConfig config;

	/** Multicast adapter (used to transfer protocol message to neighbors) */
	public I_MulticastAdapter m_adapter;

	/** True if joinOverlay has been called without matching leaveOverlay. */
	protected boolean joined;
	
	/** The property prefix, it can be changed by constructors. */
	//final static String PROPERTY_PROTO_PREFIX = "SPT";

	/** Beacon timer. */
	private static final int BEACON_TIMER_INDEX = 0;
	protected long BEACON_TIME;
	final static String BEACON_TIME_PROPERTY_NAME = "BeaconTime";
	//final static long BEACON_TIME_DEFAULT = 1000; // milliseconds
	
	final static String PROP_NAME_VERIFICATION = "Verification";
	final static String PROP_NAME_INFO_DISP_PERIOD = "InfoDispPeriod";

	
	/** String for the default id. */
	//private static final String SPT_ID_DEFAULT = "RANDOM";
	/** Name of property controlling id. */
	private static final String SPT_ID_PROPERTY_NAME = "Coords";

	/** The default range value of x or y coordinates. It is used only 
	 * when coordinates are randomly specified*/
	//private static final int RANDOM_COORDS_DEFAULT_RANGE = 10000;

	/** Used the calculate the burst-message-load. */
	private int messagesReceivedInLastHeartbeat;
	/** Used the calculate the burst-message-load. */
	private long timeOfLastHeartbeat;

	/** Used the calculate the burst-message-load. */
	private int messagesSentInLastHeartbeat;
	/** Max burst seen since last joinOverlay. */
	private double maxReceiveRateInAHeartbeat;

	/** Max burst seen since last joinOverlay. */
	private double maxSendRateInAHeartbeat;

	/** Time of the call to joinGroup() */
	private long startTime;

	/** Time of last change to the neighborhood.  Used to determine how long a node has been stable. */
	private long lastChangeToNeighborhood;

	/** Time of the call to leaveOverlay() */
	private long stopTime;

	/** Used to coordinate resetting and clearing of the heartbeat timer.*/
	private boolean timerCleared;

	/** Object used to prevent problems while starting and stopping the node. */
	Object startStopLock;

	/** The protocol subfield value which represents the used rendezvous and will be put 
	 * in the Proto_sub field of a SPT_Message. It will be reset in the constructors 
	 * of SPT_Node class.
	 */
	byte Proto_Sub = I_Node.PROTOSUB_NOSPEC;

	/** Notification handler. When interesting events happen, this node 
	 * reports the events to the notification handler if it is not null.
	 */
	protected NotificationHandler n_handler = null;
	
	/* for debuging purpose */
	/** the period to dispaly neighborhood information */	
	protected long INFO_DISP_PERIOD = 0;
	/** the timer index for the period */
	protected int INFO_DISP_TIMER_INDEX = 2;

	/** Check mode on the previous hop of OL messages.
	 */
	protected String checkmode;

	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	StatsProcessor statsPro;
	
	/**
	 * The statistics name for this object (as an I_Stats instance). Its default
	 * name is defined in the socket configuration schema file. It can be assigned
	 * by any other object through method setStattisticsName defined in the 
	 * I_Stats interface.
	 */
	private String statisticsName;

	
	/** 
	 * for experimental purpose: if StopBeacon is true, no beacon message is sent 
	 * neighbors.NeighborhoodFixed has to be true if we still want to maintain the topology
	 * */
	boolean StopBeacon = false;
	
	/** The physical and logical address of this node.
	 *  Since this object might be refered by several different place,
	 *  when the logical address need to be changed, try to use SPT_LogicalAddress.changeLogicalAddress()
	 *  instead of making another new address.
	 */
	private SPT_AddressPair myAddressPair;
	
	
	/** Constructor.
	 * Read configuration from the property file
	 * @param c		overlay socket config object
	 * @param m		node adapter
	 * @param prefix	the property name prefix
	 * @throws IllegalArgumentException if properties are incorrectly formatted.
	 */
	public SPT_Node(HyperCastConfig c, I_MulticastAdapter m, String prefix) {

		init(c, m, null, prefix);
	}
	

	/**
	 * Constructor
	 * @param c		overlay socket config object
	 * @param m		node adapter
	 * @param la	logical address
	 * @param prefix	property name prefix
	 * @throws IllegalArgumentException if properties are incorrectly formatted.
	 */
	
	public SPT_Node(
			HyperCastConfig c,
			I_MulticastAdapter m,
			SPT_LogicalAddress la,
			String prefix){
		init(c, m, la, prefix);
		
	}
	private void init(
		HyperCastConfig c,
		I_MulticastAdapter m,
		SPT_LogicalAddress la,
		String prefix) {

				config = c;
		m_adapter = m;
		/** spt node use multicast to exchange protocol message */
		m.setCallback(this);

		
		
		if(la==null)
			la=createLogicalAddress(config, prefix);
		
		myAddressPair = new SPT_AddressPair(m_adapter.createPhysicalAddress(), la);
		
		checkmode =
			//config.getStringProperty(prefix + PROP_NAME_VERIFICATION, "neighborcheck");
			config.getTextAttribute(XmlUtil.createXPath(prefix + PROP_NAME_VERIFICATION));
		
/*
		BEACON_TIME =
			config.getPositiveLongProperty(
				prefix + BEACON_TIME_PROPERTY_NAME,
				BEACON_TIME_DEFAULT);
*/
		BEACON_TIME =
			config.getPositiveLongAttribute
				(XmlUtil.createXPath(prefix + BEACON_TIME_PROPERTY_NAME));


		//INFO_DISP_PERIOD = config.getIntProperty(prefix +"."+ PROP_NAME_INFO_DISP_PERIOD);
		INFO_DISP_PERIOD = config.getIntAttribute
				(XmlUtil.createXPath(prefix +PROP_NAME_INFO_DISP_PERIOD));
		
		/*
		 * Get the statistics name defined in the configuration file for this I_Stats instance.
		 */
		statisticsName = config.getTextAttribute (XmlUtil.createXPath(prefix+"StatName"));

		InitStatisticsStructure();
		//neighbors = new SPT_Neighborhood(config, prefix, this, m_adapter, myAddressPair, ns);
		neighbors = new SPT_Neighborhood(config, prefix, this, m_adapter, myAddressPair, statsPro);

		maxReceiveRateInAHeartbeat = 0;
		maxSendRateInAHeartbeat = 0;
		timeOfLastHeartbeat = -1;
		messagesSentInLastHeartbeat = 0;
		messagesReceivedInLastHeartbeat = 0;

		startStopLock = new Object();

		joined = false;

		routing_table = new SPT_RoutingTable(config, prefix, m_adapter, neighbors);
		
		
	}

	/** Creates the node's logical address given properties. 
	 * @return a new logical address
	 * @throws IllegalArgumentException if properties are incorrectly formatted.
	 */
	protected SPT_LogicalAddress createLogicalAddress(HyperCastConfig config, String prefix) {
		//prefix+="/";
		
		String propString =
			//config.getStringProperty(prefix + SPT_ID_PROPERTY_NAME, SPT_ID_DEFAULT);
			config.getTextAttribute
				(XmlUtil.createXPath(prefix + SPT_ID_PROPERTY_NAME));

		if (propString.startsWith("RANDOM")) {

/*
			int maxID =
				config.getNonNegativeIntProperty(
						prefix + "RANDOM.base",
					RANDOM_COORDS_DEFAULT_RANGE);
*/
			int maxID =
				config.getNonNegativeIntAttribute
					(XmlUtil.createXPath(prefix+SPT_ID_PROPERTY_NAME+"/RANDOM/base"));


			int id = Math.abs(rand.nextInt()) % maxID + 1;

			//to prevent zero value
			return new SPT_LogicalAddress(id);

		} else if (propString.equals("USE_IP")) {

			I_PhysicalAddress pa = m_adapter.createPhysicalAddress();
			if (pa instanceof INETv4OnePort)
			{
				return new SPT_LogicalAddress((INETv4OnePort) pa);
			}
			else
			{
				throw new IllegalArgumentException(
					SPT_ID_PROPERTY_NAME
						+ " property = USE_IP, but the adapter is not using type IP_Address!!!");
			}

		}
		else if (propString.equals("FIXED")) {
			String coords =
				//config.getStringProperty(prefix + "FIXED.coordinate", "");
				config.getTextAttribute
					(XmlUtil.createXPath(prefix+SPT_ID_PROPERTY_NAME+"/FIXED/coordinate"));

			try {
				return new SPT_LogicalAddress(coords);
			} catch (Exception e) {
				throw new IllegalArgumentException(
					prefix
						+ SPT_ID_PROPERTY_NAME
						+ " property caused an exception -- probably poorly formated.");
			}
		} else {
                    throw new HyperCastFatalRuntimeException ("Unknown method of creating logical address is specified.");
		}
	}

	/** Joins the overlay multicast group.
	 * @see I_Node#joinOverlay()
	 */
	public void joinOverlay() {
		if (joined)
			throw new IllegalStateException("SPT_Node.joinOverlay() called twice without leaveOverlay()!");

		neighbors.clear();

		synchronized (startStopLock) {
			m_adapter.Start();
			m_adapter.startTimer();
		}

		// random time prevents synchronisity if large numbers are started at the same time.
		synchronized (this) {
			m_adapter.setTimer(
				new Integer(BEACON_TIMER_INDEX),
				(long) (rand.nextDouble() * BEACON_TIME));

			//for displaying node states
			if (INFO_DISP_PERIOD > 0)
				m_adapter.setTimer(
					new Integer(INFO_DISP_TIMER_INDEX),
					(long) (rand.nextDouble() * INFO_DISP_PERIOD));

			timerCleared = false;
			notifyAll();
		}

		startTime = m_adapter.getCurrentTime();
		lastChangeToNeighborhood = startTime;
		stopTime = -1;

		config.log.println(
			"SPT_Node started: "
				+ this.getLogicalAddress().toString()
				+ " at time "
				+ lastChangeToNeighborhood);
		config.log.flush();

		joined = true;

	}

	/** Leaves the overlay multicast group.
	 * @see I_Node#leaveOverlay
	 */
	public synchronized void leaveOverlay() {
		if (!joined)
			throw new IllegalStateException("SPT_Node.leaveOverlay() called without calling joinGroup!");

		m_adapter.clearTimer(new Integer(BEACON_TIMER_INDEX));
		m_adapter.clearTimer(new Integer(INFO_DISP_TIMER_INDEX));

		timerCleared = true;
		notifyAll();
		sendGoodbye();
		
		stopTime = m_adapter.getCurrentTime();

		// should not be synchronized while stopping adapter
		m_adapter.Stop();
		m_adapter.stopTimer();
		
		neighbors.clear();

		config.log.println(
			"SPT_Node left group: "
				+ getLogicalAddress().toString()
				+ " at time "
				+ stopTime);
		// joined should only be set after adapter is stopped.
		joined = false;

		if (n_handler != null) {
			NODE_LEAVEOVERLAY ne = new NODE_LEAVEOVERLAY(m_adapter.getCurrentTime(), null);
			n_handler.eventOccurred(ne);
		}
	}

	/** Creates a protocol message from a byte[].  
	 * Required in order to implement I_AdapterCallback.
	 * @see I_AdapterCallback#restoreMessage
	 */
	public I_Message restoreMessage(byte[] buffer, int[] startValidBytes, int endValidBytes) {
		return SPT_Message.restoreMessage(
			buffer,
			startValidBytes,
			endValidBytes,
			m_adapter,
			config.getOverlayHash());
	}

	/** @return the physical address/logical address pair of this node. 
	*/
	public synchronized I_AddressPair getAddressPair() {
		return neighbors.getMyAddressPair();
	}
	/**
	 * @return	the current logical address
	 */
	public I_LogicalAddress getLogicalAddress() {
		return getAddressPair().getLogicalAddress();
	}
	/**
	 * @return	the current SPT logical address
	 */
	public SPT_LogicalAddress getSPTLogicalAddress() {
		return (SPT_LogicalAddress) getAddressPair().getLogicalAddress();
	}
	/**
	 * Transfer a single address pair to an array with one element
	 * @param o		the address pair	
	 * @return		the resulting array containing this address pair
	 */
	private I_AddressPair[] singleByteArray(I_AddressPair o) {
		I_AddressPair[] result = new I_AddressPair[1];
		result[0] = o;
		return result;
	}
	/** 
	 *  @see I_Node#getParent(I_LogicalAddress)
	 */
	public synchronized I_AddressPair[] getParent(I_LogicalAddress root) {
		
		if (!joined)
			throw new IllegalStateException("Must joinGroup before calling getParent!");

		
		if (root.equals(getLogicalAddress())){
			//I am the destination
			return singleByteArray(neighbors.getMyAddressPair());
		}
		
		if (root.equals(neighbors.getRootLogicalAddress())){
			//destination is root
			//send to parent
			return singleByteArray(neighbors.getParentAddressPair());
		}
		SPT_RoutingEntry re =
			routing_table.getRoutingEntry(
				(SPT_LogicalAddress) root
				);
		

		I_AddressPair nextHop = null;

		if (neighbors.containsNeighbor((SPT_LogicalAddress) root)) {
			/* it's a direct neighbor */
			SPT_NeighborhoodEntry ne=neighbors.getNeighborEntry((SPT_LogicalAddress) root);
			nextHop = ne.getAddressPair();

		} else {
			
			if (re != null) {
				SPT_NeighborhoodEntry ne=neighbors.getNeighborEntry(re.getNextHop());
				nextHop = ne.getAddressPair();
				re.update();
			}
		}

		//on-demand unicast path maintenance
		//added Jan 28, 2004
		if (nextHop == null) {
			//not routable
			//send route request to all neighbors
			//return all neighbors as nexthop
			this.sendRouteRequest((SPT_LogicalAddress) root);
			return this.getAllNeighbors();
		} else{
			//routable
			//return nexthop
			return singleByteArray(nextHop);
		}
	}

	/** 
	 * @see I_Node#getChildren(I_LogicalAddress)
	 */
	public synchronized I_AddressPair[] getChildren(I_LogicalAddress root) {

		if (!joined)
			throw new IllegalStateException("Must joinGroup before calling getChildren!");
		/*
		 * Message is broadcasted to all neighbors
		 */
		return this.getAllNeighbors();

	}

	/** 
	 * @return the AddressPairs of all the neighbors
	 * @see I_Node#getAllNeighbors 
	 */
	public synchronized I_AddressPair[] getAllNeighbors() {
		if (!joined)
			throw new IllegalStateException("Must joinGroup before calling getAllNeighbors!");

		
		return neighbors.getAllNeighbors();
	}

	/** 
	 * Constructs a SPT_LogicalAddress from a byte array.
	 * @see I_Node#createLogicalAddress(byte[], int)
	 */
	public I_LogicalAddress createLogicalAddress(byte[] array, int offset) {
		return new SPT_LogicalAddress(array, offset);
	}

	/** This function, a member of I_AdapterCallback, is called for
	 * the heartbeat timer. When the timer occurs, the node sends 
	 * Hello messages to all of its neighbors.  At the same time, the
	 * node checks if any of its neighbors has timed out.
	 * @see I_AdapterCallback#timerExpired
	 */
	public synchronized void timerExpired(int timerIndex) {
		if (timerIndex == BEACON_TIMER_INDEX) {
			// NB: Getting the time must be done before it is reset
			// because, otherwise, the thread could become unscheduled
			// between the call and setting the timer - causing the next
			// heartbeatDuration to be less than the SLOW_BEACON_TIME.
			long timeOfThisHeartbeat = m_adapter.getCurrentTime();

			//modified in 02/04/04 to fit in Hypercast
			//m_adapter.setTimer(BEACON_TIMER_INDEX, BEACON_TIME);
			m_adapter.setTimer(new Integer(BEACON_TIMER_INDEX), BEACON_TIME);

			// wake all threads wait()ing because timer was not set.
			notifyAll();

			//clear timeout neighbors
			handleTimeoutTimer(timeOfThisHeartbeat);

			// send Hello messages to all neighbors
			sendHello();

			//for statistics
			double heartbeatDuration;
			if (timeOfLastHeartbeat > 0)
				heartbeatDuration = (timeOfThisHeartbeat - timeOfLastHeartbeat) / 1000.0;
			else
				heartbeatDuration = BEACON_TIME / 1000.0;

			double receiveRate = messagesReceivedInLastHeartbeat / heartbeatDuration;
			double sendRate = messagesSentInLastHeartbeat / heartbeatDuration;

			maxReceiveRateInAHeartbeat = Math.max(maxReceiveRateInAHeartbeat, receiveRate);
			maxSendRateInAHeartbeat = Math.max(maxSendRateInAHeartbeat, sendRate);

			timeOfLastHeartbeat = timeOfThisHeartbeat;
			messagesSentInLastHeartbeat = 0;
			messagesReceivedInLastHeartbeat = 0;
			

		}
		else if (timerIndex == INFO_DISP_TIMER_INDEX) {
			/* for debuging: displaying neighborhood information */
			if(INFO_DISP_PERIOD>0)
				m_adapter.setTimer(new Integer(INFO_DISP_TIMER_INDEX), INFO_DISP_PERIOD);

			System.out.println(peerInfo());
			
		} else {
			throw new IllegalArgumentException(
				"Timer index " + timerIndex + " unexpected in SPT_Node");
		}

	}
	
	/** display the information about the neighborhood */
	private String peerInfo() {
		String result = "\n";

		result
			+= ("Node "
				+ this.getLogicalAddress()
				+ " parent="
				+ neighbors.getParentLogicalAddress()
				+ " root="
				+ neighbors.getRootLogicalAddress()
				+ " cost="
				+ neighbors.getCost())
			+ "\n";

		SPT_AddressPair children[] = neighbors.getChildren();
		if (children != null && children.length > 0) {
			result += ("- Children:");
			for (int i = 0; i < children.length; i++) {
				result += (" " + children[i].getSPTLogicalAddress());
			}
			result += "\n";
		}

		for (Enumeration e = neighbors.adjacency_table.elements(); e.hasMoreElements();) {
			SPT_NeighborhoodEntry entry = (SPT_NeighborhoodEntry) e.nextElement();

			long ts = m_adapter.getCurrentTime();

			result
				+= ("- Peer "
					+ entry.getLogicalAddress()
					+ " ACK= "
					+ entry.getLinkQuality().getPingNum()
					+ "/"
					+ entry.getLinkQuality().getReversedPingNum())
					+ " "
					+ entry.getPathMetric()
					+ " "
					+ (neighbors.containsNeighbor(entry.getLogicalAddress())?"neighbor":"")
				+ "\n";

			//sendPing(entry.getPA(), ts);
		}

		return result;
	}

	/** 
	 * Remove neighbors that have expired.
	 * @param currentTime	current time
	 * 
	 */
	protected void handleTimeoutTimer(long currentTime) {
		if(neighbors.clearExpired()){
			lastChangeToNeighborhood = m_adapter.getCurrentTime();
			
			/* Report event NEIGHBORHOODCHANGED to notification handler. */
			if ((n_handler != null)) {
				NODE_NEIGHBORHOODCHANGED ne =
					new NODE_NEIGHBORHOODCHANGED(m_adapter.getCurrentTime(), null);
				n_handler.eventOccurred(ne);
			}
		}
		
		routing_table.clearExpired();
	}

	/** Sends a Hello message(beacon) to all of the node's neighbors. */
	protected final void sendHello() {

		/** beacon not sent if StopBeaon is set to true */
		if (StopBeacon)
			return;

		messagesSentInLastHeartbeat++;

		m_adapter.sendMulticastMessage(createSPTMessage(SPT_Message.Hello, null, null));
	}

	/** 
	 * This function, a member of I_AdapterCallback, is called whenever a
	 * protocol message arrives.  The function determines the proper
	 * reaction to each message type.  
	 * @see I_AdapterCallback#messageArrivedFromAdapter
	 */
	public synchronized void messageArrivedFromAdapter(I_Message a) {
		if (!(a instanceof SPT_Message))
			throw new IllegalArgumentException("SPT_Node.messageArrivedFromAdapter() passed non-SPT_Message I_Message.");

		SPT_Message m = (SPT_Message) a;
		
		//m.myLA=neighbors.getLogicalAddress();
		
		if (m.getOverlayHash() != config.getOverlayHash()) {
			/* not from the same ovelray */
			//  config.log.println("Received message with improper Group identifier: " + m.getOverlayHash() + " != " + config.getOverlayHash());
			return; // ignore the message
		}
		
		if (neighbors.getMyAddressPair().equals(m.getSrc())){
			/* sent by myself */
			return; 
		}
		
		if (neighbors.getMyAddressPair().equals(m.getSrc()))
			return; //coming from myself
		
		messagesReceivedInLastHeartbeat++;

		/* judge the type */
		switch (m.getType()) {

			case SPT_Message.Hello :
				handleHello(m);
				break;

			case SPT_Message.Goodbye :
				handleGoodbye(m);
				break;

			case SPT_Message.RouteReply :
				handleRouteReply(m);
				break;

			case SPT_Message.RouteRequest :
				handleRouteRequest(m);
				break;

			default :
				config.err.println("SPT_Node: message with unknown type arrived: " + m.getType());
		}

	}
	/**
	 * Handle the RouteRequest message
	 * @param m		the message
	 */
	protected void handleRouteRequest(SPT_Message m) {
		//is it not coming from a neighbor?
		if (!neighbors.containsNeighbor(m.getSrcLA()))
			return;
		//is the destination the root?
		if (m.getPathDst().equals(neighbors.getRootLogicalAddress()))
			return;
		
		SPT_RoutingEntry re =
			routing_table.getRoutingEntry(
					m.getPathDst()
					);
		
		if (m.getPathDst().equals(getLogicalAddress())) {
			//I am the destination
			sendRouteReply((SPT_LogicalAddress) getLogicalAddress(), m.getPathDst());
			return;
		}
		
		if (neighbors.containsNeighbor(m.getPathDst())) {
			//destination is my neighbor
			
			sendRouteReply(m.getPathDst(), m.getPathDst());
			return;
		}

		if (re != null) {
			sendRouteReply(re.getNextHop(), m.getPathDst());
	
		}

	}

	/**
	 * Handle the RouteReply message
	 * @param m		the message
	 */
	protected void handleRouteReply(SPT_Message m) {
		//is it not coming from a neighbor?
		if (!neighbors.containsNeighbor(m.getSrcLA()))
			return;
		//is the destination the root?
		if (m.getPathDst().equals(neighbors.getRootLogicalAddress()))
			return;
		
		//process the message
		SPT_RoutingEntry re =
			routing_table.getRoutingEntry(m.getSrcLA());

		//I am the next hop?
		if (m.getNextHop().equals(getLogicalAddress())) {
			if (re != null && re.getNextHop().equals(m.getNextHop()))
				routing_table.removeRouting(re.getDest());
			return;
		}
		//I am the destination?
		if (m.getPathDst().equals(getLogicalAddress()))
			return;

		//is the destnation my neighbor?
		if (neighbors.containsNeighbor(m.getPathDst()))
			return;

		//process the message

		if (re == null) {
			//add an entry
			routing_table.updateRoutingEntry(m.getPathDst(), m.getSrcLA());
			sendRouteReply(m.getSrcLA(), m.getPathDst());
		} else {
			routing_table.updateRoutingEntry(
					m.getPathDst(),
					m.getSrcLA());
			if (re.getNextHop().equals(m.getSrcLA())) {
				//the same next hop
				//dont forward RouteReply
				routing_table.updateRoutingEntry(
					m.getPathDst(),
					m.getSrcLA());
				
				return;
			}
			
			//send RouteReply
			sendRouteReply(m.getSrcLA(), m.getPathDst());

		}

	}
	/**
	 * Handle the Goodbye message
	 * Removes the sender of the message from the neighborhood.
	 * @param m		the message
	 *   
	 */
	protected void handleGoodbye(SPT_Message m) {

		if (neighbors.containsAdjacentNode(m.getSrcLA()))
			neighbors.removeAdjacentNode(m.getSrcLA());

		//Report event NEIGHBORHOODCHANGED to notification handler.
		if (n_handler != null) {
			NODE_NEIGHBORHOODCHANGED ne =
				new NODE_NEIGHBORHOODCHANGED(m_adapter.getCurrentTime(), null);
			n_handler.eventOccurred(ne);
		}
	}

	/**
	 * Handel Hello message
	 * @param m		the message
	 */
	protected void handleHello(SPT_Message m) {
		//System.out.println("Beacon received from "+m.getSrcLA()+" PathMetric="+m.getPathMetric());

		long now = m_adapter.getCurrentTime();
		
		if(neighbors.update(m)){
			lastChangeToNeighborhood = now;

			//Report event LACHANGED to notification handler.
			if (n_handler != null) {
				NODE_NEIGHBORHOODCHANGED ne =
					new NODE_NEIGHBORHOODCHANGED(m_adapter.getCurrentTime(), null);
				n_handler.eventOccurred(ne);
			}
		}
	

	}

	/** Sends a Goodbye message to all of the node's neighbors. */
	protected final void sendGoodbye() {
		messagesSentInLastHeartbeat++;

		m_adapter.sendMulticastMessage(createSPTMessage(SPT_Message.Goodbye, null, null));

	}

	/**
	 * Send a RouteReply Message
	 * @param nextHop	unicast path next hop address
	 * @param pathDst	unicast path destination address
	 */
	protected final void sendRouteReply(SPT_LogicalAddress nextHop, SPT_LogicalAddress pathDst) {

		m_adapter.sendMulticastMessage(createSPTMessage(SPT_Message.RouteReply, nextHop, pathDst));

	}
	/**
	 * Send a RouteRequest Message
	 * @param pathDst	unicast path destination address
	 */
	protected final void sendRouteRequest(SPT_LogicalAddress pathDst) {

		m_adapter.sendMulticastMessage(createSPTMessage(SPT_Message.RouteRequest, null, pathDst));

	}

	
	/**
	 * create a SPT message
	 * @param type			message type
	 * @param nextHop		only for RouteReply message
	 * @param pathDst		only for RouteReply & RouteRequest message
	 */
	SPT_Message createSPTMessage(
		byte type,
		SPT_LogicalAddress nextHop,
		SPT_LogicalAddress pathDst) {
		SPT_Message spt_msg = null;

		switch (type) {
			case SPT_Message.Hello :
				/* use multicast to send hello */
				SPT_LogicalAddress root = neighbors.getRootLogicalAddress();
				SPT_AddressPair self = neighbors.getMyAddressPair();
				SPT_LogicalAddress parent = neighbors.getParentLogicalAddress();
				int cost = neighbors.getCost();

				int selfId = self.getSPTLogicalAddress().getId();

				//if( neighbors.isRoot() )selfId=-selfId;

				spt_msg =
					new SPT_Message(
						I_Node.PROTOSUB_MULTICAST,
						SPT_Message.Hello,
						config.getOverlayHash(),
						self,
						root,
						parent,
						cost,
						neighbors.getTimeStamp(),
						neighbors.getMsgAdjacencyTable(),
						(short)neighbors.getPathMetric()
					);

				break;

			case SPT_Message.Goodbye :
				spt_msg =
					new SPT_Message(
						I_Node.PROTOSUB_MULTICAST,
						SPT_Message.Goodbye,
						config.getOverlayHash(),
						neighbors.getMyAddressPair());

				break;
			case SPT_Message.RouteReply :
				spt_msg =
					new SPT_Message(
						I_Node.PROTOSUB_MULTICAST,
						SPT_Message.RouteReply,
						config.getOverlayHash(),
						neighbors.getMyAddressPair(),
						nextHop,
						pathDst);

				break;

			case SPT_Message.RouteRequest :
				spt_msg =
					new SPT_Message(
						I_Node.PROTOSUB_MULTICAST,
						SPT_Message.RouteRequest,
						config.getOverlayHash(),
						neighbors.getMyAddressPair(),
						pathDst);
				break;

			default :
				throw new IllegalArgumentException("couldn't create new SPT_Message (illegal msg type)");

		}

		return spt_msg;

	}


	//private NeighborhoodStatsTable ns;
	private I_Stats logicalAddressStats;

	/**
	 * Initialize the data structure for handling the statistics defined in this object.
	 * In this method, a StatsProcessor instance is created which contains all statistics
	 * , with each represented by a StatsElement instance, supported in this object.
	 */
	private void InitStatisticsStructure()
	{
		/*
		 * Initial StatsProcessor instance.
		 */
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
		
		statsPro.addStatsElement("NodeAdapter"/* name */, m_adapter/* I_Stats object */, 1/* minoccurs */, 1/* maxoccurs */);

		statsPro.addStatsElement("PhysicalAddress", new PhysicalAddressOperator(), 1, 1);

		statsPro.addStatsElement("StopBeacon", new StopBeaconTimeLength(), 1, 1);

		statsPro.addStatsElement(SPT_Node.BEACON_TIME_PROPERTY_NAME, new BeaconTimePeriod(), 1, 1);
		
		statsPro.addStatsElement(SPT_Node.PROP_NAME_INFO_DISP_PERIOD, new InfoDispPeriodValue(), 1, 1);

		statsPro.addStatsElement(SPT_Neighborhood.PROP_NAME_NEIGHBOR_TIMEOUT, new NeighborTimeoutValue(), 1, 1);

		statsPro.addStatsElement(SPT_RootHistory.PROP_NAME_RH_MAXAGE, new MsgMaxAgeValue(), 1, 1);

		statsPro.addStatsElement(SPT_Neighborhood.PROP_NAME_ADJACENCY_TIMEOUT, new AdjacencyTimeoutValue(), 1, 1);

		statsPro.addStatsElement(SPT_RoutingTable.PROP_NAME_ROUTE_ENTRY_TIMEOUT, new RouteCacheTimeoutValue(), 1, 1);

		statsPro.addStatsElement(SPT_RootHistory.PROP_NAME_RH_TIMEOUT, new RootHistoryTimeoutValue(), 1, 1);

		statsPro.addStatsElement("StartTime", new NodeStartTime(), 1, 1);

		statsPro.addStatsElement("LastChangeToNeighborhood", new LastChangeToNeighborhoodQuery(), 1, 1);

		statsPro.addStatsElement("StopTime", new NodeStopTime(), 1, 1);

		statsPro.addStatsElement("MaxReceiveBurstRate", new MaxReceiveBurstRateLimit(), 1, 1);

		statsPro.addStatsElement("MaxSendBurstRate", new MaxSendBurstRateLimit(), 1, 1);
		
		logicalAddressStats = new LogicalAddressOperator();
		statsPro.addStatsElement("LogicalAddress", logicalAddressStats, 1, 1);

		statsPro.addStatsElement("Root", new RootQuery(), 1, 1);

		statsPro.addStatsElement("Parent", new ParentOperator(), 1, 1);

		statsPro.addStatsElement("Cost", new CostQuery(), 1, 1);

		statsPro.addStatsElement("NumOfNeighbors", new NumOfCurrentNeighbors(), 1, 1);

		statsPro.addStatsElement("NeighborhoodFixed", new FixedNeighborhoods(), 1, 1);
		
		statsPro.addStatsElement(SPT_TopologyPolicy.PROP_NAME_TOPO_POLICY, new TopologyPolicyQuery(), 1, 1);
		
		statsPro.addStatsElement(SPT_TopologyPolicy.PROP_NAME_JUMP_THRESHOLD, new JumpThresholdValue(), 1, 1);
		
		statsPro.addStatsElement("PeerInfo", new PeerInfoQuery(), 1, 1);
	}


	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}

	/**
	 * Gets the information of a scalar statistics from an object which 
	 * implements this interface.
	 * @throws HyperCastStatsException if the statistics with given name is not
	 * supported by the object.
	 */
/*
	public String getStatsValue(String name) throws HyperCastStatsException
	{
		if (name.equals("LogicalAddress"))
		{
			return myAddressPair.getSPTLogicalAddress().toString();
		}
		else if (name.equals("PeerInfo")) {
			return "" + this.peerInfo();
		} else if (name.equals("StopBeacon")) {
			return "" + StopBeacon;
		} else if (name.equals("PhysicalAddress")) {
			return "" + neighbors.getMyAddressPair().getPhysicalAddress();
		} else if (name.equals(SPT_Node.BEACON_TIME_PROPERTY_NAME)) {
			return "" + BEACON_TIME;
		} else if (name.equals(SPT_Node.PROP_NAME_INFO_DISP_PERIOD)) {
			return ""+ INFO_DISP_PERIOD;
		} else if (name.equals(SPT_Neighborhood.PROP_NAME_NEIGHBOR_TIMEOUT)) {
			return "" + neighbors.neighborTimeout;
		} else if (name.equals(SPT_RootHistory.PROP_NAME_RH_MAXAGE)) {
			return "" + neighbors.root_history.msgMaxAge;
		} else if (name.equals(SPT_Neighborhood.PROP_NAME_ADJACENCY_TIMEOUT)) {
			return "" + neighbors.adjacencyTimeout;
		} else if (name.equals(SPT_RoutingTable.PROP_NAME_ROUTE_ENTRY_TIMEOUT)) {
			return "" + routing_table.routeEntryTimeout;
		} else if (name.equals(SPT_RootHistory.PROP_NAME_RH_TIMEOUT)) {
			return "" + neighbors.root_history.rootHistoryTimeout;
		} else if (name.equals("StartTime")) {
			return "" + startTime;
		} else if (name.equals("LastChangeToNeighborhood")) {
			return "" + lastChangeToNeighborhood;
		} else if (name.equals("StopTime")) {
			return "" + stopTime;
		} else if (name.equals("MaxReceiveBurstRate")) {
			return "" + maxReceiveRateInAHeartbeat;
		} else if (name.equals("MaxSendBurstRate")) {
			return "" + maxSendRateInAHeartbeat;
		} else if (name.equals("Root")){
			return "" + neighbors.getRootLogicalAddress();
		} else if (name.equals("Parent")){
			if (neighbors.getParentLogicalAddress() != null)
				return neighbors.getParentLogicalAddress().toString();
			else
				return getLogicalAddress().toString(); //return my address
		} else if (name.equals("Cost")){
			return "" + neighbors.getCost();
		} else if (name.equals("NumOfNeighbors")){
			return Integer.toString(neighbors.getNumOfNeighbors());
		} else if(name.equals("NeighborhoodFixed")){
			return ""+neighbors.NeighborhoodFixed;
		} else if(name.equals(SPT_TopologyPolicy.PROP_NAME_TOPO_POLICY)){
			return neighbors.Topo_Policy.Topo_Policy;
		} else if(name.equals(SPT_TopologyPolicy.PROP_NAME_JUMP_THRESHOLD)){
			return ""+neighbors.Topo_Policy.Jump_Threshold;
		}else
		{
			throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);
		}
		
	}
*/

	/**
	 * Sets the information of s writable scalar statistics to an object which 
	 * implements this interface.
	 * @throws HyperCastStatsException if the statistics with given name is not
	 * supported by the object or read-only, or if value does not conform with 
	 * expected format.
	 */
/*
	public String setStatsValue(String name, String newValue) throws HyperCastStatsException
	{
		String returnValue = newValue;
		
		if (name.equals("LogicalAddress"))
		{
			this.setLogicalAddress(this.createLogicalAddress(newValue));
		}
		else if (name.equals("StopBeacon")) {
			if (newValue.equals("true"))
				StopBeacon = true;
			else
				StopBeacon = false;
		} else if (name.equals(SPT_Node.BEACON_TIME_PROPERTY_NAME)) {
			BEACON_TIME = Integer.parseInt(newValue);
		} else if (name.equals(SPT_Node.PROP_NAME_INFO_DISP_PERIOD)) {
			INFO_DISP_PERIOD=Integer.parseInt(newValue);
			if(INFO_DISP_PERIOD>0){
				m_adapter.setTimer(new Integer(INFO_DISP_TIMER_INDEX), INFO_DISP_PERIOD);
			}
			return ""+INFO_DISP_PERIOD;
		} else if (name.equals(SPT_Neighborhood.PROP_NAME_NEIGHBOR_TIMEOUT)) {
			neighbors.neighborTimeout = Integer.parseInt(newValue);
		} else if (name.equals(SPT_RootHistory.PROP_NAME_RH_MAXAGE)) {
			neighbors.root_history.msgMaxAge = Integer.parseInt(newValue);
		} else if (name.equals(SPT_Neighborhood.PROP_NAME_ADJACENCY_TIMEOUT)) {
			neighbors.adjacencyTimeout = Integer.parseInt(newValue);
		} else if (name.equals(SPT_RoutingTable.PROP_NAME_ROUTE_ENTRY_TIMEOUT)) {
			routing_table.routeEntryTimeout = Integer.parseInt(newValue);
		} else if (name.equals(SPT_RootHistory.PROP_NAME_RH_TIMEOUT)) {
			neighbors.root_history.rootHistoryTimeout = Integer.parseInt(newValue);
		} else if (name.equals("Parent")){	
			SPT_LogicalAddress la=new SPT_LogicalAddress(newValue);
			neighbors.Topo_Policy.NextParent=la;
		}
		else if(name.equals("NeighborhoodFixed")){
			if(newValue.equals("true")){
				neighbors.NeighborhoodFixed=true;
			}else{
				neighbors.NeighborhoodFixed=false;
			}
			
		} else if(name.equals(SPT_TopologyPolicy.PROP_NAME_JUMP_THRESHOLD)){
			neighbors.Topo_Policy.Jump_Threshold = Integer.parseInt(newValue);
		} else
		{
			throw new HyperCastStatsException(name, HyperCastStatsException.StatisticNotFound);
		}
		
		return returnValue;
	}
*/

	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("SPT_Node:getStats: xpath string is "+xpath.getPatternString());
		//System.out.println("SPT_Node:getStats: call statsPro.getStatsResult(doc, xpath)");
		return statsPro.getStatsResult(doc, xpath);
	}

	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		//System.out.println("SPT_Node:setStats: xpath string is "+xpath.getPatternString());
		//System.out.println("SPT_Node:setStats: call statsPro.setStatsResult(doc, xpath, newValue)");
		return statsPro.setStatsResult(doc, xpath, newValue);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		//System.out.println("SPT_Node:getReadSchema: xpath string is "+xpath.getPatternString());
		//System.out.println("SPT_Node:getReadSchema: call statsPro.getReadSchemaResult(doc, xpath)");
		return statsPro.getReadSchemaResult(doc, xpath);
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return statsPro.getWriteSchemaResult(doc, xpath);
	}


	/* (non-Javadoc)
	 * @see hypercast.I_AdapterCallback#timerExpired(java.lang.Object)
	 */
	public void timerExpired(Object timerID) {
		int timer_index = ((Integer) timerID).intValue();
		timerExpired(timer_index);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_Node#getParent(hypercast.I_LogicalAddress, int)
	 */
	public I_AddressPair[] getParent(I_LogicalAddress root, int index) {
		return getParent(root);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_Node#getChildren(hypercast.I_LogicalAddress, int)
	 */
	public I_AddressPair[] getChildren(I_LogicalAddress root, int index) {
		return getChildren(root);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_Node#getAllNeighbors(int)
	 */
	public I_AddressPair[] getAllNeighbors(int index) {
		return getAllNeighbors();
	}

	/* (non-Javadoc)
	 * @see hypercast.I_Node#getMyAddressPair(int)
	 */
	public I_AddressPair getMyAddressPair(int index) {
		return getAddressPair();
	}

	/* (non-Javadoc)
	 * @see hypercast.I_Node#createLogicalAddress(byte[], int, int)
	 */
	public I_LogicalAddress createLogicalAddress(byte[] laddr, int offset, int index) {
		return createLogicalAddress(laddr, offset);
	}
	/** all these API are added by jianping 3/20/04 */
	/* (non-Javadoc)
	 * @see hypercast.I_Node#getPhysicalAddress()
	 */
	public I_PhysicalAddress getPhysicalAddress() {
		return this.getAddressPair().getPhysicalAddress();
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Node#getAllNeighbors(hypercast.I_LogicalAddress)
	 */
	public I_AddressPair[] getAllNeighbors(I_LogicalAddress root) {
		return this.getAllNeighbors();
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Node#getMyAddressPair(hypercast.I_LogicalAddress)
	 */
	public I_AddressPair getMyAddressPair(I_LogicalAddress root) {
		return this.getAddressPair();
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Node#getPhysicalAddress(int)
	 */
	public I_PhysicalAddress getPhysicalAddress(int index) {
		return this.getPhysicalAddress();
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Node#getLogicalAddress(int)
	 */
	public I_LogicalAddress getLogicalAddress(int index) {
		return this.getLogicalAddress();
	}

	/** Verify the previous hop of the message. If checkmode is set to "neighborCheck",
	 * this method checks if the previous hop is a neighbor; otherwise this method checks
	 * if the previous hop is a valid sender of OL messages.
	 */
	public boolean previousHopCheck(
		I_LogicalAddress src,
		I_LogicalAddress dst,
		I_LogicalAddress prehop) {
		//update the routing table to learning the routing info from this packet
		if (neighbors.containsNeighbor((SPT_LogicalAddress) prehop)) {
			SPT_LogicalAddress sptsrc = (SPT_LogicalAddress) src;
			SPT_LogicalAddress sptdst = (SPT_LogicalAddress) dst;
			SPT_LogicalAddress sptprehop = (SPT_LogicalAddress) prehop;
			
			routing_table.updateRoutingEntry(sptsrc, sptprehop);
			
			if(sptdst!=null){
				SPT_RoutingEntry re=routing_table.getRoutingEntry(sptdst);
				if(re!=null && re.getNextHop().equals(sptprehop)){
					//it's a loop
					//remove this entry
					routing_table.removeRouting(sptdst);
				}
			}
			
		}

		if (checkmode.equals("neighborCheck")) {
			return neighbors.containsNeighbor((SPT_LogicalAddress) prehop);
		} else {
			if (dst == null) //multicast or flooding
				{
				I_AddressPair[] Parent = getParent(src);
				if (Parent.length != 1)
					return false; //no parent found

				I_LogicalAddress Parent_LA = Parent[0].getLogicalAddress();
				return ((SPT_LogicalAddress) Parent_LA).equals(prehop);
			} else //unicast
				{
				I_AddressPair[] Children = getChildren(dst);
				if (Children == null)
					return false;

				for (int i = 0; i < Children.length; i++) {
					if ((Children[i].getLogicalAddress()).equals((SPT_LogicalAddress) prehop))
						return true;
				}
			}
			return false;
		}
	}

	/**
	 * Set notification handler.
	 */
	public void setNotificationHandler(NotificationHandler nh) {
		n_handler = nh;
		//neighbors.setNotificationHandler(nh);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_Node#createLogicalAddress(java.lang.String)
	 */
	public I_LogicalAddress createLogicalAddress(String laStr) {
		return new SPT_LogicalAddress(laStr);
	}


	/** Sets the logical address of this node to the specified one.
	 */
	public void setLogicalAddress(I_LogicalAddress la)
	{
		if (!(la instanceof SPT_LogicalAddress))
		{
		   throw new IllegalArgumentException("SPT_Node: setLogicalAddress: the logical address type of the parameter is incorrect.");
		}

		neighbors.clear();
		
		this.myAddressPair.getSPTLogicalAddress().changeLogicalAddress((SPT_LogicalAddress)la);

		//Report event LACHANGED to notification handler.
		if (n_handler != null) {
			NODE_LOGICALADDRESSCHANGED ne = new NODE_LOGICALADDRESSCHANGED(m_adapter.getCurrentTime(), null);
			n_handler.eventOccurred(ne);
		}

	}
	

	/**
	 * This class manages simple statistic "LogicalAddress".
	 */
	class LogicalAddressOperator extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return myAddressPair.getSPTLogicalAddress().toString();
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			setLogicalAddress(createLogicalAddress(newValue));
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:simpleType"/* type */, 
												"xsd:String"/* restriction base */, "\\d+"/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "PhysicalAddress".
	 */
	class PhysicalAddressOperator extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + neighbors.getMyAddressPair().getPhysicalAddress();
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}


	/**
	 * This class manages simple statistic "StartTime".
	 */
	class NodeStartTime extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + startTime;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "StopTime".
	 */
	class NodeStopTime extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return ""+stopTime;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}


	/**
	 * This class manages simple statistic "NumOfNeighbors".
	 */
	class NumOfCurrentNeighbors extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return Integer.toString(neighbors.getNumOfNeighbors());
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "MaxReceiveBurstRate".
	 */
	class MaxReceiveBurstRateLimit extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + maxReceiveRateInAHeartbeat;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Double"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "MaxSendBurstRate".
	 */
	class MaxSendBurstRateLimit extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + maxSendRateInAHeartbeat;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Double"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "StopBeacon".
	 */
	class StopBeaconTimeLength extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + StopBeacon;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			if (newValue.equals("true"))
				StopBeacon = true;
			else
				StopBeacon = false;
				
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}

		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "BeaconTime".
	 */
	class BeaconTimePeriod extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + BEACON_TIME;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			BEACON_TIME = Integer.parseInt(newValue);
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "InfoDispPeriod".
	 */
	class InfoDispPeriodValue extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return ""+ INFO_DISP_PERIOD;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			INFO_DISP_PERIOD=Integer.parseInt(newValue);
			if(INFO_DISP_PERIOD>0){
				m_adapter.setTimer(new Integer(INFO_DISP_TIMER_INDEX), INFO_DISP_PERIOD);
			}
			return ""+INFO_DISP_PERIOD;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "NeighborTimeout".
	 */
	class NeighborTimeoutValue extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + neighbors.neighborTimeout;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			neighbors.neighborTimeout = Integer.parseInt(newValue);
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "MsgMaxAge".
	 */
	class MsgMaxAgeValue extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + neighbors.root_history.msgMaxAge;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			neighbors.root_history.msgMaxAge = Integer.parseInt(newValue);
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "AdjacencyTimeout".
	 */
	class AdjacencyTimeoutValue extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + neighbors.adjacencyTimeout;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			neighbors.adjacencyTimeout = Integer.parseInt(newValue);
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "RouteCacheTimeout".
	 */
	class RouteCacheTimeoutValue extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + routing_table.routeEntryTimeout;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			routing_table.routeEntryTimeout = Integer.parseInt(newValue);
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "RootHistoryTimeout".
	 */
	class RootHistoryTimeoutValue extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + neighbors.root_history.rootHistoryTimeout;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			neighbors.root_history.rootHistoryTimeout = Integer.parseInt(newValue);
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "LastChangeToNeighborhood".
	 */
	class LastChangeToNeighborhoodQuery extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + lastChangeToNeighborhood;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}
		
		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "Root".
	 */
	class RootQuery extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + neighbors.getRootLogicalAddress();
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:simpleType"/* type */, 
												"xsd:String"/* restriction base */, "\\d+"/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "Parent".
	 */
	class ParentOperator extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			if (neighbors.getParentLogicalAddress() != null)
				return neighbors.getParentLogicalAddress().toString();
			else
				return getLogicalAddress().toString(); //return my address
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			SPT_LogicalAddress la=new SPT_LogicalAddress(newValue);
			neighbors.Topo_Policy.NextParent=la;
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:simpleType"/* type */, 
												"xsd:String"/* restriction base */, "\\d+"/* pattern value */);
		}

		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:simpleType"/* type */, 
												"xsd:String"/* restriction base */, "\\d+"/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "Cost".
	 */
	class CostQuery extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + neighbors.getCost();
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "NeighborhoodFixed".
	 */
	class FixedNeighborhoods extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return ""+neighbors.NeighborhoodFixed;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			if(newValue.equals("true")){
				neighbors.NeighborhoodFixed=true;
			}else{
				neighbors.NeighborhoodFixed=false;
			}
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Boolean"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "TopologyPolicy".
	 */
	class TopologyPolicyQuery extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return neighbors.Topo_Policy.Topo_Policy;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "JumpThreshold".
	 */
	class JumpThresholdValue extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return ""+neighbors.Topo_Policy.Jump_Threshold;
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			neighbors.Topo_Policy.Jump_Threshold = Integer.parseInt(newValue);
			return newValue;
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
		
		/** 
		 * @see I_Stats#getWriteSchema(Document, XPath)
		 */
		public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Integer"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "PeerInfo".
	 */
	class PeerInfoQuery extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return "" + peerInfo();
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchema(Document, XPath)
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:String"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	public boolean isRoot(){
		return neighbors.isRoot();
	}
}
