/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/
package hypercast.SPT;

import java.util.Enumeration;
import java.util.Hashtable;

import hypercast.HyperCastConfig;
import hypercast.I_UnicastAdapter;
import hypercast.util.XmlUtil;


/** 
 * Routing Table 
 *  This is a hash table of routing entries
 * 
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, Jan 13, 2004
 */
class SPT_RoutingTable {
	/** the hash table of entries */
	private Hashtable routing_table;
	
	/** the neighbors table */
	SPT_Neighborhood neighbors;
	
	/** for getting time */
	I_UnicastAdapter adapter;
	
	/** config object */
	HyperCastConfig config;
	
	final static String PROP_NAME_ROUTE_ENTRY_TIMEOUT= "RouteCacheTimeout";
	
	/** timeout value for an entry */
	long routeEntryTimeout;
	
	/**
	 * Constructor
	 *
	 */
	public SPT_RoutingTable(HyperCastConfig c, String prefix, I_UnicastAdapter adpt, SPT_Neighborhood nb) {
		config=c;
		adapter=adpt;
		neighbors=nb;
		
		routing_table = new Hashtable();
		//prefix+=".";

		//routeEntryTimeout = config.getIntProperty (prefix + PROP_NAME_ROUTE_ENTRY_TIMEOUT);
		routeEntryTimeout = config.getIntAttribute
				(XmlUtil.createXPath(prefix + PROP_NAME_ROUTE_ENTRY_TIMEOUT));
	
	}
	/**
	 * Clear entiries that have expired
	 */
	public void clearExpired() {
		for (Enumeration rt = routing_table.elements(); rt.hasMoreElements();) {
			SPT_RoutingEntry re = (SPT_RoutingEntry) rt.nextElement();

			if(!isValid(re)){
				removeRouting(re.getDest());
			}
		}
	}
	
	/**
	 * Remove an routing entry
	 * @param dest	the destination Id
	 */
	public void removeRouting(SPT_LogicalAddress dest) {
		routing_table.remove(dest);
	}
	
	/**
	 * Update an routing entry
	 * @param dId	destination Id
	 * @param nId	next hop Id
	 */
	public void updateRoutingEntry(SPT_LogicalAddress dId, SPT_LogicalAddress nId) {
		//update or add a routing entry
		SPT_RoutingEntry entry=new SPT_RoutingEntry(dId, nId, routeEntryTimeout, adapter);
		if(isValid(entry))
			routing_table.put(dId, entry);
	}
	/**
	 * get an routing entry
	 * @param dId	destination Id
	 * @return		the routing entry
	 */
	private SPT_RoutingEntry getEntry(SPT_LogicalAddress dId) {
		//find an routing entry without considering the timeout
		//try to search in two direction of routing

		SPT_RoutingEntry entry = (SPT_RoutingEntry) routing_table.get(dId);

		return entry;
	}
	/**
	 * get an routing entry that has not expired
	 * @param dId		destination Id
	 */
	public SPT_RoutingEntry getRoutingEntry(SPT_LogicalAddress dId) {
		//find an routing entry with considering the timeout

		SPT_RoutingEntry re = getEntry(dId);

		if (re != null) {
			if(!isValid(re)){
				//if the entry is too old or next hop not neighbor
				//remove it;
				removeRouting(re.getDest());
				re = null;
			}

		}

		return re;
	}
	private boolean isValid(SPT_RoutingEntry re){
		if(re.isExpired())
			return false;
		
		if(!neighbors.containsNeighbor(re.getNextHop()))
			return false;
			
		if(neighbors.containsNeighbor(re.getDest()))
			return false;
		
		if(neighbors.getRootLogicalAddress().equals(re.getDest()))
			return false;
		
		if(neighbors.getMyLogicalAddress().equals(re.getDest()))
			return false;
		
		if(neighbors.getMyLogicalAddress().equals(re.getNextHop()))
			return false;
		
		return true;
		
	}
	/**
	 * Show the entire table
	 */
	public String toString(){
		String result="";
		for(Enumeration rt=routing_table.elements();rt.hasMoreElements();){
			SPT_RoutingEntry re= (SPT_RoutingEntry)rt.nextElement();
			result=result+"Dst="+re.getDest()+" NextHop="+re.getNextHop()+" TS="+re.getLastUpdateTime()+"\n";
		}
		return result;
	}

}
/**
 * Routing Table Entry
 * 
 * @author Guangyu Dong, Jan 13, 2004
 */
class SPT_RoutingEntry {
	/** the destination Id */
	private SPT_LogicalAddress Dest;
	/** the last time the entry being updated */
	private long lastUpdateTime;
	/** the next hop Id */
	private SPT_LogicalAddress NextHop;
	
	/** for getting time */
	I_UnicastAdapter adapter;
	
	/** expiring time */
	private long timeout;

	/**
	 * Constructor
	 * @param dId	the destination
	 * @param rId	the next hop
	 * @param to	timeout
	 */
	public SPT_RoutingEntry(SPT_LogicalAddress dId, SPT_LogicalAddress rId, long to, I_UnicastAdapter adpt) {
		adapter=adpt;
		timeout=to;
		
		Dest = dId;
		lastUpdateTime = adapter.getCurrentTime();
		NextHop = rId;
	}
	/**
	 * Update the time stamp
	 */
	public void update() {
		lastUpdateTime = adapter.getCurrentTime();
	}
	/**
	 * Update next hop	
	 * @param rId	the next hop
	 */
	public void update(SPT_LogicalAddress rId) {
		NextHop = rId;
		lastUpdateTime = adapter.getCurrentTime();
	}
	/**
	 * @return	the destination Id
	 */
	public SPT_LogicalAddress getDest() {
		return Dest;
	}
	/**
	 * @return	the next hop Id
	 */
	public SPT_LogicalAddress getNextHop() {
		return NextHop;
	}
	/**
	 * @return	the last time the entry being updated
	 */
	public float getLastUpdateTime() {
		return lastUpdateTime;
	}
	/**
	 * Check if this entry expires
	 * @return		if this entry expires
	 */
	public boolean isExpired() {
		return (lastUpdateTime + timeout) < adapter.getCurrentTime();
	}

}
